﻿import time

import app.util
import app.servicecontrol
import app.defaults
import app.logutil
import app.network

AWAIT_TIMEOUT = 10*60

class AwaitStartupDependenciesAction(app.servicecontrol.ControllerAction):
    def __init__(self, services, logger):
        self.__services = services
        self.__servicesStartedSuccessfully = []
        super(AwaitStartupDependenciesAction, self).__init__(logger)

    def chain(self, action):
        assert isinstance(action, AwaitStartupDependenciesAction)

        return AwaitStartupDependenciesAction(
            self.__services + action.getInvolvedServices(),
            self.logger)

    def getInvolvedServices(self):
        return self.__services

    def confirm(self, isRedeployment=False):
        # Starting a service is nothing worth asking for confirmation for.
        return True

    def __getEndpointsToAwait(self):
        requiredEndpoints = []

        # Gather all dependencies of all services. Note that transitive
        # dependencies does not matter here.
        for service in self.__services:
            for endpoint in service.getStartupEndpointDependencies():
                # Ignore local dependencies
                if not endpoint.getProviderEndpoint().getMachine().isLocal():
                    requiredEndpoints.append(endpoint)

        return set(requiredEndpoints)

    def __awaitEndpoint(self, endpoint, timeoutSecsTotal, timeoutSecsPerTest=2):
        providerEndpoint = endpoint.getProviderEndpoint()

        warningIssued = False

        while timeoutSecsTotal > 0:
            if app.network.isPortListening(
                    providerEndpoint.getHost(),
                    providerEndpoint.getPort(),
                    self.logger,
                    timeoutSecsPerTest):

                if providerEndpoint.getPort() == 80 or providerEndpoint.getPort() == 443:
                    # It is a good sign that the port is listening, but it may be
                    # only a load balancer that we are talking to here. So also do a
                    # quick check on whether the endpoint responds to HTTP requests
                    # in any reasonable way that indicates it is available.
                    if app.network.isHttpListening(
                        providerEndpoint.getHost(),
                        providerEndpoint.getPort(),
                        self.logger,
                        timeoutSecsPerTest):
                        # Yep, HTTP looks ok.
                        return True
                else:
                    # Some other protocol, assume having an open port is enough.
                    return True

            # Back off before trying again.
            time.sleep(timeoutSecsPerTest)
            timeoutSecsTotal -= timeoutSecsPerTest

            if not warningIssued:
                # As we need to wait, issue a warning... once.
                self.printLog("warn", "Waiting for service '%s' on %s to become available" % (
                    endpoint.getProviderService().getName(),
                    endpoint.getProviderService().getMachine().getPhysicalName()))
                warningIssued = True

        self.printLog("error", "Giving up waiting for endpoint '%r' to become available" % endpoint)
        return False

    def execute(self):
        endpointsToAwait = self.__getEndpointsToAwait()

        for endpoint in endpointsToAwait:
            if not self.__awaitEndpoint(endpoint, AWAIT_TIMEOUT):
                return False

        return True

class AwaitException(Exception):
    pass
