﻿import app.util
import app.servicecontrol
import app.logutil
import app.integrity

class CheckDataIntegrityAction(app.servicecontrol.ControllerAction):
    def __init__(self, services, logger):
        self.__services = services
        super(CheckDataIntegrityAction, self).__init__(logger)

    def chain(self, action):
        assert isinstance(action, CheckDataIntegrityAction)

        return CheckDataIntegrityAction(
            self.__services + action.getInvolvedServices(),
            self.logger)

    def getInvolvedServices(self):
        return self.__services

    def __checkIntegrityOfServiceData(self, service):

        localOk = True
        unequalMatches = False
        dataInt = app.integrity.DataIntegrity([service])
        failList, failPatterns, successfulMatches = dataInt.checkDataIntegrity()

        if len(failList) != 0:
            self.printLog("error", "Checking data integrity of %s failed." % service.getName())
            self.printLog("warn", "The following file(s) failed data integrity check:")
            count = 1
            for fileName in failList:
                self.printLog("warn", "%s. %s" %(count, fileName))
                count += 1
            localOk = False

        if len(failPatterns) != 0:
            self.printLog("error", "The following pattern(s) didn't match any file(s):")
            count = 1
            for pattern in failPatterns:
                self.printLog("warn", "%s. %s" % (count, pattern))
                count += 1

            localOk = False

        if len(successfulMatches) > 1:
            matches = iter(successfulMatches.values())
            rand = next(matches)
            if not all(match == rand for match in matches):
                unequalMatches = True
                localOk = False

        if unequalMatches and service.getName() == "subversion":
            self.printLog("error", "Not all repostories have all the hooks.")


        if localOk:
            return True
        else:
            return False

    def execute(self):
        allOk = True

        failedServices = []
        for service in self.__services:
            try:
                if not self.__checkIntegrityOfServiceData(service):
                    failedServices.append(service)
                    self.printLog("error", "%s Failed" % service.getName().ljust(50))
            except Exception as e:
                self.printLog("error", str(e))
                self.logger.error(''.join(app.util.getStackTraceAndException()).strip())
                allOk = False

        if not allOk:
            self.printLog("error", "\nData Integrity Check failed. See %s for details." % (app.logutil.getFileName(self.logger) or "log"))
        else:
            if len(self.__services) > 1:
                self.printLog("title", "\nSummary of integrity checks:")
                for service in self.__services:
                    if service not in failedServices:
                        self.printLog("success", "%s Passed" % service.getName().ljust(50))
                    else:
                        self.printLog("error", "%s Failed" % service.getName().ljust(50))

                if len(failedServices) != 0:
                    self.printLog("error", "\nSome services have failed integrity checks. See above.")

        return allOk
