﻿import os
import time

import app.util
import app.servicecontrol
import app.batch
import app.defaults
import app.autodata
import app.logutil
import app.integrity
import app.options
import app.actions.lockedaction
import app.linuxservice
import app.selinuxutil
import app.mixins

class UndeployServiceAction(app.actions.lockedaction.LockedControllerAction):
    def __init__(self, services, logger):
        self.__services = services
        super(UndeployServiceAction, self).__init__(logger)

    def filterDeployedServices(self):
        deployedServices = [service for service in self.__services
                    if service.getController(self.logger).isDeployed()]

        return UndeployServiceAction(deployedServices, self.logger)

    def chain(self, action):
        assert isinstance(action, UndeployServiceAction)

        return UndeployServiceAction(
            self.__services + action.getInvolvedServices(),
            self.logger)

    def getInvolvedServices(self):
        return self.__services

    def confirm(self, isRedeployment=False):
        servicesToBeStopped = self.__getServicesToBeStopped()

        if not len(self.__services):
            # Empty action - there is nothing to confirm.
            return True

        self.printLog(
            "warn",
            "The following services will be %s: \n  %s" % (
                    "redeployed" if isRedeployment else "undeployed",
                    "\n  ".join(sorted([service.getName() for service in self.__services]))))

        if len(servicesToBeStopped):
            self.printLog(
                "warn",
                "This requires the following services to be stopped first: \n  %s" %
                        ("\n  ".join([service.getName() for service in servicesToBeStopped])))

        return "y" == app.util.readFromUser("Do you wish to stop these services and continue? [y/N]:", ["y","N"])

    def __getServicesToBeStopped(self):
        servicesToBeStopped = []

        for service in self.__services:
            if not service.getController(self.logger).canUndeploy():
                servicesToBeStopped.append(service)

        return servicesToBeStopped

    def __stopServicesForUndeploy(self):

        if len(self.__getServicesToBeStopped()) > 0:
            stopAction = app.landscape.Landscape.getSite() \
                .getLocalMachine() \
                .getController(self.logger) \
                .createStopAction(self.__getServicesToBeStopped())

            if not stopAction:
                return

            if not stopAction.execute():
                raise app.servicecontrol.ServiceControllerException("Failed to stop services")

    def __undeploy(self):
        self.printLog("title", "Undeploying Services");

        allOk = True
        try:
            app.selinuxutil.Policy.unloadSelinuxPolicy(self.logger, self.__services)
        except Exception:
            self.logger.error(''.join(app.util.getStackTraceAndException()).strip())
            allOk = False
        for service in self.__services:
            # A separate batch is run for each service so that we the order
            # corresponds to the order of the services list

            scriptsFolder = os.path.join(
                app.defaults.DEFAULT_SERVICE_DIR,
                service.getName(),
                "hooks",
                "undeploy")

            if os.path.isdir(scriptsFolder):
                self.printLog('info', "  Undeploying %s..." % service.getName())

                batch = app.batch.CreateRuntimeBatch(
                    self.logger,
                    app.landscape.Landscape.getSiteOptions())

                batch.addFolder(scriptsFolder)

                try:
                    batch.execute()
                except Exception:
                    self.logger.error(''.join(app.util.getStackTraceAndException()).strip())
                    allOk = False

            # Consider this service undeployed from now on.
            app.servicecontrol.DeploymentManager.markServiceAsUndeployed(service.getName())

        if not allOk:
            self.printLog("error", "\nUndeployment failed. See %s for details." % (app.logutil.getFileName(self.logger) or "log"))


    def __checkAllServiceReadyForUndeployment(self):
        for service in self.__services:
            controller = service.getController(self.logger)
            if not controller.canUndeploy():
                 raise app.servicecontrol.ServiceControllerException("Cannot undeploy service '%s' while in status '%s'" %
                                (service.getName(), controller.getStatus()))

    def lockedExecute(self):
        self.__stopServicesForUndeploy()
        self.__checkAllServiceReadyForUndeployment()
        self.__undeploy()



class DeployServiceAction(app.actions.lockedaction.LockedControllerAction, app.mixins.SelinuxMixin):
    def __init__(self, services, logger):
        self.__services = services
        self.__runtimeOptions = None

        super(DeployServiceAction, self).__init__(logger)

    def chain(self, action):
        assert isinstance(action, DeployServiceAction)

        return DeployServiceAction(
            self.__services + action.getInvolvedServices(),
            self.logger)

    def getInvolvedServices(self):
        return self.__services

    def confirm(self):
        if not os.path.exists(app.defaults.DEFAULT_RUNTIME_DIR):
            return True

        else:
            self.printLog("info", "Checking integrity of TeamForge configurations ...".ljust(60), False)
            checksumCfgFile = app.defaults.CHECKSUM_CONFIG_FILE

            if os.path.isfile(app.defaults.OLD_CHECKSUM_CONFIG_FILE):
                checksumCfgFile = app.defaults.OLD_CHECKSUM_CONFIG_FILE

            confInt = app.integrity.ConfigurationIntegrity(self.__services, checksumCfgFile)
            intCheckfailList = confInt.checkRuntimeIntegrity()

            if len(intCheckfailList) == 0:
                self.printLog("success", "Passed")
                return True
            else:
                self.printLog("error", "Failed")
                self.printLog("warn", "The following configuration file(s) have local modifications that will be lost:")
                count = 1
                for fileName in intCheckfailList:
                    self.printLog("warn", "%s. %s" %(count, fileName))
                    count += 1
                return "y" == app.util.readFromUser("Do you wish to proceed anyway? [y/N]:", ["y","N"])

    def __checkAllServiceReadyForDeployment(self):
        for service in self.__services:
            controller = service.getController(self.logger)
            if not controller.canDeploy():
                 raise app.servicecontrol.ServiceControllerException("Cannot deploy service '%s' while in status '%s'" %
                                (service.getName(), controller.getStatus()))

    def __getServicesForRuntimeOptionGeneration(self):
        # When generating runtime options, we need to include the runtime options of dependencies
        # as well. This makes sure that any "global" options introduced by dependencies are
        # available. This is primarily to keep backwards compatibility.

        servicesIncludingDependencies = set([])
        for service in self.__services:
            servicesIncludingDependencies.add(service)

            for dependency in service.getDeploymentDependencies(True):
                servicesIncludingDependencies.add(dependency)

        # Now that we have the list of services, we need to get them into
        # deployment order. This is important because the scriptlets may refer
        # to options set by other services it depends on.

        services = []
        for service in app.landscape.Landscape.getSite().getLocalMachine().getServicesInDeploymentOrder():
            if service in servicesIncludingDependencies:
                services.append(service)

        return services

    def __generateRuntimeOptions(self):
        siteOptions = app.landscape.Landscape.getSiteOptions()
        siteOptionsPath = app.landscape.Landscape.getSiteOptionsPath()

        self.printLog("title", "Runtime Options Generation");

        tokenDef = app.tokens.defines.TokenDef()
        runtimeOptions = siteOptions.deriveRuntimeOptions(tokenDef)

        self.printLog('info', "  %s"%"Generating runtime options ...".ljust(58), False)
        startTime = time.time()

        servicesIncludingDependencies = self.__getServicesForRuntimeOptionGeneration()

        # runtime-option-defaults scripts should be getting executed before
        # runtime-option-overrides scripts because some tokens which are
        # configured in overrides may depended on runtime-option-defaults.

        # Apply defaults for anything the user did not set explicitly in his site-options.conf
        runtimeOptionDefaultsBatch = app.batch.RuntimeOptionsBatch(
            app.batch.OverrideSemantics.KEEP_EXISTING_OPTIONS,
            self.logger,
            runtimeOptions)
        for service in servicesIncludingDependencies:
            scriptsFolder = os.path.join(
                app.defaults.DEFAULT_SERVICE_DIR,
                service.getName(),
                "hooks",
                "runtime-option-defaults")

            if os.path.isdir(scriptsFolder):
                runtimeOptionDefaultsBatch = runtimeOptionDefaultsBatch.addFolder(scriptsFolder)

        runtimeOptions = runtimeOptionDefaultsBatch.createRuntimeOptions()

        # Apply overrides to tweak whatever the user set in his site-options.conf
        runtimeOptionOverridesBatch = app.batch.RuntimeOptionsBatch(
            app.batch.OverrideSemantics.OVERRIDE_EXISTING_OPTIONS,
            self.logger,
            runtimeOptions)
        for service in servicesIncludingDependencies:
            scriptsFolder = os.path.join(
                app.defaults.DEFAULT_SERVICE_DIR,
                service.getName(),
                "hooks",
                "runtime-option-overrides")


            if os.path.isdir(scriptsFolder):
                runtimeOptionOverridesBatch = runtimeOptionOverridesBatch.addFolder(scriptsFolder)

        runtimeOptions = runtimeOptionOverridesBatch.createRuntimeOptions()

        elapsedTime = "{0:.2f}".format(time.time() - startTime)
        self.printLog('success', 'Done'.ljust(8), False)
        self.printLog('info', '%s secs' % elapsedTime.rjust(6))

        # Apply auto data and update AUTO_DATA in site-options.conf. This covers
        # both tokens set to $auto$ in site-options.conf and tokens set to $auto$
        # in option snippets.
        autoData = app.autodata.AutoData.readFromSiteOptions(tokenDef, siteOptions)

        self.printLog("info", "  Populating auto-generatable runtime options ...".ljust(60), False)
        startTime = time.time()

        runtimeOptions = runtimeOptions.applyAutoData(autoData)
        autoData.updateSiteOptionsFile(siteOptionsPath)

        elapsedTime = "{0:.2f}".format(time.time() - startTime)
        self.printLog('success', 'Done'.ljust(8), False)
        self.printLog('info', '%s secs' % elapsedTime.rjust(6))

        self.__runtimeOptions = runtimeOptions

    def __deploy(self):
        # From this point on, it is okay if scripts call getRuntimeConfiguration() as the
        # runtime options are fully populated and immutable now.
        #
        # Note that there are still many scriptlets that call getRuntimeConfiguration.
        app.util.setRuntimeConfiguration(self.__runtimeOptions)

        try:
            for service in self.__services:
                # A separate batch is run for each service so that we the order
                # corresponds to the order of the services list

                app.selinuxutil.Policy.deployServicePolicy(self.logger, service)

                scriptsFolder = os.path.join(
                    app.defaults.DEFAULT_SERVICE_DIR,
                    service.getName(),
                    "hooks",
                    "deploy")

                if os.path.isdir(scriptsFolder):
                    self.printLog('title', "Deploying %s..." % service.getName())

                    batch = app.batch.CreateRuntimeBatch(
                        self.logger,
                        self.__runtimeOptions)

                    batch.addFolder(scriptsFolder)

                    batch.execute()

                # Consider this service deployed from now on.
                app.servicecontrol.DeploymentManager.markServiceAsDeployed(service.getName())

                postDeployAction = service.getController(self.logger).createPostDeployAction()
                if postDeployAction:
                    if not postDeployAction.execute():
                        raise Exception("Post Deploy action for service '%s' failed" % service.getName())
        except Exception as e:
            self.printLog("error", "\nDeployment failed. See %s for details." % (app.logutil.getFileName(self.logger) or "log"))
            raise e

    def __updateIntegrityChecksums(self):
        # This function used to calculate the checksum of teamforge configuration files
        # and write the checksum in file
        confInt = app.integrity.ConfigurationIntegrity(self.__services, app.defaults.CHECKSUM_CONFIG_FILE)
        confInt.writeIntegrityChecksums()

    def __reloadSystemd(self):
        # Let Systemd reload configuration - we do this here (rather than in a deploy script)
        # to make sure it happens at the end of the deployment process.

        app.linuxservice.ServiceManager.reload()

    def __installLegacyAddOns(self):
        # Find the ctcore service.
        ctfcoreService = next((s for s in self.__services if s.getName() == "ctfcore-app"), None)
        if not ctfcoreService:
            # Service is not being deployed, so addons can be ignored.
            return

        addOnsInstPath = os.path.join(app.defaults.DEFAULT_SITE_DIR, "add-ons")
        addOnsRequiringLegacyTokens = [
            "teamforge_cli_server",
            "CN-cliserver",
            "ctf_customization_installer",
            "teamforge_cli"]
        if self.__runtimeOptions.get("DATABASE_TYPE", None) != "oracle" and  \
            any(os.path.isdir(os.path.join(addOnsInstPath, name)) for name in addOnsRequiringLegacyTokens):

            # These addons need access to the database but do not declare proper required endpoints.
            # So we take the endpoints which CTF Core uses and use their value for generating the
            # legacy DATABASE tokens.
            # Note that the DATABASE_{HOST, PORT} tokens always refer to Postgres. Therefore, we do not
            # write these tokens when CTF Core is configured to use Oracle.

            self.logger.info("Legacy addon detected, adding additional tokens to runtime-options.conf")
            databaseEndpoint = ctfcoreService.getRequiredEndpoint("jdbc", "ctfcore-database").getProviderEndpoint()

            app.options.PropertiesFile.updateFile(
                app.defaults.RUNTIME_OPTIONS_CONF,
                "DATABASE_HOST",
                databaseEndpoint.getHost(),
                "This token is deprecated and is intended for legacy addons only")
            app.options.PropertiesFile.updateFile(
                app.defaults.RUNTIME_OPTIONS_CONF,
                "DATABASE_PORT",
                databaseEndpoint.getPort(),
                "This token is deprecated and is intended for legacy addons only")

        addOnsPkgPath  = os.path.join(app.defaults.DEFAULT_SITE_DIR, "etc", "add-ons")
        if not (os.path.isdir(addOnsPkgPath)):
            # No AddOns to (re) run installer for.
            return

        # Get all files ending with *.pkg in addOnsPkgPath
        addOnsPkgFiles = filter (lambda pkgfile: pkgfile.endswith(".pkg"),  os.listdir(addOnsPkgPath))

        if len(addOnsPkgFiles) == 0:
            return

        self.printLog("title", "Installing AddOns")

        for pkgFile in (sorted(addOnsPkgFiles)):
            # Extract add-on package name from package file name
            pkgName  = pkgFile.split(".pkg")[0]
            addOnDir = os.path.join(addOnsInstPath, pkgName)

            # Skip if add-on install script doesnot exist in add-on directory
            if not os.path.exists(os.path.join(addOnDir, "install")):
                self.logger.info("AddOn '%s' does not have an install script, skipping" % pkgName)
                continue

            self.printLog('info', "  %s" % pkgName.ljust(58), False)
            try:
                output = app.util.executeCommand("cd %s; ./install --runtime --verbose" % addOnDir)
                self.printLog('success', 'Done'.ljust(8))
                self.logger.info(output)
            except Exception, msg:
                self.printLog('error', 'Failed')
                self.logger.error(str(msg) + ': ' + ''.join(app.util.getStackTraceAndException()).strip())
                # Continue installing addons anyway.

    def lockedExecute(self):
        try:
            self.preSelinuxMode = self.getSelinuxMode()
            # Set selinux mode to permissive in deploy
            if self.preSelinuxMode == "Enforcing":
                self.setSelinuxPermissive()

            self.__checkAllServiceReadyForDeployment()
            self.__generateRuntimeOptions()
            self.__deploy()
            self.__installLegacyAddOns()
            self.__updateIntegrityChecksums()
            self.__reloadSystemd()
        finally:
            # Reset to the mode whatever it was before
            self.resetSelinuxMode(self.preSelinuxMode)

