﻿import os
import app.servicecontrol
import app.servicemanifest
import app.batch
import app.defaults
import app.logutil
import app.batch
from  app.selinuxutil import Policy
import app.util

class FoldersActionMixin(object):

    def applyServiceFolders(self,
                            service,
                            forceFolderRecreation,
                            forceReapplyPermissions,
                            forceReapplySelinuxPermissions,
                            checkAndApply=False,
                            checkAndApplySELinux=False):
        # Run hooks, if any.
        scriptsFolder = os.path.join(
            app.defaults.DEFAULT_SERVICE_DIR,
            service.getName(),
            "hooks",
            "folder")

        if os.path.isdir(scriptsFolder):
            if forceFolderRecreation:
                # Delete...
                app.batch.FolderDeletionBatch(
                    self.logger,
                    app.util.getRuntimeConfiguration()) \
                    .addFolder(scriptsFolder) \
                    .execute()

            if not forceReapplyPermissions or not forceReapplySelinuxPermissions:
                app.batch.FolderCreationBatch(
                    self.logger,
                    app.util.getRuntimeConfiguration()) \
                    .addFolder(scriptsFolder) \
                    .execute()

            if checkAndApply:
                # Apply permissions...
                app.batch.CheckAndApplyFolderPermissionsBatch(
                    self.logger,
                    app.util.getRuntimeConfiguration()) \
                    .addFolder(scriptsFolder) \
                    .execute()

            elif forceReapplyPermissions:
                # Apply permissions...
                self.printLog('title', "Applying permissions for %s..." % service.getName())
                app.batch.FolderPermissionsBatch(
                    self.logger,
                    app.util.getRuntimeConfiguration()) \
                    .addFolder(scriptsFolder) \
                    .execute()

            if checkAndApplySELinux:
                # Apply selinux permissions
                app.batch.CheckAndApplySELinuxPermissionsBatch(
                    self.logger,
                    app.util.getRuntimeConfiguration()) \
                    .addFolder(scriptsFolder) \
                    .execute()

            elif forceReapplySelinuxPermissions:
                # Apply selinux policy
                self.printLog('title', "Applying selinux policy for %s..." % service.getName())
                app.batch.FolderSelinuxPolicyBatch(
                    self.logger,
                    app.util.getRuntimeConfiguration()) \
                    .addFolder(scriptsFolder) \
                    .execute()

        # Handle folders described in manifest.
        for folder in service.getFolders():
            if forceReapplyPermissions:
                self.logger.info("Calling Apply for %s, with checkAndApply=%s" % (folder,checkAndApply))
            if forceReapplySelinuxPermissions:
                self.logger.info("Calling Apply Selinux for %s" % folder)
            folder.apply(forceFolderRecreation, forceReapplyPermissions, forceReapplySelinuxPermissions, checkAndApply, checkAndApplySELinux)

class ApplyPermissionsAction(app.servicecontrol.ControllerAction, FoldersActionMixin):
    def __init__(self, services, logger):
        self.__services = services
        super(ApplyPermissionsAction, self).__init__(logger)

    def chain(self, action):
        assert isinstance(action, ApplyPermissionsAction)

        return ApplyPermissionsAction(
            self.__services + action.getInvolvedServices(),
            self.logger)

    def getInvolvedServices(self):
        return self.__services

    def execute(self):
        allOk = True

        for service in self.__services:
            try:
                self.applyServiceFolders(service, False, True, False)
            except Exception as e:
                self.printLog("error", str(e))
                self.logger.error(''.join(app.util.getStackTraceAndException()).strip())
                allOk = False

        if not allOk:
            self.printLog("error", "\nSetting permissions failed. See %s for details." % (app.logutil.getFileName(self.logger) or "log"))

        return allOk

class ApplySelinuxPolicyAction(app.servicecontrol.ControllerAction, FoldersActionMixin):
    def __init__(self, services, logger):
        self.__services = services
        super(ApplySelinuxPolicyAction, self).__init__(logger)

    def chain(self, action):
        assert isinstance(action, ApplySelinuxPolicyAction)

        return ApplySelinuxPolicyAction(
            self.__services + action.getInvolvedServices(),
            self.logger)

    def getInvolvedServices(self):
        return self.__services

    def checkAndLoadPolicy(self, service):
        srvcLst = service.getDeploymentDependencies(True) + [service]
        srvcLst = [s for s in srvcLst if len(Policy.listPolicy('tf_'+s.getName())) == 0]
        map(lambda s:Policy.deployServicePolicy(self.logger, s), srvcLst)

    def execute(self):
        allOk = True

        for service in self.__services:
            try:
                self.checkAndLoadPolicy(service)
                self.applyServiceFolders(service, False, False, True)
            except Exception as e:
                self.printLog("error", str(e))
                self.logger.error(''.join(app.util.getStackTraceAndException()).strip())
                allOk = False

        if not allOk:
            self.printLog("error", "\nApplying selinux policy failed. See %s for details." % (app.logutil.getFileName(self.logger) or "log"))

        return allOk

class UnloadSelinuxPolicyAction(app.servicecontrol.ControllerAction, FoldersActionMixin):
    def __init__(self, services, logger):
        self.__services = services
        super(UnloadSelinuxPolicyAction, self).__init__(logger)

    def chain(self, action):
        assert isinstance(action, UnloadSelinuxPolicyAction)

        return UnloadSelinuxPolicyAction(
            self.__services + action.getInvolvedServices(),
            self.logger)

    def getInvolvedServices(self):
        return self.__services

    def execute(self):
        allOk = True

        try:
            Policy.unloadSelinuxPolicy(self.logger, self.__services)
        except Exception, e:
            self.printLog("error", str(e))
            self.logger.error(''.join(app.util.getStackTraceAndException()).strip())
            allOk = False

        if not allOk:
            self.printLog("error", "\nUnloading selinux policy failed. See %s for details." % (app.logutil.getFileName(self.logger) or "log"))
        else:        
            self.printLog("info", "\n Site needs a restart for this unload selinux policy to take effect...")

        return allOk

