﻿import os

import app.util
import app.servicecontrol
import app.batch
import app.defaults
import app.logutil
import app.network

class BasePortAvailabilityCheckServiceAction(app.servicecontrol.ControllerAction):
    def __init__(self, services, logger):
        self.services = services
        super(BasePortAvailabilityCheckServiceAction, self).__init__(logger)

    def isListening(self, host, port):
        return app.network.isPortListening(host, port, self.logger)

    def chain(self, portAvailabilityCheck):
        assert isinstance(portAvailabilityCheck, BasePortAvailabilityCheckServiceAction)

        return BasePortAvailabilityCheckServiceAction(
            self.services + portAvailabilityCheck.getInvolvedServices(),
            self.logger)

    def getInvolvedServices(self):
        return self.services

    def execute(self):
        raise Exception("this class is abstract")

class ProvidedEndpointAvailableCheckServiceAction(BasePortAvailabilityCheckServiceAction):
    def execute(self):
        allOk = True

        for service in self.services:
            # Check if all ports the service intends to use are available
            for endpoint in service.getProvidedEndpoints():
                if endpoint.isViaOtherPort():
                    # Ignore this as the port is owned by a different service.
                    pass
                else:
                    # To be extra thorugh, we check both 'localhost' and the
                    # physical machine name.
                    if self.isListening("localhost", endpoint.getPort()):
                        self.printLog("error", "Port %d in use by different service" %
                                      (endpoint.getPort()))
                        allOk = False
                    elif self.isListening(endpoint.getPhysicalHost(), endpoint.getPort()):
                        self.printLog("error", "Port %d on %s in use by different service" %
                                      (endpoint.getPort(), service.getName()))
                        allOk = False

        return allOk

class ProvidedEndpointListeningCheckServiceAction(BasePortAvailabilityCheckServiceAction):
    def execute(self):
        allOk = True

        for service in self.services:
            # Check if all ports the service intends to use are available
            for endpoint in service.getProvidedEndpoints():
                if endpoint.isViaOtherPort():
                    # Ignore this as the port is owned by a different service.
                    pass
                else:
                    # Only check 'localhost' as we do not really know which IP
                    # the service is supposed to bind to.
                    if not self.isListening("localhost", endpoint.getPort()):
                        self.printLog("error", "Port %d on localhost not listening" %
                                      (endpoint.getPort()))
                        allOk = False

        return allOk

class RequiredEndpointListeningCheckServiceAction(BasePortAvailabilityCheckServiceAction):
    def execute(self):
        allOk = True

        for service in self.services:
            # Check if all ports the service intends to use are available
            for reqEndpoint in service.getRequiredEndpoints():
                if not reqEndpoint.isRequiredForStartup():
                    # Nevermind this one.
                    continue

                endpoint = reqEndpoint.getProviderEndpoint()

                if not self.isListening(endpoint.getHost(), endpoint.getPort()):
                    self.printLog("error", "Port %d on %s needed by %s not listening" %
                                    (endpoint.getPort(), endpoint.getPhysicalHost(), service.getName()))
                    allOk = False

        return allOk

class EnvironmentVerificationCheckAction(app.servicecontrol.ControllerAction):
    def __init__(self, services, logger):
        self.__services = services
        super(EnvironmentVerificationCheckAction, self).__init__(logger)

    def chain(self, envVerificationAction):
        assert isinstance(envVerificationAction, EnvironmentVerificationCheckAction)

        return EnvironmentVerificationCheckAction(
            self.__services + envVerificationAction.getInvolvedServices(),
            self.logger)

    def getInvolvedServices(self):
        return self.__services

    def execute(self):
        allOk = True

        for service in self.__services:
            # A separate batch is run for each service so that we the order
            # corresponds to the order of the services list

            scriptsFolder = os.path.join(
                app.defaults.DEFAULT_SERVICE_DIR,
                service.getName(),
                "hooks",
                "verification")

            if os.path.isdir(scriptsFolder):
                self.printLog('title', "Verifying environment for %s..." % service.getName())

                batch = app.batch.VerificationBatch(
                    self.logger,
                    app.landscape.Landscape.getSiteOptions())
                batch.addFolder(scriptsFolder)

                try:
                    batch.execute()
                except Exception:
                    self.logger.error(''.join(app.util.getStackTraceAndException()).strip())
                    allOk = False

        if not allOk:
            self.printLog("error", "\nVerification failed. See %s for details." % (app.logutil.getFileName(self.logger) or "log"))

        return allOk
