﻿import time

import app.util
import app.servicecontrol
import app.defaults
import app.logutil
import app.actions.lockedaction

# Seconds we give a service to publish a proper status - if a service needs more than
# this, its controller needs to override isStarting()
DAEMON_START_GRACE_PERIOD = 5

class StartServiceAction(app.actions.lockedaction.LockedControllerAction):
    def __init__(self, services, logger):
        self.__services = services
        self.__servicesStartedSuccessfully = []
        super(StartServiceAction, self).__init__(logger)

    def chain(self, action):
        assert isinstance(action, StartServiceAction)

        return StartServiceAction(
            self.__services + action.getInvolvedServices(),
            self.logger)

    def getInvolvedServices(self):
        return self.__services

    def confirm(self, isRedeployment=False):
        # Starting a service is nothing worth asking for confirmation for.
        return True

    def __killDaemon(self, service):
        pid = service.getController(self.logger).getPid()
        if pid:
            try:
                app.util.runCommand("kill -9 " + pid + " 2> /dev/null", self.logger)
            except:
                pass

    def __awaitServiceStartup(self, service, controller, timeout):
        timeStart = time.time()

        sleepInterval = 2
        try:
            self.printLog("info", "  Waiting for %s to come up..." % service.getName(), False)

            while time.time() < timeStart + timeout:
                status = controller.getStatus()
                if status == app.servicecontrol.ServiceStatus.AVAILABLE:
                    # Great, we are set.
                    self.logger.info("Service '%s' started successfully in %d seconds" %
                                     (service.getName(), time.time() - timeStart))
                    return True

                elif service.getController(self.logger).isDemandStart() and status == app.servicecontrol.ServiceStatus.STOPPED_DEMANDSTART:
                    # Demand start services may immediately go back to this state when there is
                    # nothing to do.
                    return True

                elif status == app.servicecontrol.ServiceStatus.DOOMED:
                    # It is futile to keep this service running, kill it.
                    self.printLog("error", "Killing service '%s' as it entered 'doomed' status" % service.getName())
                    self.__killDaemon(service)
                    return False

                elif status == app.servicecontrol.ServiceStatus.UNHEALTHY or \
                     status == app.servicecontrol.ServiceStatus.STARTING:
                    # Give it some more time.

                    self.printLog("info", ".", False)
                    sleepInterval = sleepInterval + 1
                    time.sleep(sleepInterval)

                else:
                    if time.time() < timeStart + DAEMON_START_GRACE_PERIOD:
                        # Give the service some time to publish a reasonable status.
                        time.sleep(sleepInterval)
                    else:
                        self.logger.error("Service '%s' unexpectedly entered status %s during startup - killing it" %
                                          (service.getName(), status))
                        self.__killDaemon(service)
                        return False

        finally:
            self.printLog("info", " ")

        # Timeout elapsed.
        self.printLog("error", "Service '%s' failed to start up within %d seconds - leaving it running" %
                      (service.getName(), timeout))
        return False

    def __startAllServices(self):
        self.printLog("title", "Starting Services")

        allOk = True

        for service in self.__services:
            try:
                startTimeout = service.getDaemonStartupInfo().getStartupTimeout()
                if not startTimeout:
                    startTimeout = 60

                controller = service.getController(self.logger)

                if not controller.startDaemonAsync():
                    self.printLog("error", "Starting process for service '%s' failed" % service.getName())
                    allOk = False
                elif not self.__awaitServiceStartup(service, controller, startTimeout):
                    self.printLog("error", "Starting service '%s' failed" % service.getName())
                    allOk = False
                else:
                    self.__servicesStartedSuccessfully.append(service)

                statusAction = service.getController(self.logger).createPrintStatusAction()
                statusAction.setHeaderEnabled(False)
                statusAction.execute()

            except Exception as err:
                self.printLog("error", "Caught exception while trying to start service '%s': %s" % (service.getName(), str(err)))
                self.logger.error(''.join(app.util.getStackTraceAndException()).strip())
                allOk = False

        if not allOk:
            self.printLog("error", "\nStartup failed. See %s for details." % (app.logutil.getFileName(self.logger) or "log"))

        return allOk

    def __initializeAllServices(self):
        for service in self.__services:
            if not service in self.__servicesStartedSuccessfully:
                self.printLog("warn", "Skipping initialization of '%s' as it did not start up successfully" % service.getName())

        import app.actions.initialize
        return app.actions.initialize.InitializeServiceAction(self.__servicesStartedSuccessfully, self.logger).execute()

    def lockedExecute(self):
        startupResult = self.__startAllServices()
        initResult = self.__initializeAllServices()

        return startupResult and initResult
