﻿import time

import app.util
import app.servicecontrol
import app.defaults
import app.logutil
import app.actions.lockedaction

class StopServiceAction(app.actions.lockedaction.LockedControllerAction):
    def __init__(self, services, logger):
        self.__services = services
        super(StopServiceAction, self).__init__(logger)

    def chain(self, action):
        assert isinstance(action, StopServiceAction)

        return StopServiceAction(
            self.__services + action.getInvolvedServices(),
            self.logger)

    def getInvolvedServices(self):
        return self.__services

    def confirm(self, isRedeployment=False):
        # Stopping a service is nothing worth asking for confirmation for.
        return True

    def __killDaemon(self, service):
        service.getController(self.logger).killDaemon()

    def __awaitServiceShutdown(self, service, timeout):
        timeStart = time.time()
        controller = service.getController(self.logger)

        sleepInterval = 2
        try:
            self.printLog("info", "  Waiting for %s to stop..." % service.getName(), False)

            while time.time() < timeStart + timeout:
                status = controller.getStatus()
                if status == app.servicecontrol.ServiceStatus.STOPPED_AUTOSTART or \
                   status == app.servicecontrol.ServiceStatus.STOPPED_DEMANDSTART:
                    # Great, we are set.
                    self.logger.info("Service '%s' stopped successfully in %d seconds" %
                                     (service.getName(), time.time() - timeStart))
                    return True

                elif status == app.servicecontrol.ServiceStatus.DEAD:
                    # Well, that was not smooth -- the service probably did not
                    # delete its PID file correctly.
                    self.logger.info("Service '%s' stopped abruptly in %d seconds" %
                                     (service.getName(), time.time() - timeStart))
                    return True

                elif status == app.servicecontrol.ServiceStatus.AVAILABLE or \
                     status == app.servicecontrol.ServiceStatus.UNHEALTHY or \
                     status == app.servicecontrol.ServiceStatus.STOPPING:
                    # Give it some more time.

                    self.printLog("info", ".", False)
                    sleepInterval = sleepInterval + 1
                    time.sleep(sleepInterval)

                elif status == app.servicecontrol.ServiceStatus.PORTS_BLOCKED:
                    # The service might have already removed its PID file, but it is
                    # still running. Give it some more time.

                    self.printLog("info", ".", False)
                    sleepInterval = sleepInterval + 1
                    time.sleep(sleepInterval)

                else:
                    # Huh?
                    self.logger.error("Service '%s' unexpectedly entered status %s during shutdown - killing it" %
                                      (service.getName(), status))
                    self.__killDaemon(service)
                    return False

        finally:
            self.printLog("info", " ")

        # Timeout elapsed.
        self.printLog("error", "Service '%s' failed to stop up within %d seconds - killing it" %
                      (service.getName(), timeout))
        self.__killDaemon(service)
        return False

    def lockedExecute(self):
        self.printLog("title", "Stopping Services");

        allOk = True

        for service in self.__services:
            try:
                stopTimeout = service.getDaemonStartupInfo().getStartupTimeout()
                if not stopTimeout:
                    stopTimeout = 60

                controller = service.getController(self.logger)

                if controller.getStatus() == app.servicecontrol.ServiceStatus.STOPPED_AUTOSTART or \
                   controller.getStatus() == app.servicecontrol.ServiceStatus.STOPPED_DEMANDSTART:
                    # Great, nothing to do here
                    pass
                elif controller.getStatus() == app.servicecontrol.ServiceStatus.DOOMED:
                    self.printLog("error", "Service '%s' is in doomed state " % service.getName() +
                                           "- a clean shutdown will probably be futile, terminating it")
                    self.__killDaemon(service)
                elif not controller.stopDaemonAsync():
                    self.printLog("error", "Stopping process for service '%s' failed" % service.getName())
                    allOk = False
                elif not self.__awaitServiceShutdown(service, stopTimeout):
                    self.printLog("error", "Stopping service '%s' failed" % service.getName())
                    allOk = False

                statusAction = service.getController(self.logger).createPrintStatusAction()
                statusAction.setHeaderEnabled(False)
                statusAction.execute()

            except Exception as err:
                self.printLog("error", "Caught exception while trying to stop service '%s': %s" % (service.getName(), str(err)))
                self.logger.error(''.join(app.util.getStackTraceAndException()).strip())
                allOk = False

        if not allOk:
            self.printLog("error", "\nStopping failed. See %s for details." % (app.logutil.getFileName(self.logger) or "log"))

        return allOk
