﻿#!/usr/bin/env python
# Legacy code
# pylint: skip-file

"Mini-library of common code shared by adapter-svn Python scripts."


import glob
import os
import os.path
import sys
import re
import stat
import urllib
import string
import shutil

import app.vc
import app.util
import app.preprocessor
import svn.core, svn.repos, svn.fs, svn.delta

### Callers of this module's run_cmd() should be using
### app.vc.run_cmd() instead.  Once they all do, we can remove
### this crutch.
run_cmd = app.vc.run_cmd

def usage(before, long_opts, after):
  """Print BEFORE, then format and print LONG_OPTS, then print AFTER.

LONG_OPTS is a list of lists, each sublist contains three elements: the
option name, the placeholder name of its argument if any (else None),
and a description string.  If the option name ends with '=', the
option takes an argument, so the placeholder arg name will be used
(defaulting to 'FOO'); else if the option does not take an argument,
the placeholder argument name is ignored.

If BEFORE or AFTER is None, it is not printed."""
  if before:
    print before
    print ""

  for x in long_opts:
    if x[0][-1:] == '=':
      print "   --" + x[0] + (x[1] or "FOO")
    else:
      print "   --" + x[0]
    print "     ", x[2]

  print ""
  if after:
    print after


def checkSafeUser(cfg):
  import pwd
  httpd_user = cfg.get('HTTPD_USER')
  if os.getuid() != pwd.getpwnam(httpd_user)[2]:
      print "ERROR:  You are not '%s', the Subversion-safe user." \
             % (httpd_user)
      sys.exit(1)


class CommitInfoError(Exception):
  pass


class SVNLink:
  """A link to a specific Subversion resource associated with a
  revision, clever enough to turn itself into a URL."""

  def __init__(self, scheme, port, project, domain, path, change, newRev):
    self.scheme = scheme
    self.port = int(port)
    self.project = project
    self.domain = domain
    self.change = change
    self.newRev = newRev
    self.path = path

  def __str__(self):
    """Creates a URL to ViewCVS, providing diffs for modified
    files."""

    # Easy out -- no diff URLs for directories
    if self.change.item_kind == svn.core.svn_node_dir:
      return ''

    # Calculate our base URL.
    portstr = ''
    if not (self.port == 80 or self.port == 443):
      portstr = ':%d' % (self.port)
    base_url = "%s://%s.%s%s/source/browse/%s" \
           % (self.scheme, self.project, self.domain,
            portstr, self.project)

    # And then calculate our full diff URL, if any.
    diff_url = ''
    if not self.change.path: ## delete
      diff_url = '%s/%s?view=auto&rev=%d' \
             % (base_url,
              urllib.quote(self.change.base_path[1:]),
              self.change.base_rev)
    elif self.change.added: ## add...
      if self.change.base_path \
           and (self.change.base_rev != -1): ## ...with history
        diff_url = '%s/%s?view=diff&rev=%d&p1=%s&r1=%d&p2=%s&r2=%d' \
               % (base_url,
                urllib.quote(self.change.path),
                self.newRev,
                urllib.quote(self.change.base_path[1:]),
                self.change.base_rev,
                urllib.quote(self.change.path),
                self.newRev)
      else: ## ...without history
        diff_url = '%s/%s?view=auto&rev=%d' \
               % (base_url,
                urllib.quote(self.change.path),
                self.newRev)
    elif not self.change.text_changed: ## no change
      return
    else: ## modification
      diff_url = '%s/%s?view=diff&rev=%d&p1=%s&r1=%d&p2=%s&r2=%d' \
             % (base_url,
              urllib.quote(self.change.path),
              self.newRev,
              urllib.quote(self.change.base_path[1:]),
              self.change.base_rev,
              urllib.quote(self.change.path),
              self.newRev)
    return diff_url

# CVS could use a similar API if we ever get rid of the Perl scripts.
# Right now, it's embedded in logchange.pl.  This implementation would
# be cleaner using Subversion's Python bindings.
def getCommitLinks(cfg, repository, project, revnum):
  """Returns three lists of SVNLink objects to the deltas associated
  with a revision.  The lists hold URLs to the added, modified, and
  removed versioned resources, respectively."""

  domain = cfg["DOMAIN"]
  scheme = cfg.get("HTTP_SCHEME", "http")
  port = int(cfg.get("APACHE_PORT", 80))

  def _get_commit_links(pool):
    repos_obj = svn.repos.svn_repos_open(repository, pool)
    fs_obj = svn.repos.svn_repos_fs(repos_obj)
    editor = svn.repos.RevisionChangeCollector(fs_obj, revnum, pool)
    e_ptr, e_baton = svn.delta.make_editor(editor, pool)
    root = svn.fs.revision_root(fs_obj, revnum, pool)
    svn.repos.svn_repos_replay(root, e_ptr, e_baton, pool)
    changelist = editor.changes.items()
    changelist.sort()
    adds = []
    dels = []
    mods = []
    for path, change in changelist:
      status = 'M'
      if not change.path: ## delete
        status = 'D'
      elif change.added:
        status = 'A'
      elif not change.text_changed: ## no change
        pass
      else: ## modification
        pass

      url = SVNLink(scheme, port, project, domain, path, change, revnum)
      if status == 'A':
        adds.append(url)
      elif status == 'D':
        dels.append(url)
      else:
        mods.append(url)
    return adds, mods, dels

  try:
    return svn.core.run_app(_get_commit_links)
  except Exception as e:
    err_msg = str(e) + '\n'
    sys.stderr.write(err_msg)
    raise CommitInfoError, err_msg


def getCommitInfo(repository, revision):
  """Return strings author, date, and log message for REPOSITORY at
  REVISION.  If any error, print the error to stderr and raise
  CommitInfoError. """

  def _get_commit_info(pool):
    repos_obj = svn.repos.svn_repos_open(repository, pool)
    fs_obj = svn.repos.svn_repos_fs(repos_obj)
    author = svn.fs.revision_prop(
      fs_obj, revision, svn.core.SVN_PROP_REVISION_AUTHOR, pool) or ''
    log_msg = svn.fs.revision_prop(
      fs_obj, revision, svn.core.SVN_PROP_REVISION_LOG, pool) or ''
    date = svn.fs.revision_prop(
      fs_obj, revision, svn.core.SVN_PROP_REVISION_DATE, pool)
    if date is not None:
      date = svn.core.svn_time_to_human_cstring(
        svn.core.svn_time_from_cstring(date, pool),
        pool)
    return author, date, log_msg

  try:
    return svn.core.run_app(_get_commit_info)
  except Exception as e:
    err_msg = str(e) + '\n'
    sys.stderr.write(err_msg)
    raise CommitInfoError, err_msg


DB_CONFIG_TEMPLATE = """
# This is the configuration file for the Berkeley DB environment
# used by your Subversion repository.
#
# You should run 'svnadmin recover' on the repository whenever you
# modify this file, to ensure that your changes take effect.

### Lock subsystem
#
# Make sure you read the documentation at:
#
#   http://www.sleepycat.com/docs/ref/lock/max.html
#
# before tweaking these values.
set_lk_max_locks   %d
set_lk_max_lockers %d
set_lk_max_objects %d

### Log file subsystem
#
# Make sure you read the documentation at:
#
#   http://www.sleepycat.com/docs/api_c/env_set_lg_bsize.html
#   http://www.sleepycat.com/docs/api_c/env_set_lg_max.html
#   http://www.sleepycat.com/docs/ref/log/limits.html
#
# Increase the size of the in-memory log buffer from the default
# of 32 Kbytes to 256 Kbytes.  Decrease the log file size from
# 10 Mbytes to 1 Mbyte.  This will help reduce the amount of disk
# space required for hot backups.  The size of the log file must be
# at least four times the size of the in-memory log buffer.
#
# Note: Decreasing the in-memory buffer size below 256 Kbytes
# will hurt commit performance. For details, see this post from
# Daniel Berlin <dan@dberlin.org>:
#
# http://subversion.tigris.org/servlets/ReadMsg?list=dev&msgId=161960
set_lg_bsize    %d
set_lg_max      %d
#
# If you see 'log region out of memory' errors, bump lg_regionmax.
# See http://www.sleepycat.com/docs/ref/log/config.html and
# http://svn.haxx.se/users/archive-2004-10/1001.shtml for more.
set_lg_regionmax %d
#
# The default cache size in BDB is only 256k. As explained in
# http://svn.haxx.se/dev/archive-2004-12/0369.shtml, this is too
# small for most applications. Bump this number if 'db_stat -m'
# shows too many cache misses.
set_cachesize    0 %d 1
#
# Disable fsync of log files on transaction commit. Read the
# documentation about DB_TXN_NOSYNC at:
#
#   http://www.sleepycat.com/docs/api_c/env_set_flags.html
#
# [requires Berkeley DB 4.0]
# set_flags DB_TXN_NOSYNC
#
# Enable automatic removal of unused transaction log files.
# ### NOTE: You probably don't want to turn on automatic logfile
# ### removal for CEE in production.  Our backup process uses the
# ### log files, and is also responsible for cleaning them up.  If
# ### files are accumulating on a production site, it may mean
# ### something has gone wrong with the backups, so look there first.
# ### See PCN 27367.
# ###
# ### Of course, if you're not in production, or for some other reason
# ### are not running backups, then you do want to autoremove logfiles,
# ### otherwise the repository will grow too quickly.
# Read the documentation about DB_LOG_AUTOREMOVE at:
#
#   http://www.sleepycat.com/docs/api_c/env_set_flags.html
#
# [requires Berkeley DB 4.2]
%s

# We store logfiles and datafiles in a separate directory from the
# rest of the Berkeley DB environment, to facilite CEE backup
# procedures.  Otherwise they remain in the main DB environment
# directory.  See PCN27580 and PCN28803.
#
set_lg_dir %s
set_data_dir %s
# We store certain BDB environment files in a separate directory, too,
# but that's so we can keep them on local disk instead of possibly on
# NFS where the rest of the repository often lives.  See PCN65841.
%s
"""

def writeDBConfigFile(db_config_path, lg_dir, data_dir,
                      db_log_autoremove=0, max_locks=8000, max_lockers=8000,
                      max_objects=8000, lg_bsize=262144, lg_max=1048576,
                      lg_regionmax=131072, cachesize=1048576, env_dir=None):
  if db_log_autoremove:
    log_flags = "# set_flags DB_LOG_AUTOREMOVE"
  else:
    log_flags = "set_flags DB_LOG_AUTOREMOVE"

  if env_dir:
    env_line = "set_env_dir " + env_dir
  else:
    env_line = "# set_env_dir /FIX/ME/PLEASE"

  open(db_config_path, 'w').write(DB_CONFIG_TEMPLATE % \
                                  (max_locks, max_lockers, max_objects,
                                   lg_bsize, lg_max, lg_regionmax, cachesize,
                                   log_flags, lg_dir, data_dir, env_line))


def createProjectRepository(cfg, projectname, renameRepos=1):
  """Create a new svn repository PROJECTNAME in $repository_home.
  Use CFG to obtain run-time configuration data."""

  svnadmin = cfg.get('SVNADMIN_BINARY')

  # The parent directory of all CollabNet Subversion repositories.
  repository_home = cfg.get('SVN_REPOSITORY_HOME')

  # Is this a production environment?
  in_production = cfg.get('IN_PRODUCTION') == 'true'

  # We build the repository under a temporary name, so it can't be
  # accessed until we're done setting it up.  The temporary name
  # includes characters not legal in project names, so conflict with
  # a real project is not possible.
  tmp_repository = os.path.join(repository_home, '_TMP_' + projectname)
  repository = os.path.join(repository_home, projectname)

  # Create a repository.  (### Do something with stdout?  stderr?)
  command = svnadmin + " create " + tmp_repository + " --fs-type bdb"
  stdout_lines, stderr_lines = app.vc.run_cmd(command)
  if len(stderr_lines):
    raise Exception(str(stderr_lines))

  # Now we have a repository at revision 0, containing just an empty
  # root directory.  But, we have to tweak it a bit... In CollabNet,
  # subversion repositories keep their database files outside the
  # repository so that the database files don't get backed up as
  # though they were regular static data files in DATA_DIR.  Logfiles
  # are also stored in a separate location, outside DATA_DIR.  See PCN
  # 27580 for details.
  datafile_dir = os.path.join(cfg.get('SVN_BDB_DATAFILE_LIVE'), projectname)
  logfile_dir = os.path.join(cfg.get('SVN_BDB_LOGFILE_LIVE'), projectname)
  old_db_dir = os.path.join(tmp_repository, 'db')

  # Create our DATAFILE and LOGFILE directories, and populate them
  # with the database files and logfiles, respectively.

  os.mkdir(logfile_dir)
  os.mkdir(datafile_dir)

  def _my_move(old, new):
    # Note: don't use shutil.move() because it's Python >= 2.3.
    shutil.copy2(old, new)
    os.remove(old)

  db_archive = os.path.join(cfg.get('BDB_BINARIES_DIR'), 'db_archive')
  stdout_lines, stderr_lines = app.vc.run_cmd(
      db_archive + ' -l -h ' + old_db_dir)
  if len(stderr_lines):
    raise Exception(str(stderr_lines))
  for logfile in stdout_lines:
    _my_move(os.path.join(old_db_dir, string.rstrip(logfile)), logfile_dir)
  for dbfile in ['changes',
                 'copies',
                 'nodes',
                 'representations',
                 'revisions',
                 'strings',
                 'transactions',
                 'uuids']:
    _my_move(os.path.join(old_db_dir, dbfile), datafile_dir)

  # Write a new DB_CONFIG with parameters that support CollabNet use.
  # When IN_PRODUCTION, we'll turn on automatic logfile removal.
  # Otherwise, we wanna keep those suckers around.
  writeDBConfigFile(os.path.join(old_db_dir, 'DB_CONFIG'),
                    logfile_dir, datafile_dir, not in_production)

  # ... and 'svnadmin recover' to make the changes stick.
  command = svnadmin + " recover " + tmp_repository
  stdout_lines, stderr_lines = app.vc.run_cmd(command)
  if len(stderr_lines):
    raise Exception(str(stderr_lines))

  # rename to projectname only if the caller asks for it
  if renameRepos:
    # Make the new repository visible to the world.
    os.rename(tmp_repository, repository)


def installHookScripts(cfg, projectname):
  """Install the CollabNet default hook scripts into the repository
  for PROJECTNAME.  Use CFG to figure out where those scripts and that
  repository exist."""
  repository = os.path.join(cfg.get('SVN_REPOSITORY_HOME'), projectname)

  # We don't use pre-commit yet, so don't bother to copy it.
  hook_tmpl_dir = cfg.get('SVN_HOOK_TEMPLATES')
  hook_dest_dir = os.path.join(repository, "hooks")

  dirents = os.listdir(hook_tmpl_dir)
  for dirent in dirents:
      hook_tmpl = os.path.join(hook_tmpl_dir, dirent)
      hook_dest = os.path.join(hook_dest_dir, dirent)
      if os.path.isfile(hook_tmpl):
          app.preprocessor.copyScriptWithTransform(cfg, hook_tmpl,
                                                          hook_dest)


def _enumerater(func, repos_home, repos_bdb_data_dir,
                repos_bdb_logs_dir, repos_bdb_env_dir):
  """Call FUNC for each Subversion repository in REPOS_HOME.  FUNC is
  a callback function whose argument list is (REPOS_NAME, REPOS_HOME,
  REPOS_BDB_DATA_DIR, REPOS_BDB_LOGS_DIR)."""

  # Get a list of repositories in REPOS_HOME.
  for repos in os.listdir(repos_home):
    if os.path.isdir(os.path.join(repos_home, repos)):
      # Call our callback function with the repository information.
      func(repos,
           os.path.join(repos_home, repos),
           os.path.join(repos_bdb_data_dir, repos),
           os.path.join(repos_bdb_logs_dir, repos),
           os.path.join(repos_bdb_env_dir, repos))


def enumerateLiveRepositories(cfg, func):
  """Crawl the CEE data locations looking for live Subversion
  repositories, and call FUNC for each one.  FUNC is a callback
  function whose argument list is (REPOS_NAME, REPOS_HOME,
  REPOS_BDB_DATA_DIR, REPOS_BDB_LOGS_DIR)."""

  _enumerater(func,
              cfg.get('SUBVERSION_REPOSITORY_BASE'),
              cfg.get('BDB_DATAFILE_DIR'),
              cfg.get('BDB_LOGFILE_DIR'),
              cfg.get('BDB_ENVFILE_DIR'),
              cfg.get('SUBVERSION_BRANDING_REPOSITORY_BASE'),
              cfg.get('CVS_REPOSITORY_BASE'))


def enumerateArchivedRepositories(cfg, func):
  ###This function is broken and unused in CTF. In the next major sync up with
  ###CEE source should bless this function.
  """Crawl the CEE data locations looking for archived Subversion
  repositories, and call FUNC for each one.  FUNC is a callback
  function whose argument list is (REPOS_NAME, REPOS_HOME,
  REPOS_BDB_DATA_DIR, REPOS_BDB_LOGS_DIR)."""

  _enumerater(func,
              cfg.get('SVN_REPOSITORY_ARCHIVE'),
              cfg.get('SVN_BDB_DATAFILE_ARCHIVE'),
              cfg.get('SVN_BDB_LOGFILE_ARCHIVE'))


def enumerateRepositories(cfg, func):
  """Crawl the CEE data locations looking for Subversion repositories,
  and call FUNC for each one.  FUNC is a callback function whose
  argument list is (REPOS_NAME, REPOS_HOME, REPOS_BDB_DATA_DIR,
  REPOS_BDB_LOGS_DIR)."""
  enumerateLiveRepositories(cfg, func)
  #enumerateArchivedRepositories(cfg, func)


def refreshSvnWC(cfg, dirs):
  SVNBinary = cfg.get('SVN_BINARY')
  if dirs:
      for dir in dirs:
          originaldir = os.getcwd()
          try:
              os.chdir(dir)
              for entry in glob.glob("*"):
                  if os.path.isdir(entry):
                      shutil.rmtree(entry)
                  else:
                      os.remove(entry)
              cmd = '%s %s' % (user_group_string(cfg), SVNBinary)
              cmd = cmd + ' --password="avoid-prompt" update'
              original_lc_all = None
              try:
                  original_lc_all = app.util.setEnvHttpEncoding(cfg)
                  stdout_lines, stderr_lines = app.vc.run_cmd(cmd)
                  if len(stderr_lines):
                    raise Exception(str(stderr_lines))
              finally:
                  app.util.restoreEnvEncoding(original_lc_all)
          finally:
              os.chdir(originaldir)

def user_group_string(cfg):
    ug_string = ""
    if app.util.invokedBySuperUser():
      ug_string = "setuidgid %s" % cfg.get('HTTPD_USER')
    return ug_string


def check_path(txn_root, path, recurse, pool):
    """Check PATH in TXN_ROOT for illegal states.  RECURSE if so
    asked, and use POOL for temporary allocations."""

    # Fetch the 'svn:externals' property value.  Explicitly test
    # against None becuase we don't want to allow empty-string values
    # either.
    if svn.fs.node_prop(txn_root, path, svn.core.SVN_PROP_EXTERNALS, \
                        pool) is not None:
        raise Exception, "Cannot add or change '%s' property " \
                         "on or under www/ directory:\n  ('%s')\n" \
                         % (svn.core.SVN_PROP_EXTERNALS, path)

    # Fetch the 'svn:special' property value (used to make symlinks
    # into symlinks).  Explicitly test against None becuase we don't
    # want to allow empty-string values either.
    if svn.fs.node_prop(txn_root, path, svn.core.SVN_PROP_SPECIAL, \
                        pool) is not None:
        raise Exception, "Cannot add or change '%s' property " \
                         "on or under www/ directory:\n  ('%s')\n" \
                         % (svn.core.SVN_PROP_SPECIAL, path)

    # If we aren't recursing (or the path isn't recursible), we're done.
    if (not recurse) \
        or (svn.fs.check_path(txn_root, path, pool) != svn.core.svn_node_dir):
        return

    # Get the children, and loop over them, checking each one.
    children = svn.fs.dir_entries(txn_root, path, pool).keys()
    subpool = svn.core.svn_pool_create(pool)
    for child in children:
        svn.core.svn_pool_clear(subpool)
        check_path(txn_root, path + '/' + child, 1, subpool)
    svn.core.svn_pool_destroy(subpool)


def compare_paths(path1, path2):
    """Compare PATH1 and PATH2 from an ordering perspective.  Suitable
    for use with sort()."""

    path1_len = len (path1)
    path2_len = len (path2)
    min_len = min(path1_len, path2_len)
    i = 0

    # Are the paths exactly the same?
    if path1 == path2:
        return 0

    # Skip past common prefix
    while (i < min_len) and (path1[i] == path2[i]):
        i = i + 1

    # Children of paths are greater than their parents, but less than
    # greater siblings of their parents
    char1 = '\0'
    char2 = '\0'
    if (i < path1_len):
        char1 = path1[i]
    if (i < path2_len):
        char2 = path2[i]

    if (char1 == '/') and (i == path2_len):
        return 1
    if (char2 == '/') and (i == path1_len):
        return -1
    if (i < path1_len) and (char1 == '/'):
        return -1
    if (i < path2_len) and (char2 == '/'):
        return 1

    # Common prefix was skipped above, next character is compared to
    # determine order
    return cmp(char1, char2)


def subsumes(path1, path2):
    """Return non-zero iff PATH1 is, or is a parent of, PATH2."""
    if path1 == '' \
       or path1 == '/' \
       or path1 == path2 \
       or path2.find(path1 + '/') == 0:
        return 1
    return 0


def verifyNoChangesOutsideOf(txn_t, repo_fs, fs_path, pool):
    # Return 0 if TXN_T in REPO_FS contains any modifications to paths
    # not at or under FS_PATH, non-zero otherwise.

    # Get all the fun roots and stuff we need, and our changed path list.
    txn_root = svn.fs.txn_root(txn_t, pool)
    changes = svn.fs.paths_changed(txn_root, pool)

    # Loop over paths, seeing if any aren't subsumed by FS_PATH.
    for path in changes.keys():
        if not subsumes(fs_path, path):
            return 0
    return 1


def assertValidWWWChanges(txn_t, repo_fs, pool):
    # Okay, for the sake of security and sanity, we need to ensure
    # that no svn:externals or svn:special (symlink) properties are
    # being added under the /www directory.  So what are we looking for:
    #   - existing files modified to have these properties.
    #   - additions/replacements of nodes with this property, checked
    #     recursively (because directory copies into www/ could have
    #     the gotcha properties inside of them).

    # Get all the fun roots and stuff we need, and our changed path list.
    txn_root = svn.fs.txn_root(txn_t, pool)
    changes = svn.fs.paths_changed(txn_root, pool)
    paths = changes.keys()

    # Sort the changed_paths depth-first-ily.
    paths.sort(lambda a, b: compare_paths(a, b))

    # Create an iteration subpool.
    subpool = svn.core.svn_pool_create(pool)

    # Loop over paths, checking as needed.  LAST_PATH will always
    # contain the path we most recently checked recursively for
    # naughty stuff.
    last_path = None
    for path in paths:
        # Clear our iteration subpool.
        svn.core.svn_pool_clear(subpool)

        # If PATH is subsumed by the last path we recursively checked,
        # skip it.
        if last_path and subsumes(last_path, path):
            continue

        # Fetch the CHANGE for path.
        change = changes[path]

        # We don't care about deleted paths, or modified paths that
        # don't have property mods.
        if (change.change_kind == svn.fs.path_change_delete) \
           or (change.change_kind == svn.fs.path_change_modify \
               and not change.prop_mod):
            continue

        # We don't care about paths outside the auto-publish area.
        if not path.startswith('/trunk/www'):
            continue

        # We are left with modified+propchange paths, and
        # added/replaced paths, in the auto-publish area.
        if change.change_kind == svn.fs.path_change_modify:
            check_path(txn_root, path, 0, subpool)
        else:
            check_path(txn_root, path, 1, subpool)
            last_path = path

    # Destroy the iteration subpool.
    svn.core.svn_pool_destroy(subpool)

def rewriteWCUrls(wcDir, rootUrl, targetUrl):
    """Crawl Subversion working copy WCDIR, replacing entry URLs
    stored in the administrative subdirectories with new URLs rooted
    at TARGETURL, and replacing any repository root URLs with
    ROOTURL."""

    if not (os.path.isdir(wcDir)
            and os.path.isdir(os.path.join(wcDir, '.svn'))
            and os.path.isfile(os.path.join(wcDir, '.svn', 'entries'))):
        raise Exception, "ERROR: Not a working copy directory: %s\n" % (wcDir)

    if rootUrl[-1] == '/':
        rootUrl = rootUrl[:-1]
    if targetUrl[-1] == '/':
        targetUrl = targetUrl[:-1]
    if targetUrl != rootUrl and targetUrl.find(rootUrl + '/') != 0:
        raise Exception, "ERROR: Target URL '%s' is not equal to or a child " \
                         "of root URL '%s'\n" % (targetUrl, rootUrl)
    wcDir = os.path.normpath(wcDir)
    topDirLen = len(wcDir)

    _url_re = re.compile('^( *url=").*(" *)$')
    _repos_re = re.compile('^( *repos=").*(" *)$')

    def _walk_cb(baton, root, names):
        # If we're in a .svn directory and have an entries file to
        # play with, we'll do what we came here to do.
        if os.path.basename(root) == '.svn' and 'entries' in names:
            thisRelDir = os.path.dirname(root)[topDirLen:]
            entriesPath = os.path.join(root, 'entries')
            lines = open(entriesPath, 'r').readlines()

            # Replace the bits in our entries file that need replacin'.
            for i in range(len(lines)):
                match = _url_re.match(lines[i])
                if match and match.groups():
                    lines[i] = "%s%s%s\n" \
                               % (match.group(1),
                                  targetUrl + urllib.quote(thisRelDir),
                                  match.group(2))
                else:
                    match = _repos_re.match(lines[i])
                    if match and match.groups():
                        lines[i] = "%s%s%s\n" \
                                   % (match.group(1), rootUrl, match.group(2))

            # Get the chmod()able permission bits.  We'll need to
            # tweak them to allow write access, write the new lines
            # out, and then restore the permissions.
            s = os.stat(entriesPath)
            mode = stat.S_IMODE(s.st_mode)
            os.chmod(entriesPath, mode | stat.S_IWUSR)
            open(entriesPath, 'w').write(''.join(lines))
            os.chmod(entriesPath, mode)

        # Else if this directory has a .svn directory, make it the
        # only dir we care about.
        elif '.svn' in names and os.path.isdir(os.path.join(root, '.svn')):
            names = ['.svn']

    os.path.walk(wcDir, _walk_cb, None)

def getRepositoryFSType(cfg, projectname, parentpath=None):
  """Return the file-system type of a existing Subversion repository.
  if PARENTPATH not given, PROJECTNAME will be looked in the CTF
  default SVN repository location."""

  # The projects repository home.
  if parentpath is None:
    proj_home = os.path.join(cfg.get('SUBVERSION_REPOSITORY_BASE'), projectname)
  else:
    proj_home = os.path.join(parentpath, projectname)

  if os.path.isdir(proj_home):
    fstype_file = os.path.join(proj_home, 'db', 'fs-type')
    if os.path.exists(fstype_file):
      return open(fstype_file).readlines()[0].strip()
    else:
      return "bdb"
  else:
    raise Exception, "ERROR:  '%s', Subversion repository missing." \
          % (projectname)
