﻿import app.util
import app.preprocessor
import time
import new
import sys
import os

class ScriptException(Exception):
    pass

class ScriptProcessor(object):
    def __init__(self, scriptFile, moduleName, className, methodName, logger):
        self.scriptFile = scriptFile
        self.scriptBaseName = os.path.basename(scriptFile)
        self.moduleName = moduleName
        self.className = className
        self.methodName = methodName
        self.logger = logger

    def __repr__ (self):
        return self.scriptFile

    def __evaluatePython(self, config):
        """
            Evaluate a Python script
        """
        scriptContent = open(self.scriptFile).read()
        sys.modules[self.moduleName] = mod = new.module(self.moduleName)
        self.logger.debug('start -- script %s' % self.scriptFile)
        mod.__file__ = self.scriptFile


        try:
            exec scriptContent in mod.__dict__
            exec """
from %s import %s
obj = %s(self.logger, config)
__file__ = '%s'
obj.__file__ = '%s'
mod.__dict__['result'] = obj.%s()
""" % (self.moduleName, self.className, self.className, self.scriptFile, self.scriptFile, self.methodName)
            self.logger.debug("result -- %s" % mod.__dict__['result'])
            self.logger.debug('end -- script %s' % self.scriptFile)
            return mod.__dict__['result']
        except Exception as e:
            raise ScriptException, ScriptException("Script %s failed: %s" % (self.scriptFile, e)), sys.exc_info()[2]

    def evaluate(self, config):
        """
            Evaluate Python, shell script, or a static file
        """
        startTimeStamp = time.time()

        if self.scriptFile[-3:] == '.py':
            # Python - evaluate as script.
            result = self.__evaluatePython(config)
        else:
            # Assume static file.
            result = open(self.scriptFile).read()

        elapsed = time.time() - startTimeStamp

        self.logger.debug(result)
        self.logger.info('Completed %s in %f seconds' % (self.scriptFile, elapsed))

        return result

class Batch(object):
    def __init__(self, moduleName, className, methodName, logger, config):
        self.moduleName = moduleName
        self.className = className
        self.methodName = methodName
        self.logger = logger
        self.config = config

        self.scripts = []

    def __isHookScriptFile(self, filePath):
        filePath = filePath.lower()
        
        # Every file that is not a Python turd is a viable hook script.
        return os.path.isfile(filePath) and \
            not filePath.endswith(".pyo") and \
            not filePath.endswith(".pyc")

    def addFolder(self, scriptFolder):
        filePaths = []

        for f in os.walk(scriptFolder):
            filePaths += [os.path.join(f[0], file) for file in f[2]]

        for scriptFile in filePaths:
            if self.__isHookScriptFile(scriptFile):
                self.scripts.append(ScriptProcessor(
                    scriptFile,
                    self.moduleName,
                    self.className,
                    self.methodName,
                    self.logger))
        return self

    def addFile(self, scriptFile):
        self.scripts.append(ScriptProcessor(
            scriptFile,
            self.moduleName,
            self.className,
            self.methodName,
            self.logger))
        return self

    def removeFile(self, scriptBaseName):
        self.scripts = [script for script in self.scripts if script.scriptBaseName != scriptBaseName]
        return self

    def evaluateScript(self, script):
        """ Hook for subclasses to customize the result """

        return script.evaluate(self.config)

    def evaluateInto(self, fileHandle):
        """ Run scripts and write output to file """

        # evaluate all scripts, sorted by file name (not path).
        for script in sorted(self.scripts, key=lambda s: s.scriptBaseName):
            try:
                fileHandle.write(self.evaluateScript(script))
                fileHandle.write("\n")
            except:
                self.logger.debug('Error in runScripts')
                raise Exception, ''.join(app.util.getStackTraceAndException())
        fileHandle.write("\n")


    def evaluateIntoFile(self, outputFile, accessMode = 'w'):
        """ Run scripts and write output to file """

        with open(outputFile, accessMode) as fh:
            self.evaluateInto(fh)

    def execute(self):
        """ Run scripts, ignoring output """

        # run all scripts, sorted by file name (not path).
        for script in sorted(self.scripts, key=lambda s: s.scriptBaseName):
            try:
                self.evaluateScript(script)
            except Exception as e:
                self.logger.error('Error in runScripts')
                self.logger.error(''.join(app.util.getStackTraceAndException()))
                raise e

    def size(self):
        """ Returns the number of scripts in the batch """
        return len(self.scripts)

class ConfigBatch(Batch):
    def __init__(self, logger, runtimeConfig = None):
        if runtimeConfig is None:
            runtimeConfig = app.util.getRuntimeConfiguration()

        self.processor = app.preprocessor.TokenProcessor(runtimeConfig)

        super(ConfigBatch, self).__init__('ConfigScript', 'ConfigScript', 'run', logger, runtimeConfig)

    def evaluateScript(self, script):
        # replace tokens like {__TOKEN__} in result
        return self.processor.substitute(super(ConfigBatch, self).evaluateScript(script))

class BootstrapBatch(Batch):
    def __init__(self, logger, config):
        super(BootstrapBatch, self).__init__(
            'BootstrapScript', 'BootstrapScript', 'run', logger, config)

class MigrateBatch(Batch):
    def __init__(self, logger, config):
        super(MigrateBatch, self).__init__(
            'MigrationScript', 'MigrationScript', 'run', logger, config)
class CronBatch(Batch):
    def __init__(self, logger, config):
        super(CronBatch, self).__init__(
            'CronScript', 'CronScript', 'run', logger, config)

class OverrideSemantics:
    OVERRIDE_EXISTING_OPTIONS = 1
    KEEP_EXISTING_OPTIONS = 2

class RuntimeOptionsBatch(Batch):
    def __init__(self, overrideSemantics, logger, initialOptions):
        self.__options = initialOptions
        self.__overrideSemantics = overrideSemantics

        super(RuntimeOptionsBatch, self).__init__(
            'OptionScript', 'OptionScript', 'run', logger, initialOptions)

    def evaluateScript(self, script):
        # evaluate script using the config of the previous iteration
        #
        # Most snippets are not well behaved and just call app.util.getRuntimeConfiguration
        # even though runtime-options.conf does not exist yet.
        #
        # Make sure this returns the same options.
        app.util.setRuntimeConfiguration(self.__options)
        rawSnippet = script.evaluate(self.__options)
        app.util.setRuntimeConfiguration(None)

        newOptions = app.options.RuntimeOptions.readFromSnippet(rawSnippet)

        if self.__overrideSemantics == app.batch.OverrideSemantics.KEEP_EXISTING_OPTIONS:
            self.__options = newOptions \
                .overlayBy(self.__options) \
                .resolveReferences()
        elif self.__overrideSemantics == app.batch.OverrideSemantics.OVERRIDE_EXISTING_OPTIONS:
            self.__options = self.__options \
                .overlayBy(newOptions) \
                .resolveReferences()
        else:
            raise Exception("Unrecognized override semantics specified")

    def createRuntimeOptions(self):
        self.execute()
        return self.__options

class CreateRuntimeBatch(Batch):
    def __init__(self, logger, config):
        super(CreateRuntimeBatch, self).__init__(
            'CreateRuntimeScript', 'CreateRuntimeScript', 'run', logger, config)

class VerificationBatch(Batch):
    def __init__(self, logger, config):
        super(VerificationBatch, self).__init__(
            'VerificationScript', 'VerificationScript', 'run', logger, config)

class InitializationBatch(Batch):
    def __init__(self, logger, config):
        super(InitializationBatch, self).__init__(
            'InitializationScript', 'InitializationScript', 'run', logger, config)

class FolderCreationBatch(Batch):
    def __init__(self, logger, config):
        super(FolderCreationBatch, self).__init__(
            'FolderScript', 'FolderScript', 'createFolders', logger, config)

class FolderDeletionBatch(Batch):
    def __init__(self, logger, config):
        super(FolderDeletionBatch, self).__init__(
            'FolderScript', 'FolderScript', 'deleteFolders', logger, config)

class FolderPermissionsBatch(Batch):
    def __init__(self, logger, config):
        super(FolderPermissionsBatch, self).__init__(
            'FolderScript', 'FolderScript', 'applyFolderPermissions', logger, config)

class CheckAndApplyFolderPermissionsBatch(Batch):
    def __init__(self, logger, config):
        super(CheckAndApplyFolderPermissionsBatch, self).__init__(
            'FolderScript', 'FolderScript', 'checkAndApplyFolderPermissions', logger, config)

class FolderSelinuxPolicyBatch(Batch):
    def __init__(self, logger, config):
        super(FolderSelinuxPolicyBatch, self).__init__(
            'FolderScript', 'FolderScript', 'applySELinuxPermissions', logger, config)

class CheckAndApplySELinuxPermissionsBatch(Batch):
    def __init__(self, logger, config):
        super(CheckAndApplySELinuxPermissionsBatch, self).__init__(
            'FolderScript', 'FolderScript', 'checkAndApplySELinuxPermissions', logger, config)
