﻿import os
import shutil
import ssl
import app.defaults
import app.util
import tempfile
import app.logutil

class JavaCertStore(object):
    ALIAS = "node1"
    PASSPHRASE = "changeit"

    def __init__(self, keystorePath, ):
        self.__keystorePath = keystorePath

    @staticmethod
    def generateSelfSignedCertificate(keystorePath, commonName, validityInDays):
         javaCommand = [
            os.path.join(app.defaults.DEFAULT_JAVA_HOME, 'bin/keytool'),
            "-genkeypair",
            "-keystore", keystorePath,
            "-keyalg", "RSA",
            "-alias", JavaCertStore.ALIAS,
            "-dname", commonName,
            "-storepass", JavaCertStore.PASSPHRASE,
            "-keypass", "changeit",
            "-validity", str(validityInDays)
         ]

         app.util.executeCommand(javaCommand, useShell=False)

         return JavaCertStore(keystorePath)

    def exportCertificateToDerFile(self, path):
        """ Export in DER format """

        javaCommand = [
            os.path.join(app.defaults.DEFAULT_JAVA_HOME, 'bin/keytool'),
            "-exportcert",
            "-keystore", self.__keystorePath,
            "-alias", JavaCertStore.ALIAS,
            "-storepass", JavaCertStore.PASSPHRASE,
            "-file", path
        ]

        app.util.executeCommand(javaCommand, useShell=False)

    def exportCertificateToPemFile(self, path):
        """ Export in PEM format """
        javaCommand = [
            os.path.join(app.defaults.DEFAULT_JAVA_HOME, 'bin/keytool'),
            "-exportcert",
            "-keystore", self.__keystorePath,
            "-alias", JavaCertStore.ALIAS,
            "-storepass", JavaCertStore.PASSPHRASE,
            "-file", path,
            "-rfc"
        ]

        app.util.executeCommand(javaCommand, useShell=False)

    def exportPrivateKeyToPemFile(self, path):
        """ Export in PEM format """
        javaCommand = [
            os.path.join(app.defaults.DEFAULT_JAVA_HOME, 'bin/java'),
            "-cp", app.defaults.CTFUTIL_CLASSPATH,
            "com.collabnet.ctf.base.security.tools.ExportPrivateKey",
            JavaCertStore.ALIAS,
            self.__keystorePath,
            JavaCertStore.PASSPHRASE
            ]

        key = app.util.executeCommand(javaCommand, useShell=False)

        with open(path, "w") as f:
            f.write(key)

class JavaTrustStore(object):
    DEFAULT_PASSWORD = "changeit"

    def __init__(self, path=None):
        self.__path = path

        defaultPaths = [
            os.path.join(app.defaults.DEFAULT_JAVA_HOME, "lib/security/jssecacerts"),
            os.path.join(app.defaults.DEFAULT_JAVA_HOME, "lib/security/cacerts")
        ]

        if not self.__path:
            for path in defaultPaths:
                if os.path.isfile(path):
                    self.__path = path

        if not self.__path:
            raise Exception("No path specified and no default store found on system")

    def getPath(self):
        return self.__path

    @staticmethod
    def createTrustStore(path, caRootPemFile):
        javaCommand = [
            os.path.join(app.defaults.DEFAULT_JAVA_HOME, 'bin/keytool'),
            "-importcert",
            "-keystore", path,
            "-alias", "CA-cert",
            "-storepass", JavaTrustStore.DEFAULT_PASSWORD,
            "-file", caRootPemFile,
            "-noprompt"
            ]

        app.util.executeCommand(javaCommand, useShell=False)

        return JavaTrustStore(path)

    def listTrustedOwners(self):
        javaCommand = [
            os.path.join(app.defaults.DEFAULT_JAVA_HOME, 'bin/keytool'),
            "-list",
            "-v",
            "-keystore", self.__path,
            "-storepass", JavaTrustStore.DEFAULT_PASSWORD,
            ]

        output = app.util.executeCommand(" ".join(javaCommand) + "|grep Owner|cut -d : -f 2").split("\n")
        return [line.strip() for line in output if line]

    def importCertificateFromServer(self, host, port):
        javaCommand = [
            os.path.join(app.defaults.DEFAULT_JAVA_HOME, 'bin/java'),
            "-cp", app.defaults.CTFUTIL_CLASSPATH,
            "com.collabnet.ctf.base.security.tools.ImportSslCertificateFromHost",
            host,
            str(port)
            ]

        if self.__path:
            javaCommand += [
                self.__path,
                JavaTrustStore.DEFAULT_PASSWORD
                ]

        app.util.executeCommand(javaCommand, useShell=False)

    def deleteCertificate(self, alias):
        javaCommand = [
            os.path.join(app.defaults.DEFAULT_JAVA_HOME, 'bin/keytool'),
            "-delete",
            "-alias", alias,
            "-keystore", self.__path,
            "-storepass", JavaTrustStore.DEFAULT_PASSWORD,
            "-noprompt"
            ]

        app.util.executeCommand(javaCommand, useShell=False)

    def importCertificateFile(self, alias, pemCertPath):
        # Importing fails if there if a certificate with the same alias is already
        # available. Therefore, delete first.
        try:
            self.deleteCertificate(alias)
        except:
            pass

        javaCommand = [
            os.path.join(app.defaults.DEFAULT_JAVA_HOME, 'bin/keytool'),
            "-import",
            "-alias", alias,
            "-keystore", self.__path,
            "-storepass", JavaTrustStore.DEFAULT_PASSWORD,
            "-file", pemCertPath,
            "-noprompt"
            ]

        app.util.executeCommand(javaCommand, useShell=False)

    def importCertificateChainFile(self, alias, pemChainFile):
        self.importCertificateFile(alias, pemChainFile)

class LinuxTrustStore(object):
    ANCHORS_DIR = "/etc/pki/ca-trust/source/anchors/"
    CAPATH_DIR = "/etc/pki/tls/certs/"

    @staticmethod
    def __readSubjectNameHash(pemFile):
        openSslCommand = [
            "openssl",
            "x509",
            "-noout",
            "-hash", 
            "-in",
            pemFile
            ]

        return app.util.executeCommand(openSslCommand, useShell=False).rstrip()

    def __importToCaPath(self, alias, pemPath):
        # Copy PEM file to CAPath and make sure it is world-readable so that
        # all processes/users can benefit from it.
        certPath = os.path.join(LinuxTrustStore.CAPATH_DIR, alias + ".pem")
        shutil.copy(pemPath, certPath)
        os.chmod(certPath, 0644)

        # For OpenSSL to use <CAPATH_DIR> as CAPath, certificates need to be
        # made accessible by having their subject name hash as file name, cf
        # https://www.openssl.org/docs/man1.0.2/ssl/SSL_CTX_load_verify_locations.html.
        subjectNameHash = LinuxTrustStore.__readSubjectNameHash(pemPath)
        for i in range(0, 9):
            linkPath = os.path.join(LinuxTrustStore.CAPATH_DIR, "%s.%d" % (subjectNameHash, i))
            if os.path.exists(linkPath) or os.path.islink(linkPath):
                if os.readlink(linkPath) == certPath:
                    # Link exists already, fine.
                    return
                else:
                    # Name clash, use different suffix
                    pass
            else:
                os.symlink(certPath, linkPath)
                return

        raise Exception("Multiple name clashes in %s" % LinuxTrustStore.ANCHORS_DIR)

    def __importToAnchors(self, alias, pemCertPath):
        shutil.copy(pemCertPath, os.path.join(LinuxTrustStore.ANCHORS_DIR, alias + ".pem"))
        app.util.executeCommand(["update-ca-trust", "extract"])

    def importCertificateFile(self, alias, pemCertPath):
        self.__importToCaPath(alias, pemCertPath)
        self.__importToAnchors(alias, pemCertPath)

    def importCertificateChainFile(self, alias, pemChainFile):
        self.importCertificateFile(alias, pemChainFile)

    def importCertificateFromServer(self, alias, host, port):
        try:
            tempCertFilePath = tempfile.NamedTemporaryFile().name

            with open(tempCertFilePath, "w") as tempCertFile:
                tempCertFile.write(ssl.get_server_certificate(
                    (host, port),
                    ssl_version=ssl.PROTOCOL_SSLv23))   # Python 2.6/CentOS 6 compatibility
            
            self.importCertificateFile(alias, tempCertFilePath)

        finally:
            if os.path.isfile(tempCertFilePath):
                os.remove(tempCertFilePath)

class CertificateGenerator(object):
    log = app.logutil.getRuntimeLogger()

    @staticmethod
    def deriveCommonDomain(allDomains):
        assert isinstance(allDomains, set)

        if len(allDomains) == 1:
            # Easy, only a single domain is used.
            return allDomains.pop()
        else:
            # Remove hostname in case there are dots in the name (host.example.com -> example.com).
            # In case there are no dots, use full name (example -> example).
            commonSubdomains = set([d[d.find(".")+1:] for d in allDomains])
            if len(commonSubdomains) == 1:
                # All domains have a common subdomain - so we use a wildcard to match them all.
                wildcardDomain = "*." + commonSubdomains.pop()

                CertificateGenerator.log.info("Using %s as common subdomain for %s", wildcardDomain, ", ".join(allDomains))
                return wildcardDomain

            else:
                raise Exception("The domains %s do not share a common subdomain, " % ", ".join(allDomains) +
                                "so it is not possible to use a single, non-SNI SSL certificate for all domains")

    @staticmethod
    def createSelfSignedCertificate(domainName, validityInDays, targetCertFile, targetKeyFile):
        try:
            tempKeystore = tempfile.NamedTemporaryFile().name

            app.certstore.JavaCertStore.generateSelfSignedCertificate(
                tempKeystore,
                "cn=%s"%domainName,
                validityInDays)

            javaCertStore = app.certstore.JavaCertStore(tempKeystore)
            javaCertStore.exportCertificateToPemFile(targetCertFile)
            javaCertStore.exportPrivateKeyToPemFile(targetKeyFile)

        finally:
            os.remove(tempKeystore)

    @staticmethod
    def provisionCertificates(primaryDomainName, effectiveDomainName, userCertPath=None, userKeyPath=None, userChainFile=None):
        # Note: <primaryDomainName> is the (canonical) domain name of this machine. But the machine
        # may have additional domain names, so <effectiveDomainName> is the domain that matches all
        # of them - usually a wildcard domain.

        targetCertPath = os.path.join(app.defaults.DEFAULT_SSLCERTS_PATH, primaryDomainName + ".crt")
        targetKeyPath  = os.path.join(app.defaults.DEFAULT_SSLCERTS_PATH, primaryDomainName + ".key")
        targetChainFile = os.path.join(app.defaults.DEFAULT_SSLCERTS_PATH, "intermediate.crt")

        CertificateGenerator.log.info("Using %s to store certificates", app.defaults.DEFAULT_SSLCERTS_PATH)
        if not os.path.isdir(app.defaults.DEFAULT_SSLCERTS_PATH):
            os.makedirs(app.defaults.DEFAULT_SSLCERTS_PATH)

        if userCertPath == targetCertPath and userKeyPath == targetKeyPath and \
            os.path.isfile(targetCertPath) and os.path.isfile(targetKeyPath):
            CertificateGenerator.log.info("Keeping existing SSL certiticate %s for '%s'" %
                                          (targetCertPath, effectiveDomainName))

        elif userCertPath and os.path.isfile(userCertPath) and userKeyPath and os.path.isfile(userKeyPath):
            CertificateGenerator.log.info("Using user-provided SSL certiticates for '%s'" % effectiveDomainName)
            userCertCommonName = Certificate(userCertPath).getCommonName()
            if effectiveDomainName.startswith("*") and userCertCommonName.lower() != effectiveDomainName.lower():
                raise Exception("The SSL certificate '%s' has been issued for " % userCertPath +
                                "the domain '%s', but the site requires a certificate " % userCertCommonName +
                                "for the wildcard domain '%s'" % effectiveDomainName)

            shutil.copy(userCertPath, targetCertPath)
            shutil.copy(userKeyPath, targetKeyPath)

            if userChainFile and os.path.isfile(userChainFile):
                shutil.copy(userChainFile, targetChainFile)
                CertificateGenerator.__setFilePermission(targetChainFile)
        else:
            CertificateGenerator.log.info("Generating self-signed SSL certiticates for '%s'" % effectiveDomainName)
            CertificateGenerator.createSelfSignedCertificate(
                effectiveDomainName,
                app.defaults.DEFAULT_CERTIFICATE_VALIDITY_PERIOD,
                targetCertPath,
                targetKeyPath)

        CertificateGenerator.__setFilePermission(targetCertPath)
        CertificateGenerator.__setFilePermission(targetKeyPath)

    @staticmethod
    def provisionCombinedCertificateFile(primaryDomainName):
        targetCertPath = os.path.join(app.defaults.DEFAULT_SSLCERTS_PATH, primaryDomainName + ".crt")
        targetKeyPath  = os.path.join(app.defaults.DEFAULT_SSLCERTS_PATH, primaryDomainName + ".key")
        targetChainFile = os.path.join(app.defaults.DEFAULT_SSLCERTS_PATH, "intermediate.crt")
        targetCombinedPemPath = os.path.join(app.defaults.DEFAULT_SSLCERTS_PATH, primaryDomainName + ".pem")

        # Create combined PEM file -- note: order matters, and indentation needs to be removed as well.
        with open(targetCombinedPemPath, 'w') as outfile:
            for fname in [targetKeyPath, targetCertPath, targetChainFile]:
                if os.path.isfile(fname):
                    with open(fname) as infile:
                        for line in infile.readlines():
                            outfile.write(line.strip() + "\n")

        CertificateGenerator.__setFilePermission(targetCombinedPemPath)

    @staticmethod
    def __setFilePermission(certFile):
        os.system("chown root:root %s" %certFile)
        os.chmod(certFile, 0600)

class Certificate(object):
    def __init__(self, pemFile):
        self.__pemFile = pemFile

    def getCommonName(self):
        javaCommand = [
            os.path.join(app.defaults.DEFAULT_JAVA_HOME, 'bin/java'),
            "-cp", app.defaults.CTFUTIL_CLASSPATH,
            "com.collabnet.ctf.base.security.tools.ExtractCommonName",
            self.__pemFile
            ]

        return app.util.executeCommand(javaCommand, useShell=False)
