﻿class Database(object):

    def __init__(self, connection):
        self.__connection = connection

    @staticmethod
    def connect(type, dbHost, dbPort, dbName, dbUser, dbPassword):
        if type == 'postgresql':
            return PostgresDatabase.connect(dbHost, dbPort, dbName, dbUser, dbPassword)
        elif type == 'oracle':
            return OracleDatabase.connect(dbHost, dbPort, dbName, dbUser, dbPassword)
        else:
            raise Exception("Unrecognized database type: %s" % type)

    def close(self):
        if self.__connection:
            try:
                self.__connection.close()
            except:
                pass

    def __enter__(self):
        """ enter 'with' block """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ exit 'with' block """
        self.close()

    def preprareSql(self, sql, bindVals=None):
        pass

    def execute(self, sql, bindVals=None):
        """ Execute UPDATE/DELETE/CREATE, returning number of affected rows """
        cur = self.__connection.cursor()
        cur.execute(self.preprareSql(sql, bindVals))
        self.__connection.commit()
        return cur.rowcount

    def query(self, sql, bindVals=None):
        """ Execute SELECT, returning result set """

        cur = self.__connection.cursor()
        cur.execute(self.preprareSql(sql, bindVals))
        return cur.fetchall()

    def boolValue(self, value):
        pass

    def queryNextSequenceValue(self, sequenceName):
        pass

class OracleDatabase(Database):

    @staticmethod
    def connect(dbHost, dbPort, dbName, dbUser, dbPassword):
        import cx_Oracle
        dsn = cx_Oracle.makedsn(dbHost, dbPort, dbName)
        con = cx_Oracle.connect(u"%s/%s@%s" % (dbUser, dbPassword, dsn))

        return OracleDatabase(con)

    def preprareSql(self, sql, bindVals=None):
        """
        Binds values into SQL
        """
        if bindVals is not None:
           sql = sql%(bindVals)

        return u"%s" % sql

    def boolValue(self, value):
        return '1' if value else '0'

    def queryNextSequenceValue(self, sequenceName):
        return self.query("select %s.nextval from dual" % sequenceName)[0][0]

class PostgresDatabase(Database):

    @staticmethod
    def connect(dbHost, dbPort, dbName, dbUser, dbPassword):
        import psycopg2
        con = psycopg2.connect(user = dbUser,
                            password = dbPassword,
                            database = dbName,
                            host = dbHost,
                            port = dbPort)
        con.autocommit=True

        return PostgresDatabase(con)

    @staticmethod
    def connectAsAdmin(dbHost, dbPort, dbName):
        return PostgresDatabase.connect(dbHost, dbPort, dbName, "postgres", None)

    def preprareSql(self, sql, bindVals=None):
        """
        Binds values into SQL
        """
        if bindVals is not None:
           sql = sql%(bindVals)

        return sql

    def boolValue(self, value):
        return 't' if value else 'f'

    def queryNextSequenceValue(self, sequenceName):
        return self.query("select nextval('%s')" % sequenceName)[0][0]
