﻿import os
import app.util
import app.defaults
import glob

class ConfigurationIntegrity:

    def __init__(self, services, checksumConfigFile):
        self.services  = services
        self.confObj = app.util.configParser()
        self.checksumConfigFile = checksumConfigFile

    def checkRuntimeIntegrity(self):
        integrityFailedList = []

        if os.path.isfile(self.checksumConfigFile):
            self.confObj.read(self.checksumConfigFile)

        for service in self.services:
            runtimeConfFiles = service.getIntegrityConfFiles()
            if runtimeConfFiles is not None:
                serviceName = service.getName()
                # old checksum config file has no section legacy-runtime
                # change the section to all while parsing the old checksum config file
                if os.path.basename(self.checksumConfigFile) == os.path.basename(app.defaults.OLD_CHECKSUM_CONFIG_FILE) and serviceName == "legacy-runtime":
                    serviceName = "all"

                for fileName in runtimeConfFiles:
                   try:
                       curCheckSum = self.confObj.get(serviceName, fileName)
                       newCheckSum = app.util.getCheckSum(runtimeConfFiles[fileName])
                   except:
                       continue

                   if newCheckSum is None:
                       continue
                   if curCheckSum != newCheckSum:
                       integrityFailedList.append(runtimeConfFiles[fileName])

        return integrityFailedList


    def writeIntegrityChecksums(self):
        configFolder = os.path.dirname(self.checksumConfigFile)
        if not os.path.isdir(configFolder):
            os.makedirs(configFolder)

        for service in self.services:
            runtimeConfigFiles = service.getIntegrityConfFiles()
            if runtimeConfigFiles is not None:
                serviceName = service.getName()
                self.confObj.add_section(serviceName)
                for fileName in runtimeConfigFiles:
                    checkSum = app.util.getCheckSum(runtimeConfigFiles[fileName])
                    self.confObj.set(serviceName, fileName, checkSum)

        with open(self.checksumConfigFile, "wb") as cfgFile:
            self.confObj.write(cfgFile)

class DataIntegrity:

    def __init__(self, services):
        self.services  = services
        self.dataChecksums = app.util.configParser()
        self.dataChecksumsFile = os.path.join(app.defaults.DEFAULT_RUNTIME_DIR,
                                              "conf",
                                              "dataChecksums.ini")
        self.failedPatterns = []
        self.successfulMatches = {}

    def checkDataIntegrity(self):
        integrityFailedList = []

        self.verifyAndUpdateDataChecksumsFile(forceUpdate = False)

        self.dataChecksums.read(self.dataChecksumsFile)

        for service in self.services:
            dataChecksumConfig = service.getDataChecksumConfig()
            if dataChecksumConfig is not None:
                serviceName = service.getName()

                for pattern in dataChecksumConfig:
                    if dataChecksumConfig[pattern] in self.failedPatterns:
                        continue
                    expectedChecksum = self.dataChecksums.get(serviceName, pattern)
                    for fileName in glob.glob(dataChecksumConfig[pattern]):
                        try:
                            actualChecksum = app.util.getCheckSum(fileName)
                        except:
                            raise

                        if actualChecksum != expectedChecksum:
                            integrityFailedList.append(fileName)
                        else:
                            if pattern in self.successfulMatches:
                                self.successfulMatches[pattern] += 1
                            else:
                                self.successfulMatches[pattern] = 1


        return integrityFailedList, self.failedPatterns, self.successfulMatches

    def verifyAndUpdateDataChecksumsFile(self, forceUpdate = False):
        if os.path.isfile(self.dataChecksumsFile):
            self.dataChecksums.read(self.dataChecksumsFile)

        for service in self.services:
            serviceName = service.getName()

            dataChecksumConfig = service.getDataChecksumConfig()
            if dataChecksumConfig is not None:
                if not self.dataChecksums.has_section(serviceName):
                    self.dataChecksums.add_section(serviceName)

                for pattern in dataChecksumConfig:
                    if self.dataChecksums.has_option(serviceName, pattern) and not forceUpdate:
                        continue
                    else:
                        sampleFile = next(glob.iglob(dataChecksumConfig[pattern]), None)
                        if sampleFile:
                            checksum = app.util.getCheckSum(sampleFile)
                            self.dataChecksums.set(serviceName, pattern, checksum)
                        else:
                            self.failedPatterns.append(dataChecksumConfig[pattern])
                            # No repos at the time of writing checksums. Intimate the user
                            # to re-create checksums after creating atleast one repo. #FIXME
            else:
                self.dataChecksums.remove_section(serviceName)

        with open(self.dataChecksumsFile, 'wb') as f:
            self.dataChecksums.write(f)

