﻿import os
import socket
import shutil
import app.permsutil
import app.options
import app.servicecontrol
import app.servicemanifest
import app.mixins
import app.selinuxutil
import app.util
# Object Model
#
#                                        +-------------+ *   1 +--------------+
#                                        |Prov. Endpnt |------>|Req. Endpnt   |
#                                        +-------------+       +--------------+
#                                               * ^-----+      * +-^
#                                                       |        |
#                                                       | 1    1 |
#  +------------+ 1     * +-------------+ 1        * +--------------+
#  |  Site      +-------->+  Machine    +----------->+   Service    |
#  +------------+         +-------------+            +--------------+
#

class Address(object):
    def __init__(self, fqdn, isLocal=False, ip=None):
        assert fqdn

        if fqdn == "localhost":
            raise ConfigurationException("'localhost' is not a valid FQDN")

        self.__fqdn = fqdn
        self.__isLocal = isLocal
        self.__ip = ip

    def __eq__(self, other):
        return isinstance(other, Address) and self.__fqdn == other.__fqdn

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__fqdn)

    def __repr__(self):
        if self.__ip:
            return "%s [%s]" % (self.__fqdn, self.__ip)
        else:
            return self.__fqdn

    def isLocal(self):
        return self.__isLocal

    def getIp(self):
        if not self.__ip:
            # Resolve IP lazily.
            self.__ip = Address.resolveDnsName(self.__fqdn)

        return self.__ip

    def getName(self):
        return self.__fqdn

    @staticmethod
    def resolveDnsName(dnsName):
        try:
            return socket.gethostbyname(dnsName)
        except Exception:
            raise NetworkException("Could not resolve IP for name '%s'" % dnsName)

class ProvidedEndpoint(app.mixins.DictEqualityMixin):
    def __init__(self, id, scope, port, viaInfo = None):
        assert viaInfo is None or isinstance(viaInfo, app.servicemanifest.Via)

        self.__id = id
        self.__scope = scope
        self.__port = port
        self.__viaInfo = viaInfo
        self.__clientServices = []

        # Injected later
        self.__service = None

    def _setService(self, service):
        assert isinstance(service, Service)
        self.__service = service

    def isViaOtherPort(self):
        """ Checks if this endpoint is to be accessed via a different port """
        return self.__viaInfo is not None

    def __getViaEndpoint(self):
        service = self.getMachine().getService(self.__viaInfo.getService())
        return service.getProvidedEndpoint(self.__viaInfo.getId())

    def __repr__(self):
        if self.isViaOtherPort():
            viaInfo = " (via %s)" % self.__viaInfo.getService()
        else:
            viaInfo = ""

        return "[%s] (%s) on %s:%d%s, %s" % (
            self.getId(),
            self.getScope(),
            self.getHost(),
            self.getPort(),
            viaInfo,
            self.getHost(False))

    def getId(self):
        return self.__id

    def getScope(self):
        return self.__scope

    def getPort(self):
        if self.isViaOtherPort():
            return self.__getViaEndpoint().getPort()
        else:
            return self.__port

    def getPhysicalHost(self):
        return self.__service.getMachine().getPhysicalName()

    def getService(self):
        return self.__service

    def getMachine(self):
        return self.__service.getMachine()

    def getBindingAddress(self):
        """ Return the 'right' address to bind the server to """

        if not self.isViaOtherPort() and \
           self.__scope == app.servicemanifest.PROVIDED_ENDPOINT_SCOPE_PUBLIC:
            # Bind to public network.
            return self.__service.getPublicAddress()
        else:
            # Bind to internal network.
            return self.__service.getInternalAddress()

    def getAddress(self):
        """ Return the 'right' address to use by clients """

        if self.isViaOtherPort():
            # As this endpoint is used via a different service, make sure
            # the via service's scope is taken to determine whether the
            # internal or public FQDN is to be used.
            scope = self.__getViaEndpoint().getScope()
        else:
            scope = self.__scope

        if scope == app.servicemanifest.PROVIDED_ENDPOINT_SCOPE_PUBLIC:
            # Remote box, go thru public endpoint/load balancer.
            return self.__service.getPublicAddress()
        else:
            # Internal cross-server communication.
            return self.__service.getInternalAddress()

    def getHost(self, preferLocalhost=True):
        """ Return the 'right' host to use by clients """

        address = self.getAddress()

        if preferLocalhost and address.isLocal():
            # Always prefer 'localhost'
            return "localhost"
        else:
            return address.getName()

    def getViaServiceName(self):
        if not self.__viaInfo:
            raise ConfigurationException("endpoint does not use via")

        return self.__viaInfo.getService()

    def _snapClient(self, clientService):
        assert isinstance(clientService, Service)
        self.__clientServices.append(clientService)

    def getClients(self):
        return self.__clientServices

    def getViaTargetEndpoints(self):
        """ Determine list of provided endpoints this service serves as a Via for """
        targets = []
        for siblingService in self.__service.getMachine().getServices():
            for provEndpoint in siblingService.getProvidedEndpoints():
                if provEndpoint.__viaInfo and \
                   provEndpoint.__viaInfo.getService() == self.__service.getName() and \
                   provEndpoint.__viaInfo.getId() == self.__id:
                    # This endpoint uses as as a Via.
                    targets.append(provEndpoint)
        return targets


class RequiredEndpoint(app.mixins.DictEqualityMixin):
    def __init__(self, id, serviceName, requiredForDeployment, requiredForStartup):
        assert isinstance(serviceName, basestring)

        self.__id = id
        self.__serviceName = serviceName
        self.__requiredForDeployment = requiredForDeployment
        self.__requiredForStartup = requiredForStartup
        self.__snappedEndpoint = None
        self.__providerService = None

    def __repr__(self):
        if self.__snappedEndpoint:
            snapInfo = " -> %s" % self.__snappedEndpoint
        else:
            snapInfo = ""

        return "[%s] of %s%s" % (self.getId(), self.getServiceName(), snapInfo)

    def getId(self):
        return self.__id

    def getServiceName(self):
        return self.__serviceName

    def isRequiredForStartup(self):
        return self.__requiredForStartup

    def isRequiredForDeployment(self):
        return self.__requiredForDeployment

    def _snapProvider(self, providerService, providedEndpoint):
        self.__providerService = providerService
        self.__snappedEndpoint = providedEndpoint

    def getProviderService(self):
        return self.__providerService

    def getProviderEndpoint(self):
        return self.__snappedEndpoint

class DaemonStartupInfo(object):
    def __init__(self, user=None, group=None, startupTimeout=None):
        self.__user = user
        self.__group = group
        self.__startupTimeout = startupTimeout

    def getUser(self):
        return self.__user

    def getGroup(self):
        return self.__group

    def getStartupTimeout(self):
        return self.__startupTimeout

class FolderInfo(object):
    def __init__(self, path, state, mode, owner, group, context, recursive):
        self.__path = path
        self.__state = state
        self.__mode = mode
        self.__owner = owner
        self.__group = group
        self.__context = context
        self.__recursive = recursive

    def getPath(self):
        return self.__path

    def getState(self):
        return self.__state

    def getMode(self):
        return self.__mode

    def getOwner(self):
        return self.__owner

    def getGroup(self):
        return self.__group

    def getContext(self):
        return self.__context

    def isRecursive(self):
        return self.__recursive

class ServiceFolder(object):
    def __init__(self, folderInfo):
        assert type(folderInfo) is FolderInfo

        self.__folderInfo = folderInfo

    def __repr__(self):
        if self.__folderInfo.getState() == app.servicemanifest.Folder.STATE_ABSENT:
            return "%s (absent)" % self.__folderInfo.getPath()
        else:
            return "%s [%s:%s, %o, %s]" % (
                self.__folderInfo.getPath(),
                self.__folderInfo.getOwner(),
                self.__folderInfo.getGroup(),
                app.permsutil.Permissions.getMode(self.__folderInfo.getMode()),
                self.__folderInfo.getContext())

    def __getOwnerUid(self):
        return app.permsutil.Permissions.getOwnerUid(self.__folderInfo.getOwner())

    def __getGroupGid(self):
        return app.permsutil.Permissions.getGroupGid(self.__folderInfo.getGroup())

    def __getContext(self):
        return self.__folderInfo.getContext()

    def __applyContext(self, path, context):
        app.selinuxutil.File.setContext(path, context)

    def __ensureAbsent(self):
        path = self.__folderInfo.getPath()

        if os.path.isfile(path):
            os.remove(path)
        elif os.path.islink(path):
            os.unlink(path)
        elif os.path.isdir(path):
            shutil.rmtree(path)

    def __ensurePresent(self, applyPermissions, applySelinuxPermissions, checkAndApply=False, checkAndApplySELinux=False):
        path = self.__folderInfo.getPath()

        if os.path.isfile(path):
            raise ServiceFolderException("File '%s' is in the way" % path)


        if not os.path.isdir(path):
            os.makedirs(path, app.permsutil.Permissions.getMode(self.__folderInfo.getMode()))
            # In migration if folder does not exists will
            #  create new folder at the same time we should set the
            # proper permission too.
            applyPermissions = True
            checkAndApply = False
            checkAndApplySELinux = True

        if checkAndApply:
            applyPermissions = not app.permsutil.Permissions.isPermissionApplied(path,
                                                                self.__folderInfo.getMode(),
                                                                self.__folderInfo.getOwner(),
                                                                self.__folderInfo.getGroup())
        if checkAndApplySELinux:
            applySelinuxPermissions = not app.selinuxutil.Policy.isSELinuxPermissionsApplied(
                                                            path,
                                                            self.__folderInfo.getContext())
        if applyPermissions:
            app.permsutil.Permissions.applyPermissions(
                path,
                self.__folderInfo.getMode(),
                self.__folderInfo.getOwner(),
                self.__folderInfo.getGroup(),
                self.__folderInfo.isRecursive())

        if applySelinuxPermissions:
            app.selinuxutil.Policy.applySELinuxPermissions(path,
                                                           self.__folderInfo.getContext())
    def apply(self,
              forceFolderRecreation,
              forceReapplyPermissions,
              forceReapplySelinuxPermissions,
              checkAndApply=False,
              checkAndApplySELinux=False):
        if self.__folderInfo.getState() == app.servicemanifest.Folder.STATE_ABSENT:
            self.__ensureAbsent()
        elif forceFolderRecreation:
            self.__ensureAbsent()
            self.__ensurePresent(True, True, False, False)
        else:
            self.__ensurePresent(forceReapplyPermissions, forceReapplySelinuxPermissions, checkAndApply, checkAndApplySELinux)

class Service(object):
    def __init__(
            self,
            name,
            daemonStartupInfo,
            providedEndpoints,
            requiredEndpoints,
            integrityConfFiles,
            dataChecksumConfig = None,
            folders = [],
            isAlias=False,
            internalAddress=None,
            publicAddress=None):
        assert not daemonStartupInfo or isinstance(daemonStartupInfo, DaemonStartupInfo)
        assert not providedEndpoints or isinstance(providedEndpoints[0], ProvidedEndpoint)
        assert not requiredEndpoints or isinstance(requiredEndpoints[0], RequiredEndpoint)
        assert not folders or isinstance(folders[0], ServiceFolder)
        assert not internalAddress or isinstance(internalAddress, Address)
        assert not publicAddress or isinstance(publicAddress, Address)

        self.__name = name
        self.__daemonStartupInfo = daemonStartupInfo
        self.__providedEndpoints = providedEndpoints
        self.__requiredEndpoints = requiredEndpoints
        self.__dependenciesInDeploymentOrder = []
        self.__transitiveDependenciesInDeploymentOrder = []

        self.__containerService = None
        self.__containedServices = []
        self.__integrityConfFiles = integrityConfFiles
        self.__dataChecksumConfig = dataChecksumConfig
        self.__folders = folders
        self.__isAlias = isAlias
        self.__internalAddress = internalAddress
        self.__publicAddress = publicAddress

        for endpoint in providedEndpoints:
            endpoint._setService(self)

        # Injected later
        self.__machine = None

        # Lazy loaded
        self.__controller = None

    def _setMachine(self, machine):
        assert isinstance(machine, Machine)
        self.__machine = machine

        # Provide default addresses
        if not self.__internalAddress:
            self.__internalAddress = machine.getInternalAddress()

        if not self.__publicAddress:
            self.__publicAddress = machine.getPublicAddress()

    def __repr__(self):
        return self.__name

    def getName(self):
        return self.__name

    def getMachine(self):
        return self.__machine

    def getInternalAddress(self):
        return self.__internalAddress

    def getPublicAddress(self):
        return self.__publicAddress

    def isAlias(self):
        return self.__isAlias

    def getProvidedEndpoints(self):
        return self.__providedEndpoints

    def getRequiredEndpoints(self):
        return self.__requiredEndpoints

    def getIntegrityConfFiles(self):
        return self.__integrityConfFiles

    def getDataChecksumConfig(self):
        return self.__dataChecksumConfig

    def getProvidedEndpoint(self, id, scope=None):
        for endpoint in self.__providedEndpoints:
            if endpoint.getId() == id and (scope is None or scope == endpoint.getScope()):
                return endpoint

        return None

    def getRequiredEndpoint(self, id, service=None):
        for endpoint in self.__requiredEndpoints:
            if endpoint.getId() == id and (service is None or service == endpoint.getServiceName()):
                return endpoint

        return None

    def setDeploymentDependencies(self, dependenciesInDeploymentOrder, transitiveDependenciesInDeploymentOrder, containerService):
        self.__dependenciesInDeploymentOrder = dependenciesInDeploymentOrder
        self.__transitiveDependenciesInDeploymentOrder = transitiveDependenciesInDeploymentOrder

        if containerService:
            if self.isDaemon():
                raise ConfigurationException(
                    "Service %s is a daemon and thus must not define a container" % self.__name)

            self.__containerService = containerService
            containerService.addContainedService(self)

    def addContainedService(self, service):
        if not self in service.getDeploymentDependencies():
            raise ConfigurationException(
                "Service %s is hosted in container %s but it does not define a deployment dependency" %
                (service.getName(), self.__name))
        self.__containedServices.append(service)

    def getContainingService(self):
        return self.__containerService

    def getContainedServices(self):
        return self.__containedServices

    def getDeploymentDependencies(self, transitive=False):
        if transitive:
            return self.__transitiveDependenciesInDeploymentOrder
        else:
            return self.__dependenciesInDeploymentOrder

    def getStartupEndpointDependencies(self, filteredByMachine = None):
        """ Returns the list of endpoints that need to be available for this service to start up """

        endpoints = set()

        # Any vital endpoint counts as startup dependency
        for reqEndpoint in self.__requiredEndpoints:
            if reqEndpoint.isRequiredForStartup():
                if not reqEndpoint.getProviderEndpoint():
                    raise UnresolvedEndpointDependenciesException(
                        "Service %s requires %s of %s, but this endpoint has not been resolved yet" %
                            (self.__name, reqEndpoint.getId(), reqEndpoint.getServiceName()))
                else:
                    endpoints.add(reqEndpoint)

        return endpoints

    def getStartupDependencies(self, filteredByMachine = None):
        """ Returns the list of services that need to be available for this service to start up """

        dependencies = set()

        for reqEndpoint in self.getStartupEndpointDependencies():
            if not filteredByMachine or filteredByMachine == reqEndpoint.getProviderEndpoint().getMachine():
                service = reqEndpoint.getProviderService()
                dependencies.add(service)

                # If the service is contained by another service, consider this as a dependency too.
                serviceContainer = service.getContainingService()
                if serviceContainer:
                    dependencies.add(serviceContainer)

        # Any service via which we expose endpoints is a startup dependency.
        for provEndpoint in self.__providedEndpoints:
            if provEndpoint.isViaOtherPort():
                if self.__containerService and \
                   self.__containerService.getName() == provEndpoint.getViaServiceName():
                    # The container will start up first anyway - if we added a
                    # startup dependency here, it would create a dependency cycle.
                    continue

                viaService = None
                for service in self.__dependenciesInDeploymentOrder:
                    if service.getName() == provEndpoint.getViaServiceName():
                        viaService = service
                        break

                if viaService:
                     dependencies.add(service)
                else:
                    raise UnresolvedEndpointDependenciesException(
                        "Service %s exposes endpoint via %s, but this service is not a deployment dependency" %
                        (self.__name, provEndpoint.getViaServiceName()))

        # If this service serves as container for other services, we need to consider the
        # contained services dependencies as well.
        for containedService in self.__containedServices:
            dependencies.add(containedService)

        return dependencies

    def getMachineStartupDependencies(self):
        """ Returns the list of machines whose services need to be available for this service to start up """

        machines = set()

        # Any vital endpoint counts as startup dependency
        for reqEndpoint in self.__requiredEndpoints:
            if reqEndpoint.isRequiredForStartup():
                if not reqEndpoint.getProviderEndpoint():
                    raise UnresolvedEndpointDependenciesException(
                        "Service %s requires %s of %s, but this endpoint has not been resolved yet" %
                            (self.__name, reqEndpoint.getId(), reqEndpoint.getServiceName()))

                if reqEndpoint.isRequiredForStartup():
                    machines.add(reqEndpoint.getProviderEndpoint().getMachine())

        return machines

    def isDaemon(self):
        return self.__daemonStartupInfo is not None

    def getDaemonStartupInfo(self):
        return self.__daemonStartupInfo

    def getController(self, logger):
        if not self.__machine.isLocal():
            raise ServiceException("Remote service cannot be controlled")

        if not self.__controller:
            self.__controller = app.servicecontrol.ServiceController.forService(self, logger)

        return self.__controller

    def getFolders(self):
        return self.__folders

class Machine(object):
    def __init__(self, physicalAddress, internalAddress, publicAddress, services, deploymentOrder = [], cluster=None):
        assert not services or isinstance(services[0], Service)
        assert isinstance(physicalAddress, Address)
        assert isinstance(internalAddress, Address)
        assert isinstance(publicAddress, Address)
        assert not deploymentOrder or isinstance(deploymentOrder[0], basestring)
        assert not cluster or isinstance(cluster, Cluster)

        self.__physicalAddress = physicalAddress
        self.__internalAddress = internalAddress
        self.__publicAddress = publicAddress
        self.__services = services
        self.__deploymentOrder = deploymentOrder
        self.__cluster = cluster

        if not deploymentOrder:
            deploymentOrder = [s.getName() for s in services]

        for service in services:
            service._setMachine(self)

    def __repr__(self):
        return self.getPhysicalName()

    def __eq__(self, other):
        return isinstance(other, Machine) and other.__physicalAddress == self.__physicalAddress

    @staticmethod
    def getLocalHostName():
        return socket.gethostname()


    def getPhysicalAddress(self):
        return self.__physicalAddress

    def getInternalAddress(self):
        return self.__internalAddress

    def getPublicAddress(self):
        return self.__publicAddress



    def getPhysicalName(self):
        return self.__physicalAddress.getName()


    def isLocal(self):
        return self.__physicalAddress.isLocal()

    def setServices(self, services, deploymentOrder):
        self.__services = services
        self.__deploymentOrder = deploymentOrder

    def getServices(self):
        return self.__services

    def getServiceNames(self):
        return [service.getName() for service in self.__services]

    def hasService(self, serviceName):
        return serviceName in self.getServiceNames()

    def getService(self, serviceName):
        for service in self.__services:
            if service.getName() == serviceName:
                return service

        return None

    def getServicesInDeploymentOrder(self):
        return [self.getService(serviceName)
                for serviceName in self.__deploymentOrder
                if self.hasService(serviceName)]

    def getServicesInAlphabeticOrder(self):
        return sorted(self.__services, key=lambda s : s.getName())

    def getDaemonServices(self):
        return [service for service in self.__services if service.isDaemon()]

    def getServicesInStartOrder(self):
        # Build a dict `service : set([dependencies...])`...
        dependencyGraph = dict((service, service.getStartupDependencies(self))
                               for service in self.getServices())

        # ...and sort it topologically. Finally, throw out anything that is not a daemon.
        import ext.toposort
        return [service for service in ext.toposort.toposort_flatten(dependencyGraph)]

    def getMachineStartupDependencies(self):
        machines = set()

        for service in self.__services:
            machines |= service.getMachineStartupDependencies()

        if self in machines:
            machines.remove(self)
        return machines

    def getController(self, logger):
        if not self.isLocal():
            raise ServiceException("Remote machine cannot be controlled")

        return app.servicecontrol.MachineController.forMachine(self, logger)

    def getCluster(self):
        return self.__cluster

    def isClustered(self):
        return self.__cluster is not None

class Cluster(object):
    MAX_MACHINES = 2

    """ Represents a set of machines (not services!) that form a cluster. """

    def __init__(self, name, internalAddress, publicAddress, serviceNames):
        assert not internalAddress or isinstance(internalAddress, Address)
        assert not publicAddress or isinstance(publicAddress, Address)
        assert internalAddress or publicAddress

        self.__name = name
        self.__internalAddress = internalAddress
        self.__publicAddress = publicAddress
        self.__serviceNames = serviceNames

        self.__machines = []

    def getName(self):
        return self.__name

    def getInternalName(self):
        return self.__internalAddress.getName() if self.__internalAddress else None

    def getPublicName(self):
        return self.__publicAddress.getName() if self.__publicAddress else None

    def getMachines(self):
        return self.__machines

    def getServiceNames(self):
        return self.__serviceNames

    def addMachine(self, machine):
        assert isinstance(machine, Machine)

        if len(self.__machines) >= Cluster.MAX_MACHINES:
            raise ConfigurationException("A cluster cannot have more than %d machines" % Cluster.MAX_MACHINES)

        self.__machines.append(machine)


class Site(object):
    """ Represents the landscape configuration for the entire site (i.e., all machines) """

    def __init__(self, localMachine, allMachines, clusters = {}):
        assert isinstance(localMachine, Machine)
        assert allMachines and isinstance(allMachines[0], Machine)
        assert not clusters or isinstance(clusters[0], Cluster)

        self.__localMachine = localMachine
        self.__allMachines = allMachines
        self.__clusters = clusters

    def getMachine(self, physicalName):
        if physicalName == "localhost":
            return self.getLocalMachine()

        for machine in self.__allMachines:
            if machine.getPhysicalName() == physicalName:
                return machine

        return None

    def getAllMachines(self):
        return self.__allMachines

    def getLocalMachine(self):
        return self.__localMachine

    def getMachinesByService(self, serviceName):
        return [machine for machine in self.__allMachines if machine.hasService(serviceName)]

    def getAllClusters(self):
        return self.__clusters

    def getCluster(self, name):
        for cluster in self.__clusters:
            if cluster.getName() == name:
                return cluster

        return None

    def resolveEndpointDependencies(self):
        for machine in self.getAllMachines():
            for service in machine.getServices():
                for reqdEndpoint in service.getRequiredEndpoints():
                    # Find a machine providing this service
                    providerMachines = self.getMachinesByService(reqdEndpoint.getServiceName())

                    if len(providerMachines) == 0:
                        if not machine.isLocal():
                            # Our knowledge about remotely deployed services is limited - maybe
                            # the user omitted something in the local site-options.conf, or maybe
                            # we do not have the manifest for one of the services. As this dependency
                            # is not really relevant to our local services, ignore it.
                            continue

                        if reqdEndpoint.isRequiredForDeployment():
                            raise UnsatisfiedDependencyException(
                                "Service '%s' on '%s' needs '%s', but this is not deployed anywhere" %
                                (service.getName(), machine.getPhysicalName(), reqdEndpoint.getServiceName()))
                        else:
                            # Nevermind, the service can run without.
                            continue

                    elif len(providerMachines) > 1:
                        # Be solipsistic and assume that the local service is always used. Note that
                        # "local" changes depending on the viewpoint.
                        # This becomes important in scenarios like registering an integration server.
                        if self.getLocalMachine() in providerMachines:
                            providerMachine = self.getLocalMachine()
                        elif all(machine.isClustered() for machine in providerMachines):
                            # We are dealing with a cluster. It does not matter which machine we
                            # chose, communication will go thru the load balancer anyway.
                            providerMachine = providerMachines[0]
                        else:
                            raise AmbiguousDependencyException(
                                "Service '%s' (needed by %s) is provided on multiple machines (%s) - do not know which one to use" %
                                (reqdEndpoint.getServiceName(), service.getName(), ", ".join([m.getPhysicalName() for m in providerMachines])))
                    else:
                        providerMachine = providerMachines[0]

                    # Now that we know the machine, find the right endpoint.
                    providerService = providerMachine.getService(reqdEndpoint.getServiceName())

                    if providerMachine == machine:
                        # Same machine - in this case, prefer an 'internal' endpoint.
                        scopePreference = ["internal", "public"]
                    else:
                        # Difference machines - prefer a 'public' endpoint because an 'internal' one
                        # may either not be reachable (firewall, binding) or may not be appropriately secured.
                        scopePreference = ["public", "internal"]

                    for scope in scopePreference:
                        providerEndpoint = providerService.getProvidedEndpoint(reqdEndpoint.getId(), scope)
                        if providerEndpoint is not None:
                            break

                    if providerEndpoint is None:
                        raise MissingEndpointException(
                            "Service '%s' on '%s' needs '%s' which was found on %s, but it does not expose %s" %
                            (service.getName(), machine.getPhysicalName(), reqdEndpoint.getServiceName(),
                             providerMachine, reqdEndpoint.getId()))

                    # We have a match
                    reqdEndpoint._snapProvider(providerService, providerEndpoint)
                    providerEndpoint._snapClient(service)

    def getMachineStartupOrder(self):
        # Build a dict `service : set([dependencies...])`...
        dependencyGraph = dict((machine, machine.getMachineStartupDependencies())
                               for machine in self.getAllMachines())

        # ...and sort it topologically. Finally, throw out anything that is not a daemon.
        import ext.toposort
        try:
            return ext.toposort.toposort_flatten(dependencyGraph)
        except ValueError:
            raise CyclicDependencyInMachineStartupOrder(
                "Cyclic dependency found in machine startup order - consider using a different landscape. Details: %r" % dependencyGraph)

    def getController(self, logger):
        return app.servicecontrol.SiteController.forSite(self, logger)

    def getAllServiceNames(self):
        serviceNames = []

        for machine in self.__allMachines:
            serviceNames += machine.getServiceNames()

        return list(set(serviceNames))

    @staticmethod
    def fromSiteOrRuntimeOptions(siteOrRunimeOptions, serviceManifestCollection):
        if ServiceTokenParser.isApplicable(siteOrRunimeOptions):
            return ServiceTokenParser(siteOrRunimeOptions, serviceManifestCollection).parse()
        else:
            return LegacyHostTokenParser(siteOrRunimeOptions, serviceManifestCollection).parse()

class BaseLandscapeParser(object):
    def __init__(self, siteOrRunimeOptions, serviceManifestCollection):
        assert isinstance(serviceManifestCollection, app.servicemanifest.ServiceManifestCollection)

        self.serviceManifests = serviceManifestCollection

        # Ensure that any {__OTHER_PROP__} referenes are resolved.
        self.options = siteOrRunimeOptions.resolveReferences()

    def __loadDaemonStartupInfo(self, manifest):
        if not manifest.isDaemon():
            return None
        else:
            if (manifest.getDaemonInfo().getUser() and manifest.getDaemonInfo().getGroup() and manifest.getDaemonInfo().getStartupTimeout()) is None:
                return DaemonStartupInfo(None, None, None)
            else:
                return DaemonStartupInfo(
                    manifest.getDaemonInfo().getUser().resolveValue(self.options),
                    manifest.getDaemonInfo().getGroup().resolveValue(self.options),
                    manifest.getDaemonInfo().getStartupTimeout())

    def loadService(self, serviceName, internalAddress=None, publicAddress=None):
        manifest = self.serviceManifests.get(serviceName)
        folderInfo = []

        reqdEndpoints = [RequiredEndpoint(
                            endpInfo.getId(),
                            endpInfo.getService(),
                            endpInfo.isRequiredForDeployment(),
                            endpInfo.isRequiredForStartup())
                         for endpInfo in manifest.getRequiredEndpoints()]

        provEndpoints = [ProvidedEndpoint(
                            endpInfo.getId(),
                            endpInfo.getScope(),
                            int(endpInfo.getPort().resolveValue(self.options))
                                if endpInfo.getPort() else None,
                            endpInfo.getVia())
                         for endpInfo in manifest.getProvidedEndpoints()]

        daemonStartupInfo = self.__loadDaemonStartupInfo(manifest)
        for f in manifest.getFolders(app.defaults.DEFAULT_DATA_DIR):
            folderInfo.append(ServiceFolder(
                FolderInfo(
                    f.getPath(),
                    f.getState(),
                    f.getMode(),
                    f.getOwner().resolveValue(self.options) if f.getOwner() else None,
                    f.getGroup().resolveValue(self.options) if f.getGroup() else None,
                    f.getContext(),
                    f.isRecursive())))

        return Service(
            serviceName,
            daemonStartupInfo,
            provEndpoints,
            reqdEndpoints,
            manifest.getIntegrityConfFiles(),
            manifest.getDataChecksumConfig(),
            folderInfo,
            self.serviceManifests.isAlias(serviceName),
            internalAddress,
            publicAddress)

    def purgeUnrecognizedServices(self, serviceNames):
        return [s for s in serviceNames if self.serviceManifests.get(s) is not None]

    def determineEffectiveSetOfServicesToLoad(self, serviceNames, isLocal):
        # The list of services may contain aliases - resolve them. On remote hosts, ignore
        # any aliases that we cannot resolve because we might not have all the manifests
        # locally.
        serviceNames = self.serviceManifests.translateAliases(serviceNames, not isLocal)

        if not isLocal:
            # Purge any services which we do not have manifests for.
            serviceNames = self.purgeUnrecognizedServices(serviceNames)

        # Now that we know what gets deployed, make sure that...
        # (1) The manifest collection contains entries for both the "real" names and the alias names
        #     of all relevant services.
        # (2) The service list (of the machine) contains both the alias and real services.
        serviceNames = self.serviceManifests.resolveAliases(serviceNames)

        # Resolve deployment dependencies
        serviceNames = self.serviceManifests.getTransitiveDeploymentDependencySet(serviceNames, ignoreUnrecognizedServices=not isLocal)

        return serviceNames

    def loadMachine(self, physicalAddress, internalAddress, publicAddress, services, cluster):
        machine = Machine(
            physicalAddress,
            internalAddress,
            publicAddress,
            services,
            self.serviceManifests.getDeploymentOrder([s.getName() for s in services]),
            cluster=cluster)

        isLocal = physicalAddress.isLocal()

        # Now that all services are loaded, wire up the deployment dependencies (non-transitive)
        for service in machine.getServices():
            transitiveDependencyServiceNames = self.serviceManifests.getTransitiveDeploymentDependencySet(
                [service.getName()],
                False,
                not isLocal)
            dependencyServiceNames = self.serviceManifests.get(service.getName()).getDeploymentDependencies()
            containerServiceName = self.serviceManifests.get(service.getName()).getContainer()
            if containerServiceName and containerServiceName not in dependencyServiceNames:
                raise ConfigurationException(
                    "Service %s is hosted in container %s, but it lacks a corresponding deployment dependency" %
                    (service.getName(), containerServiceName))

            if not isLocal:
                # Purge any services which we do not have manifests for.
                dependencyServiceNames = self.purgeUnrecognizedServices(dependencyServiceNames)
                transitiveDependencyServiceNames = self.purgeUnrecognizedServices(transitiveDependencyServiceNames)
                containerServiceName = containerServiceName if self.serviceManifests.get(containerServiceName) else None

            if dependencyServiceNames:
                service.setDeploymentDependencies(
                    [machine.getService(name) for name in dependencyServiceNames],
                    [machine.getService(name) for name in transitiveDependencyServiceNames],
                    machine.getService(containerServiceName) if containerServiceName else None)

        return machine

    def getDatabaseType(self):
        type = self.options.get("DATABASE_TYPE", None)
        if not type:
            return "postgres"
        elif type == "postgresql":
            return "postgres"
        elif type == "oracle":
            return "oracle"
        else:
            raise ConfigurationException("Unrecognized database type '%s'" % type)

    @staticmethod
    def setAdditionalTokens(siteOrRunimeOptions):
        # Quirk: If SSL is on, the HTTPD_PORT default value changes.
        # Todo: Turn into a runtime-option-override
        if siteOrRunimeOptions.get("SSL", "on") == "off" and not siteOrRunimeOptions.has_key("HTTPD_PORT"):
            siteOrRunimeOptions = siteOrRunimeOptions.overlayBy(
                app.options.RuntimeOptions({"HTTPD_PORT": "80"}))

        return siteOrRunimeOptions

class LegacyHostTokenParser(BaseLandscapeParser):
    """ Parser for the old (pre 17.1) site-options.conf format """

    def __init__(self, siteOrRunimeOptions, serviceManifestCollection):
        assert isinstance(serviceManifestCollection, app.servicemanifest.ServiceManifestCollection)

        super(LegacyHostTokenParser, self).__init__(siteOrRunimeOptions, serviceManifestCollection)

        self.__hostTokens = {}

    appsToServicesMapping = {
            "app": "ctfcore",
            "database": "ctfcore-database",
            "datamart": "ctfcore-datamart",
            "cvs": "cvs",
            "subversion": "subversion",
            "etl": "etl",
            "gerrit": "gerrit",
            "binary": "binary",
            "indexer": "search",
            "codesearch": "codesearch-elasticsearch"
        }

    @staticmethod
    def isApplicable(siteOptions):
        assert isinstance(siteOptions, app.options.OptionsBase)

        for key in siteOptions.keys():
            if key.startswith("HOST_"):
                return True

        return False

    def __isDatamartAndDatabaseCohostedOnThisMachine(self, appsList):
        defaultPort = app.defaults.POSTGRES_DEFAULT_PORT

        return "database" in appsList \
            and "datamart" in appsList \
            and self.getDatabaseType() == "postgres" \
            and self.options.get("DATABASE_PORT", defaultPort) == self.options.get("REPORTS_DATABASE_PORT", defaultPort)

    def __getServicesFromHostToken(self, appsList):
        services = []

        for appToken in appsList:
            if LegacyHostTokenParser.appsToServicesMapping.has_key(appToken):
                # Translate to service
                services += [LegacyHostTokenParser.appsToServicesMapping[appToken]]
            else:
                # Not an app, must be a service
                services += [appToken]

        if "database" in appsList:
            # "database" is just an alias - replace by concrete service, this
            # will in turn trigger the database alias to be added again later
            services.remove("ctfcore-database")
            services += ["ctfcore-database-" + self.getDatabaseType()]

        if "datamart" in appsList:
            # "datamart" is just an alias - replace by concrete service, this
            # will in turn trigger the database alias to be added again later
            services.remove("ctfcore-datamart")

            if self.__isDatamartAndDatabaseCohostedOnThisMachine(appsList):
                # Use cohosted service
                services += ["ctfcore-datamart-postgres-cohosted"]
            else:
                # Use dedicated service
                services += ["ctfcore-datamart-" + self.getDatabaseType()]

        if "gerrit-database" in appsList:
            # "gerrit-database" is just an alias - replace by concrete service, this
            # will in turn trigger the database alias to be added again later
            services.remove("gerrit-database")

            # Gerrit does not support Oracle - so we have to hardcode Postgres here.
            services += ["gerrit-database-postgres"]

        if "reviewboard-database" in appsList:
            # reviewboard-database" is just an alias - replace by concrete service, this^M
            # will in turn trigger the database alias to be added again later
            services.remove("reviewboard-database")

            # Reviewboard doesnot support oracle - we have to hardcode postgres
            services += ["reviewboard-database-postgres"]

        if "binary-database" in appsList:
            # "binary-database" is just an alias - replace by concrete service, this
            # will in turn trigger the database alias to be added again later
            services.remove("binary-database")

            # Binary does not support Oracle - so we have to hardcode Postgres here.
            services += ["binary-database-postgres"]


        allApps = " ".join(self.__hostTokens.values())
        if "app" in appsList:
            # We want the database administration tools to be installed where
            # ctfcore is deployed. Technically, it does not matter where the
            # database administration tools are deployed, but this has been the
            # convention.
            services += ["ctfcore-database-admin-" + self.getDatabaseType()]

            if "datamart" in allApps:
                # The datamart is installed somewhere (does not matter on which host,
                # so add the datamart administration tools as well.
                services += ["ctfcore-datamart-admin-" + self.getDatabaseType()]

            if not "indexer" in allApps and not "search" in allApps:
                # If indexer/search is not explicitly listed anywhere, then
                # install it on the app box.
                services += ["search"]

            if not "mail" in allApps:
                # If mail is not explicitly listed anywhere, then install it on the app box
                # to mimic old behavior where it was always deployed alongside with jboss.
                services += ["mail"]

        if not "service-monitor" in allApps:
            if self.options.has_key("ENABLE_SERVICE_MONITORING") and self.options["ENABLE_SERVICE_MONITORING"] == "true":
                services += ["service-monitor"]

        if "gerrit" in appsList and not "gerrit-database" in allApps:
            # The 'gerrit' token used to imply a local database - so if we do not have
            # 'gerrit-database' listed in any HOST token, assume the database is supposed
            # to be installed locally.
            services += ["gerrit-database-postgres"]


        if "database" in appsList and "reviewboard" in allApps and not "reviewboard-database" in allApps:
            # If review-board database is not deployed anywhere then deploy it on
            # the machine where ctfcore-database is running
            services += ["reviewboard-database-postgres"]

        if "binary" in appsList and not "binary-database" in allApps:
            # The 'binary' token used to imply a local database - so if we do not have
            # 'binary-database' listed in any HOST token, assume the database is supposed
            # to be installed locally.
            services += ["binary-database-postgres"]

        return services

    def parse(self):
        self.__hostTokens = dict((key, value) for key, value in self.options.iteritems() if key.rfind("HOST_") == 0)

        machines = []
        localMachine = None

        # Load config for each machine.
        for key, value in self.__hostTokens.iteritems():
            apps = set(value.split())

            # Determine physical host name
            physicalName = key.split("_")[1]
            isLocal = physicalName == "localhost" or physicalName == Machine.getLocalHostName()
            if physicalName == "localhost":
                # We need the proper name from now on.
                physicalName = Machine.getLocalHostName()
            elif app.util.isAnIpAddress(physicalName):
                raise ConfigurationException("HostName '%s' is an IP Address. Please use only domain names." % physicalName)


            # Determine installed services from HOST_ tokens.
            serviceNames = self.__getServicesFromHostToken(apps)

            # Determine public host name
            if isLocal and self.options.has_key("DOMAIN_localhost"):
                publicName = self.options["DOMAIN_localhost"]
            else:
                publicName = self.options.get("DOMAIN_%s" % physicalName, physicalName)

            # Create machine and service objects
            serviceNames = self.determineEffectiveSetOfServicesToLoad(serviceNames, isLocal)
            services = [self.loadService(serviceName) for serviceName in serviceNames]
            machine = self.loadMachine(
                Address(physicalName, isLocal),
                Address(physicalName, isLocal),
                Address(publicName, isLocal),
                services,
                None)
            machines.append(machine)
            if isLocal:
                localMachine = machine

        if localMachine is None:
            raise ConfigurationException("Missing HOST token for local machine")

        return Site(localMachine, machines)

class ServiceTokenParser(BaseLandscapeParser):
    """ Parser for the new (post 17.1) site-options.conf format """

    def __init__(self, siteOrRunimeOptions, serviceManifestCollection):
        assert isinstance(serviceManifestCollection, app.servicemanifest.ServiceManifestCollection)

        super(ServiceTokenParser, self).__init__(siteOrRunimeOptions, serviceManifestCollection)

    @staticmethod
    def isApplicable(siteOptions):
        assert isinstance(siteOptions, app.options.OptionsBase)

        for key in siteOptions.keys():
            if key.endswith(":SERVICES") or key.endswith(":CLUSTER_SERVICES"):
                return True

        return False

    def __isDatamartAndDatabaseCohostedOnThisMachine(self, serviceNames):
        defaultPort = app.defaults.POSTGRES_DEFAULT_PORT

        return "ctfcore-database" in serviceNames \
            and "ctfcore-datamart" in serviceNames \
            and self.getDatabaseType() == "postgres" \
            and self.options.get("DATABASE_PORT", defaultPort) == self.options.get("REPORTS_DATABASE_PORT", defaultPort)

    def __getServicesFromServiceToken(self, serviceNames, isDatamartDeployed):
        # Check if we run in Postgres/cohosted mode before we change the service names.
        postgresCohosted = self.__isDatamartAndDatabaseCohostedOnThisMachine(serviceNames)

        if "ctfcore-database" in serviceNames:
            # "ctfcore-database" is just an alias - replace by concrete service, this
            # will in turn trigger the database alias to be added again later
            serviceNames.remove("ctfcore-database")
            serviceNames += ["ctfcore-database-" + self.getDatabaseType()]

        if "ctfcore-datamart" in serviceNames:
            # "ctfcore-datamart" is just an alias - replace by concrete service, this
            # will in turn trigger the database alias to be added again later
            serviceNames.remove("ctfcore-datamart")

            if postgresCohosted:
                # Use cohosted service
                serviceNames += ["ctfcore-datamart-postgres-cohosted"]
            else:
                # Use dedicated service
                serviceNames += ["ctfcore-datamart-" + self.getDatabaseType()]

        if "ctfcore" in serviceNames:
            # We want the database administration tools to be installed where
            # ctfcore is deployed. Technically, it does not matter where the
            # database administration tools are deployed, but this has been the
            # convention.
            serviceNames += ["ctfcore-database-admin-" + self.getDatabaseType()]

            if isDatamartDeployed:
                # The datamart is installed somewhere (does not matter on which host,
                # so add the datamart administration tools as well.
                serviceNames += ["ctfcore-datamart-admin-" + self.getDatabaseType()]

        return serviceNames

    def parse(self):
        clusterNames = [key.split(":")[0] for key in self.options.keys() if key.endswith(":CLUSTER_SERVICES")]
        machineNames = [key.split(":")[0] for key in self.options.keys() if key.endswith(":SERVICES")] + \
                       [key.split(":")[0] for key in self.options.keys() if key.endswith(":CLUSTER")]

        machines = []
        clusters = {}
        localMachine = None

        # Check if the datamart is installed anywhere (it is not clusterable, so nevermind
        # checking clusters as well).
        isDatamartInstalled = False
        for machineName in machineNames:
            if app.util.isAnIpAddress(machineName):
                raise ConfigurationException("HostName '%s' is an IP Address. Please use only domain names." % machineName)

            if "ctfcore-datamart" in self.options.get(machineName + ":SERVICES", "").split():
                isDatamartInstalled = True

        listenIps = [key for key in self.options.keys() if key.endswith(":LISTEN_IP")]
        for listenIp in listenIps:
            if self.options.get(listenIp) == "" or not app.util.isAnIpAddress(self.options.get(listenIp)):
                raise ConfigurationException("Please use correct ip address for LISTEN_IP token - '%s'." % listenIp)

        # Load clusters.
        for clusterName in clusterNames:
            publicName   = self.options.get(clusterName + ":PUBLIC_FQDN")
            internalName = self.options.get(clusterName + ":INTERNAL_FQDN")

            if not publicName and not internalName:
                raise ConfigurationException("Cluster '%s' neither has an internal nor public FQDN assigned" % clusterName)

            serviceNames = self.__getServicesFromServiceToken(
                self.options[clusterName + ":CLUSTER_SERVICES"].split(),
                isDatamartInstalled)

            if "service-monitor" not in serviceNames:
                raise ConfigurationException(
                    "The service 'service-monitor' needs to be deployed on all clusters. Please add this service to the CLUSTER_SERVICES token")
            clusters[clusterName] = Cluster(
                clusterName,
                Address(internalName) if internalName else None,
                Address(publicName) if publicName else None,
                serviceNames)

        # Load machines.
        for machineName in machineNames:
            isLocal = machineName == "localhost" or machineName == Machine.getLocalHostName()

            physicalName = machineName if not isLocal else Machine.getLocalHostName()
            machineListenIp = self.options.get(machineName + ":LISTEN_IP", None)
            physicalAddress = Address(physicalName, isLocal)

            clusterName = self.options.get(machineName + ":CLUSTER")
            if clusterName:
                # This machine is part of a cluster.
                cluster = clusters.get(clusterName)
                if not cluster:
                    raise ConfigurationException("Machine '%s' references nonexisting cluster '%s'" %
                                                 (machineName, clusterName))

                if self.options.has_key(machineName + ":SERVICES"):
                    raise ConfigurationException("Machine '%s' is part of a cluster and thus cannot run " % machineName +
                                                 "additional services - remove the SERVICES token")

                if self.options.has_key(machineName + ":PUBLIC_FQDN"):
                    raise ClusterNodeFqdnException("Machine '%s' is part of a cluster and must not specify a PUBLIC_FQDN" % machineName)

                if self.options.has_key(machineName + ":INTERNAL_FQDN"):
                    raise ClusterNodeFqdnException("Machine '%s' is part of a cluster and must not specify a PUBLIC_FQDN" % machineName)

                # Derive list of services from cluster.
                serviceNames = cluster.getServiceNames()

                # Derive FQDNs from cluster. Never use the physical name, even for internal communication,
                # as this would sidestep the load balancer.
                #
                # Note that as the LB is in between, these addresses are NOT local.
                publicAddress   = Address(cluster.getPublicName(), False, machineListenIp)
                internalAddress = Address(cluster.getInternalName() or cluster.getPublicName(), False, machineListenIp)

            else:
                # Independent machine.
                cluster = None
                serviceNames = self.__getServicesFromServiceToken(
                    self.options[machineName + ":SERVICES"].split(),
                    isDatamartInstalled)

                # As this is a standalone machine, the internal/public address may indeed be local.
                publicAddress   = Address(self.options.get(machineName + ":PUBLIC_FQDN", physicalName), isLocal, machineListenIp)
                internalAddress = Address(self.options.get(machineName + ":INTERNAL_FQDN", physicalName), isLocal, machineListenIp)

            # Create machine and service objects
            serviceNames = self.determineEffectiveSetOfServicesToLoad(serviceNames, isLocal)
            services = []
            for serviceName in serviceNames:
                # Check for service specific listen IP (these are always specific to one machine,
                # so disregard clusters here).
                serviceListenIp = self.options.get("%s:%s:LISTEN_IP" % (machineName, serviceName), machineListenIp)

                # Check for a service specific FQDN:
                # - In case of a cluster, that is <cluster>:<service>:INTERNAL_FQDN.
                # - In case of a standalone machine, that is <machine>:<service>:INTERNAL_FQDN.
                # Ditto for public FQDN.public

                serviceSpecificInternalFqdn = self.options.get("%s:%s:INTERNAL_FQDN" % (clusterName or machineName, serviceName))
                if serviceSpecificInternalFqdn:
                    servicInternalAddress = Address(serviceSpecificInternalFqdn, internalAddress.isLocal(), serviceListenIp)
                else:
                    servicInternalAddress  = Address(internalAddress.getName(), internalAddress.isLocal(), serviceListenIp)

                serviceSpecificPublicFqdn = self.options.get("%s:%s:PUBLIC_FQDN" % (clusterName or machineName, serviceName))
                if serviceSpecificPublicFqdn:
                    servicePublicAddress = Address(serviceSpecificPublicFqdn, publicAddress.isLocal(), serviceListenIp)
                else:
                    servicePublicAddress  = Address(publicAddress.getName(), publicAddress.isLocal(), serviceListenIp)

                services.append(self.loadService(serviceName, servicInternalAddress, servicePublicAddress))

            machine = self.loadMachine(
                physicalAddress,
                internalAddress,
                publicAddress,
                services,
                cluster)
            machines.append(machine)

            # Now that all services have been loaded and deployment dependencies resolved, check
            # if all services are properly compatible with the setup.
            #
            # Note: It is important to check the dependencies as well - therefore, we cannot do
            # this earlier.

            for serviceName in machine.getServiceNames():
                haInfo = self.serviceManifests.get(serviceName).getHighAvailabilityInfo()

                if cluster:
                    # Services deployed in a cluster must be clusterable.
                    if not haInfo.isClusterable():
                        raise ServiceNotClusterableException("Service '%s' cannot be deployed on a cluster" % serviceName)

                elif len(clusters):
                    # If there is a cluster around, the service must be compatible.
                    if not haInfo.isCompatible():
                        raise ServiceNotClusterCompatible("Service '%s' cannot be deployed in an environment with clusters" % serviceName)

            if isLocal:
                localMachine = machine

            if cluster:
                cluster.addMachine(machine)

        if localMachine is None:
            raise ConfigurationException("No service configuration for local machine found")

        return Site(localMachine, machines, clusters.values())

class Landscape(object):
    """ Singleton object describing the current landscape """

    @staticmethod
    def isInitialized():
        try:
            Landscape.getSite()
            return True
        except:
            return False

    @staticmethod
    def initializeFromSiteOptionsPath(siteOptionsPath):
        # Note: site-options.conf tokens may reference each other using {__OTHERTOKEN__} syntax,
        # so these references need to be resolved. SITE_DIR used to be an implicit token (derived
        # from the -d command line switch, so this needs to be made available for resolution to
        # work in all cases.
        siteOptions = app.options.SiteOptions \
            .readFromFilePath(siteOptionsPath, True) \
            .set("SITE_DIR", app.defaults.DEFAULT_SITE_DIR) \
            .resolveReferences()

        # Initialize statics.
        Landscape.__siteOptionsPath = siteOptionsPath
        Landscape.__siteOptions = BaseLandscapeParser.setAdditionalTokens(siteOptions)
        Landscape.__site = app.landscape.Site.fromSiteOrRuntimeOptions(
            Landscape.__siteOptions,
            app.servicemanifest.ServiceManifestCollection().loadFromDefaultLocation())
        Landscape.__site.resolveEndpointDependencies()

        # Validate that the site-options.conf does not contain any obvious errors.
        siteOptions = siteOptions.validate(
            app.tokens.defines.TokenDef(),
            Landscape.__site.getAllServiceNames())

    @staticmethod
    def reinitialize():
        Landscape.initializeFromSiteOptionsPath(Landscape.__siteOptionsPath)

    @staticmethod
    def getSite():
        return Landscape.__site

    @staticmethod
    def getSiteOptions():
        return Landscape.__siteOptions

    @staticmethod
    def getSiteOptionsPath():
        return Landscape.__siteOptionsPath


class ConfigurationException(Exception):
    pass

class ServiceNotClusterableException(ConfigurationException):
    pass

class ClusterNodeFqdnException(ConfigurationException):
    pass

class ServiceNotClusterCompatible(ConfigurationException):
    pass

class UnresolvedEndpointDependenciesException(Exception):
    pass

class UnsatisfiedDependencyException(ConfigurationException):
    pass

class AmbiguousDependencyException(ConfigurationException):
    pass

class MissingEndpointException(ConfigurationException):
    pass

class CyclicDependencyInMachineStartupOrder(ConfigurationException):
    pass

class NetworkException(ConfigurationException):
    pass

class ServiceException(Exception):
    pass

class ServiceFolderException(Exception):
    pass
