﻿import app.util
import app.logutil
import platform
import commands
import os

class ServiceManager(object):
    def __init__(self, serviceName, log):
        self.serviceName = serviceName
        self.log = log

    @staticmethod
    def isRunningInContainer():
        return os.environ.get("container", None) == "docker"

    @staticmethod
    def isAvailable():
        # If we are inside a container, then there is no service manager. Otherwise,
        # there is.
        return not ServiceManager.isRunningInContainer()

    @staticmethod
    def __getRhelVersion():
        (distname,version,id) = platform.linux_distribution()
        if distname not in ['Red Hat Enterprise Linux Server', 'CentOS Linux', 'CentOS']:
            raise Exception("Unrecognized platform")

        try:
            return int(platform.linux_distribution()[1].split('.')[0])
        except:
            raise Exception("You are running the installation on an unsupported operating system.")

    @staticmethod
    def getService(serviceName, log=None):
        if not log:
            log = app.logutil.getRuntimeLogger(app.defaults.DEFAULT_LOG_DIR, False)

        os_version = ServiceManager.__getRhelVersion()

        if not ServiceManager.isAvailable():
            return NopService(serviceName, log)

        elif os_version >= 7:
            return SystemdService(serviceName, log)

        elif os_version == 6:
            return InitdService(serviceName, log)

        else:
            raise Exception("Unsupported OS version")

    @staticmethod
    def reload():
        if ServiceManager.isAvailable() and ServiceManager.__getRhelVersion() >= 7:
            return os.system("/bin/systemctl daemon-reload") == 0
        else:
            # No need to reload anything on other platforms
            pass

class LinuxService(object):
    def __init__(self, serviceName, log):
        self.serviceName = serviceName
        self.log = log

    def runServiceCommand(self, command, failOnError=False):
        (exitCode, output) = commands.getstatusoutput(command + " 2>&1" )

        if exitCode == 0:
            self.log.info("Command '%s' completed successfully" % command)
            return True

        else:
            self.log.warn("Command '%s' failed with exit code %d: %s" % (command, exitCode, output))

            if failOnError:
                raise LinuxServiceException("Command '%s' failed with exit code %d: %s", (command, exitCode, output))
            else:
                return False

    def stopAndDisable(self, failOnError=True):
        self.stop(failOnError)
        self.disable(failOnError)

    def disable(self, failOnError=True):
        raise Exception("override me", failOnError=True)

    def enable(self, failOnError=True):
        raise Exception("override me", failOnError=True)

    def isEnabled(self):
        raise Exception("override me")

    def isRunning(self):
        raise Exception("override me")

    def stop(self, failOnError=True):
        raise Exception("override me", failOnError=True)

    def start(self, failOnError=True):
        raise Exception("override me")

class SystemdService(LinuxService):
    def __init__(self, serviceName, log):
        self.systemctlPath = '/bin/systemctl'

        super(SystemdService, self).__init__(serviceName, log)

    def disable(self, failOnError=False):
        return self.runServiceCommand(self.systemctlPath + " disable " + self.serviceName, failOnError)

    def enable(self, failOnError=False):
        return self.runServiceCommand(self.systemctlPath + " enable " + self.serviceName, failOnError)

    def isEnabled(self):
        return self.runServiceCommand(self.systemctlPath + " is-enabled " + self.serviceName)

    def isRunning(self):
        return self.runServiceCommand(self.systemctlPath + " status " + self.serviceName)

    def stop(self, failOnError=False):
        return self.runServiceCommand(self.systemctlPath + " stop " + self.serviceName, failOnError)

    def start(self, failOnError=False):
        return self.runServiceCommand(self.systemctlPath + " start " + self.serviceName, failOnError)

class InitdService(LinuxService):
    def __init__(self, serviceName, log, initdScriptPath = None):
        self.chkconfigPath = '/sbin/chkconfig'
        if not initdScriptPath:
            self.initdScriptPath = '/etc/init.d/%s' % serviceName

        super(InitdService, self).__init__(serviceName, log)

    def disable(self, failOnError=False):
        self.runServiceCommand(self.chkconfigPath + " " + self.serviceName + " off", failOnError)
        return self.runServiceCommand(self.chkconfigPath + " --del " + self.serviceName, failOnError)

    def enable(self, failOnError=False):
        return self.runServiceCommand(self.chkconfigPath + " --level 345 " + self.serviceName + " on", failOnError)

    def isEnabled(self):
        return self.runServiceCommand(self.chkconfigPath + " " + self.serviceName)

    def isRunning(self):
        return self.runServiceCommand(self.initdScriptPath + " status")

    def stop(self, failOnError=False):
        return self.runServiceCommand(self.initdScriptPath + " stop", failOnError)

    def start(self, failOnError=False):
        return self.runServiceCommand(self.initdScriptPath + " start", failOnError)

class NopService(LinuxService):
    def disable(self, failOnError=False):
        # Containers do not have services, so there is nothing to disable.
        pass

    def enable(self, failOnError=False):
        # Containers do not have services, so there is nothing to enable.
        pass

    def isEnabled(self):
        # Containers do not have services, so there is nothing to enable.
        pass

    def isRunning(self):
        # Containers do not have services, so there is nothing to run.
        pass

    def stop(self, failOnError=False):
        # Containers do not have services, so there is nothing to stop.
        pass

    def start(self, failOnError=False):
        # Containers do not have services, so there is nothing to stop.
        pass

class LinuxServiceException(Exception):
    pass
