﻿
import logging
import os
import sys
import app.util
import app.defaults
import app.colors

## Initialising Global Variables
## This global variable is used to hold the status whether the stdout is already
## streamed to receive $log messages.
## Used by the getLogger method to decide whether the stdout has to be added to
## the stream handler or not.  Without this flag checking, every logger object
## creation may add the stdout to the stream handler and results in duplicate log
## entries
global isStdoutStreamed
isStdoutStreamed = False

def getLogger(logName, logLevel, logFormat=None, logFile=None, logToStdErr =
                True, logToStdOut = True):
    """
        Return the logger with the given options
    """
    stdLogFormat = '%(pathname)s %(levelname)s %(message)s'
    hndlrs = []
    global isStdoutStreamed
    if not logFormat:
        logFormat = stdLogFormat
    logger = logging.getLogger(logName)

    if logFile:
        try:
            if not os.access(os.path.dirname(logFile), os.F_OK):
                os.makedirs(os.path.dirname(logFile))
            hndlrs.append(logging.FileHandler(logFile))
        except Exception:
            print ''.join(app.util.getStackTraceAndException())
            print "Error creating handler for logfile %s" % logFile
    if logToStdErr:
        try:
            # If we dont provide a stream to it, stderr is used
            hndlrs.append(logging.StreamHandler())
        except Exception:
            print "Error creating handler for std error"
    if (logToStdOut and not isStdoutStreamed):
        try:
            # If we dont provide a stream to it, stderr is used
            # Let's call out stdout explicitly
            hndlrs.append(logging.StreamHandler(sys.stdout))
            isStdoutStreamed = True
        except Exception:
            print "Error creating handler for std error"

    try:
        formatter = logging.Formatter(logFormat)
    except:
        print "Invalid log format %s" % logFormat

    for hndlr in hndlrs:
        hndlr.setFormatter(formatter)
        logger.addHandler(hndlr)

    if logLevel == 'WARNING':
        logger.setLevel(logging.WARNING)
    elif logLevel == 'DEBUG':
        logger.setLevel(logging.DEBUG)
    elif logLevel == 'ERROR':
        logger.setLevel(logging.ERROR)
    elif logLevel == 'CRITICAL':
        logger.setLevel(logging.CRITICAL)
    else:
        logger.setLevel(logging.INFO)
    return logger

def closeLogger(logger):
    """
    Close the logger
    """
    for hndlr in logger.handlers:
        logger.removeHandler(hndlr)

def getBootstrapLogger(logHome = '/tmp', logToStdOut=True):
    """
    Return the boot-strap logger, based on the settings in the Constants file
    """
    cfg = app.util.getRuntimeConfiguration()

    logName = 'StandardLogger'
    logLevel = logging.INFO
    if cfg.has_key('DEFAULT_LOG_LEVEL'):
        logLevel = cfg['DEFAULT_LOG_LEVEL']
    if cfg.has_key('RUNTIME_LOG_DIR'):
        logHome = cfg['RUNTIME_LOG_DIR']
    logFileName = 'bootstrap.log'
    logFormat = '%(levelname)s %(message)s'
    if cfg.has_key('DEFAULT_LOG_FORMAT'):
        logFormat = cfg['DEFAULT_LOG_FORMAT']
    logger = logging.getLogger(logName)
    if (len(logger.handlers) < 1):
        return getLogger(logName, logLevel, logFormat,
            os.path.join(logHome, logFileName), False, logToStdOut)
    else:
        return getLoggerByName(logName)

def getCronJobLogger(logHome = '/tmp', logToStdOut=True):
    """
    Return the cronjob logger, based on the settings in the Constants file
    """
    cfg = app.util.getRuntimeConfiguration()

    logName = 'StandardLogger'
    logLevel = logging.INFO
    if cfg.has_key('DEFAULT_LOG_LEVEL'):
        logLevel = cfg['DEFAULT_LOG_LEVEL']
    if cfg.has_key('RUNTIME_LOG_DIR'):
        logHome = cfg['RUNTIME_LOG_DIR']
    logFileName = 'cronjob.log'
    logFormat = '%(levelname)s %(message)s'
    if cfg.has_key('DEFAULT_LOG_FORMAT'):
        logFormat = cfg['DEFAULT_LOG_FORMAT']
    logger = logging.getLogger(logName)
    if (len(logger.handlers) < 1):
        return getLogger(logName, logLevel, logFormat,
            os.path.join(logHome, logFileName), False, logToStdOut)
    else:
        return getLoggerByName(logName)

def getRuntimeLogger(logHome = None, logToStdOut=False):
    if not logHome:
        logHome = os.path.normpath(os.path.join(app.defaults.DEFAULT_LOG_DIR, 'runtime'))

        if not(os.path.isdir(logHome)):
            os.makedirs(logHome)

    logName = 'StandardLogger'
    logger = logging.getLogger(logName)
    if (len(logger.handlers) < 1):
        return getLogger(
            logName,
            app.defaults.DEFAULT_LOG_LEVEL,
            app.defaults.DEFAULT_LOG_FORMAT,
            os.path.join(logHome, 'runtime.log'),
            False,
            logToStdOut)
    else:
        return getLoggerByName(logName)

def getLoggerByName(logName):
    """
    Return the logger with the given name
    """
    return logging.getLogger(logName)

def getNewLogger(logName, logFile, logLevel = None, logFormat = None,
            logToStdErr = False, logToStdOut = True):
    return getLogger(
        logName,
        app.defaults.DEFAULT_LOG_LEVEL,
        app.defaults.DEFAULT_LOG_FORMAT,
        logFile,
        logToStdErr,
        logToStdOut)

def printLog(logger, msgType, msgData, lineFeed=True):
    """Function for printing messages"""

    try:
        if len(msgData) == 0:
            return
    except:
        return

    suffix = "\n" if lineFeed else ""

    if msgType == "warn":
        if (logger.level <= logging.WARNING):
            sys.stdout.write(app.colors.Color.amber(msgData) + suffix)
        logger.warn(msgData)
    elif msgType == "error":
        if (logger.level <= logging.ERROR):
            sys.stdout.write(app.colors.Color.red(msgData) + suffix)
        logger.error(msgData)
    elif msgType == "debug":
        if (logger.level <= logging.DEBUG):
            sys.stdout.write(msgData + "\n")
        logger.info(msgData)
    elif msgType == "success":
        if (logger.level <= logging.INFO):
            sys.stdout.write(app.colors.Color.green(msgData) + suffix)
        logger.info(msgData)
    elif msgType == "header":
        if (logger.level <= logging.INFO):
            sys.stdout.write(app.colors.Color.blue(msgData) + suffix)
        logger.info(msgData)
    elif msgType == "title":
        if (logger.level <= logging.INFO):
            sys.stdout.write(app.colors.Color.blue(msgData) + suffix)
        logger.info(msgData)
    elif msgType == "notify" or msgType == "info":
        if (logger.level <= logging.INFO):
            sys.stdout.write(msgData + suffix)
        logger.info(msgData)
    else:
        if (logger.level <= logging.INFO):
            sys.stdout.write(msgData + suffix)
        logger.info(msgData)

    sys.stdout.flush()

def getNullLogger(logToStdErr = False, logToStdOut = True):
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

    logger = logging.getLogger("silent")
    logger.setLevel(logging.CRITICAL)
    logger.addHandler(NullHandler())
    return logger

def getFileName(logger):
    for handler in logger.handlers:
        if hasattr(handler, 'baseFilename'):
            return handler.baseFilename

    return None
