﻿import StringIO
import os
import re
import subprocess

import app.util
import app.defaults
import app.tokens.defines

class OptionsBase(object):
    def __init__(self, options):
        self.__options = options

    def __getitem__(self, key):
        return self.__options[key]

    def __repr__(self):
        return self.__options.__repr__()

    def get(self, key, defaultValue=None):
        return self.__options.get(key, defaultValue)

    def keys(self):
        return self.__options.keys()

    def items(self):
        return self.__options.items()

    def has_key(self, key):
        return self.__options.has_key(key)

    def iteritems(self):
        return self.__options.iteritems()

    def resolveReferences(self):
        """ Resolve all {__TOKEN__} references """

        processedOptions = self.__options.copy()
        while True:
            numberOfSubstitutionsMade = 0

            processor = app.preprocessor.TokenProcessor(processedOptions)
            for key, value in processedOptions.iteritems():
                processedValue = processor.substitute(value)

                if processedValue != value:
                    numberOfSubstitutionsMade += 1
                    processedOptions[key] = processedValue

            if numberOfSubstitutionsMade != 0:
                # At least one substituion made - this means we have to do at
                # least one more round as tokens can refer to other tokens.
                pass
            else:
                # No substitutions made, so there is no point in another round.
                break

        return self.new(processedOptions)

    def overlayBy(self, options):
        """ Merge options, with the passed in options taking priority """
        mergedOptions = self.__options.copy()
        mergedOptions.update(options.__options)
        return self.new(mergedOptions)

    def dict(self):
        return self.__options.copy()

    def new(self, options):
        raise Exception("Override me")

class RuntimeOptions(OptionsBase):
    """ Runtime options dictionary. """

    def __init__(self, options = {}):
        super(RuntimeOptions, self).__init__(options)

    def new(self, options):
        return RuntimeOptions(options)

    def deobfuscate(self, obfuscator):
        """ Deobfuscate everything """

        assert isinstance(obfuscator, Obfuscator)
        return obfuscator.deobfuscateOptions(self)

    def obfuscate(self, obfuscator):
        """ Obfuscate everything """

        assert isinstance(obfuscator, Obfuscator)
        return obfuscator.obfuscateOptions(self)

    def applyAutoData(self, autoData):
        return RuntimeOptions(dict((key, autoData.resolveAutoToken(key, value)) for key, value in self.iteritems()))

    def writeToFilePath(self, filePath):
        PropertiesFile.writeToFilePath(self.dict(), filePath)

    def writeToFile(self, file):
        PropertiesFile.writeToFile(self.dict(), file)

    @staticmethod
    def readFromFilePath(filePath, checkDuplicate=False):
        return RuntimeOptions(PropertiesFile.readFromFilePath(filePath, checkDuplicate))

    @staticmethod
    def readFromSnippet(snippet, checkDuplicate=False):
        return RuntimeOptions(PropertiesFile.readFromSnippet(snippet, checkDuplicate))

    @staticmethod
    def readFromFile(file, checkDuplicate=False):
        return RuntimeOptions(PropertiesFile.readFromFile(file, checkDuplicate))



class SiteOptions(OptionsBase):
    """ Site options dictionary. """

    def __init__(self, filePath, options = {}):
        super(SiteOptions, self).__init__(options)
        self.__filePath = filePath

    def new(self, options):
         return SiteOptions(self.__filePath, options)

    def set(self, key, value):
        mergedOptions = self.dict()
        mergedOptions[key] = value
        return SiteOptions(self.__filePath, mergedOptions)

    def validatePasswordOptions(self):
        # artf119721 - REQUIRE_RANDOM_ADMIN_PASSWORD, when set to 'true', configures the admin account with a random password.
        # Before first login, the 'Forgot Your Password?' link in the site home page needs to be used to reset the password.
        # This necessitates the admin account to have been configured with a valid email id.
        if self.has_key('REQUIRE_RANDOM_ADMIN_PASSWORD') and self['REQUIRE_RANDOM_ADMIN_PASSWORD'] == 'true':
            if not self.has_key('ADMIN_EMAIL') or self.get('ADMIN_EMAIL') == "":
                raise Exception("REQUIRE_RANDOM_ADMIN_PASSWORD site option has been set to 'true'. This requires ADMIN_EMAIL to be specified as well.")

        # artf199417 - When PCK is turned on by enabling REQUIRED_PASSWORD_SECURITY but user forgets to add the effective date,
        # it takes the current date as the effective date and starts disabling/deleting all those who meet the PCK criteria
        # recreate runtime should fail, when PASSWORD_CONTROL_EFFECTIVE_DATE is not set/valid when REQUIRED_PASSWORD_SECURITY=true
        if self.has_key("REQUIRE_PASSWORD_SECURITY") and self["REQUIRE_PASSWORD_SECURITY"] == "true":
            # PASSWORD_CONTROL_EFFECTIVE_DATE should not be empty/null. It should be a valid date
            if not self.has_key("PASSWORD_CONTROL_EFFECTIVE_DATE") or self.get("PASSWORD_CONTROL_EFFECTIVE_DATE") == "":
                raise Exception("Invalid PASSWORD_CONTROL_EFFECTIVE_DATE found.\nIt should be set to a valid date (30 days from today, for eg) in MM/dd/yyyy format when REQUIRE_PASSWORD_SECURITY is set to true. Please refer documentation for more info.\n")

        return self

    def validateTokens(self, tokenDef, allDeployedServiceNames):
        tokenDef.setComponentList(allDeployedServiceNames)
        tokenDef.validateConfig(self.dict())
        return self

    def validate(self, tokenDef, allDeployedServices):
        return self.validateTokens(tokenDef, allDeployedServices) \
            .validatePasswordOptions()

    def deriveRuntimeOptions(self, tokenDef):
        return RuntimeOptions(self.dict())

    @staticmethod
    def readFromFilePath(filePath, checkDuplicate=False):
        return SiteOptions(filePath, PropertiesFile.readFromFilePath(filePath, checkDuplicate))

class Obfuscator(object):
    def __init__(self, obfuscationPrefix, obfuscationKey, tokenDef):
        self.__obfuscationPrefix = obfuscationPrefix
        self.__obfuscationKey = obfuscationKey
        self.__tokenDef = tokenDef

    def __needsObfuscation(self, key, value):
        return len(value) > 0 and self.__tokenDef.getProperty(key, 'obfuscate') == "true"

    def __isObfuscated(self, value):
        return value.startswith(self.__obfuscationPrefix)

    def __invokeObfuscatorUtil(self, verb, str):
        jreDirectory = app.defaults.DEFAULT_JAVA_HOME
        if not os.path.exists(jreDirectory):
            # This should only happen on CI machines when running unit tests.
            jreDirectory = "/usr/java/latest"

        command = [
            os.path.join(jreDirectory, 'bin/java'),
            '-cp', app.defaults.CTFUTIL_CLASSPATH,
            'com.vasoftware.sf.system.utils.TeamforgeObfuscator',
           verb,
           str ]

        env = os.environ.copy()
        env['OBFUSCATION_KEY'] = self.__obfuscationKey
        env['OBFUSCATION_PREFIX'] = self.__obfuscationPrefix

        return app.util.executeCommand(command, env, False).strip('\n')

    def obfuscate(self, value):
        return self.__invokeObfuscatorUtil('-encrypt', value)

    def deobfuscate(self, value):
        return self.__invokeObfuscatorUtil('-decrypt', value)

    def obfuscateToken(self, key, value):
        if not self.__needsObfuscation(key, value):
            return value
        else:
             return self.__invokeObfuscatorUtil('-encrypt', value)

    def deobfuscateToken(self, key, value):
        if value.startswith(self.__obfuscationPrefix):
            return self.__invokeObfuscatorUtil('-decrypt', value)
        else:
            return value

    @staticmethod
    def createFromRuntimeOptions(runtimeOptions):
        return app.options.Obfuscator(
            runtimeOptions["OBFUSCATION_PREFIX"],
            runtimeOptions["OBFUSCATION_KEY"],
            app.tokens.defines.TokenDef())

    def __createCommand(self, verb, filePath=None):
        jreDirectory = app.defaults.DEFAULT_JAVA_HOME
        if not os.path.exists(jreDirectory):
            # This should only happen on CI machines when running unit tests.
            jreDirectory = "/usr/java/latest"

        command = [
            os.path.join(jreDirectory, 'bin/java'),
            '-cp', app.defaults.CTFUTIL_CLASSPATH,
            'com.collabnet.obfuscator.ObfuscatorUtil',
            verb,
            self.__obfuscationKey,
            self.__obfuscationPrefix ]

        if filePath:
            command += [filePath]

        return command

    def deobfuscateOptions(self, options):
        assert isinstance(options, RuntimeOptions)

        # We could pipe everything thru the tool -- however, this
        # will not work for multi-line options. So we need to split
        # the input into two buckets - one that needs to be piped thru
        # the obfscation tool and one that can be used verbatim.

        obfuscatedOptions = {}
        nonobfuscatedOptions = {}
        for key, value in options.iteritems():
            if self.__isObfuscated(value):
                obfuscatedOptions[key] = value
            else:
                nonobfuscatedOptions[key] = value

        # Pipe obfuscated options thru the tool.
        process = subprocess.Popen(
            self.__createCommand("-deobfuscate"),
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=None,
            bufsize=0,
            shell=False)

        RuntimeOptions(obfuscatedOptions).writeToFile(process.stdin)
        process.stdin.close()
        exitCode = process.wait()
        deobfuscatedOptions = RuntimeOptions().readFromFile(process.stdout)

        if exitCode != 0:
            raise Exception("ObfuscatorUtil failed with exit code %d" % exitCode)


        return deobfuscatedOptions.overlayBy(RuntimeOptions(nonobfuscatedOptions))

    def obfuscateOptions(self, options):
        assert isinstance(options, RuntimeOptions)

        # Split the options into two buckets - one that needs to be piped thru
        # the obfscation tool and one that can be used verbatim.

        obfuscatableOptions = {}
        nonobfuscatableOptions = {}
        for key, value in options.iteritems():
            if self.__needsObfuscation(key, value):
                obfuscatableOptions[key] = value
            else:
                nonobfuscatableOptions[key] = value

        # Pipe obfuscatable options thru the tool.
        process = subprocess.Popen(
            self.__createCommand("-obfuscate"),
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=None,
            bufsize=0,
            shell=False)

        RuntimeOptions(obfuscatableOptions).writeToFile(process.stdin)
        process.stdin.close()
        exitCode = process.wait()
        obfuscatedOptions = RuntimeOptions().readFromFile(process.stdout)

        if exitCode != 0:
            raise Exception("ObfuscatorUtil failed with exit code %d" % exitCode)

        return obfuscatedOptions.overlayBy(RuntimeOptions(nonobfuscatableOptions))

class PropertiesFile(object):
    """ Helper class to read and write property files """

    @staticmethod
    def readFromFilePath(filePath, checkDuplicate=False):
        with open(filePath) as file:
            return PropertiesFile.readFromFile(file, checkDuplicate)

    @staticmethod
    def readFromSnippet(snippet, checkDuplicate=False):
        return PropertiesFile.readFromFile(StringIO.StringIO(snippet), checkDuplicate)

    @staticmethod
    def __unescape(str):
        # Property files (at least when written with Java) escape certain characters.
        return str \
            .replace("\:", ":") \
            .replace("\=", "=") \
            .replace("\!", "!") \
            .replace("\#", "#")

    @staticmethod
    def readFromFile(file, checkDuplicate=False):
        options = {}

        regex = re.compile("^\s*([/\w\.\-\:]+)\s*=\s*(.*)\s*$")
        # fileinput module is avoided for better performance
        lineno=0
        while 1:
            line = file.readline()
            lineno += 1
            if not line:
                break
            line = line.strip()
            if line == '' or line[0:1] == '#':
                continue
            match = regex.match(line.splitlines()[0])
            if match:
                key = match.group(1)
                value = match.group(2)
                if (checkDuplicate == True and options.has_key(key)):
                   raise Exception("Duplicate Token: Token '%s' is redefined in line no %s" %
                                (key, lineno))
		        # Support multi-line property value
                if value == 'START_MULTILINE_BLOCK':
                    value = ''
                    while 1:
                        line = file.readline()
                        lineno += 1
                        if not line:
                            break
                        if line.strip() == 'END_MULTILINE_BLOCK':
                            break
                        value += line

                options[PropertiesFile.__unescape(key)] = PropertiesFile.__unescape(value)

        return options

    @staticmethod
    def writeToFilePath( dict, filePath):
        with open(filePath, "w") as file:
            PropertiesFile.writeToFile(dict, file)

    @staticmethod
    def writeToFile(dict, file):
        for key, value in dict.items():
	        # Support multi-line property value
            if value.find('\n') == -1:
                file.write("%s=%s\n" % (key, value))
            else:
                file.write("%s=START_MULTILINE_BLOCK\n%sEND_MULTILINE_BLOCK\n" % (key, value))

    @staticmethod
    def updateFile(filePath, key, value, comment = 'Below token is auto generated, please do not edit'):
        # Read file
        lines = []
        with open(filePath, "r") as readFh:
            lines = readFh.read().split("\n")

        matcher = '^\s*(%s)\s*=\s*(.*)\s*$' % key
        found = False

        for lineNumber, line in enumerate(lines):
            mObj = re.search(r'%s'%matcher, line)
            if mObj is not None:
                # Replace the token
                lines[lineNumber] = '%s=%s' % (key, value)

                # Update the comment
                if lineNumber > 0 and lines[lineNumber - 1].startswith('##'):
                    lines[lineNumber - 1] = '## %s' % comment
                else:
                    lines.insert(lineNumber, '## %s' % comment)
                found = True
                break

        if not found:
            lines.append('')
            lines.append('## %s' % comment)
            lines.append('%s=%s' % (key, value))

        # Write
        with open(filePath, "w") as writeFh:
            writeFh.write("\n".join(lines))
