﻿#!/usr/bin/env python
# tokens.py -*- Python -*-

"""Contains classes which replace tokens."""

__author__ = 'Todd Enersen <tenersen@collab.net>'

import os
import re
import shutil
import stat
import sys
import app.options

class TokenProcessor:
    """For replacing tokens within a file. Tokens have the form
    {__*__}, where the braces and underscores are literal
    characters, not regular expressions."""

    def __init__(self, conf=None):
        self.conf = conf
        self.pattern = re.compile('{__(.*?)__}')

    def copyRecursively(self, source, dest):
        """ we'd use shutil.copytree, but it doesn't like to find dirs in the
        way.  this just does what that would do, making an effort to handle
        things as well as it can.  errors remain possible.
        """
        if not os.path.isdir(dest):
            os.mkdir(dest)

        for item in os.listdir(source):
            sourcePath = os.path.join(source, item)
            destPath = os.path.join(dest, item)
            if os.path.isdir(sourcePath):
                self.copyRecursively(sourcePath, destPath)
            elif not os.path.isfile(destPath):
                self.copyWithSubstitution(sourcePath, destPath)
            elif os.stat(sourcePath).st_mtime > os.stat(destPath).st_mtime:
                # delete the file first so as to avoid permissions errors
                os.remove(destPath)
                self.copyWithSubstitution(sourcePath, destPath)


    def copyWithSubstitution(self, sourceFilePath, destFilePath, overrideCfg=None):
        """Copy from dest to source with substitution of site option
        tokens.  Does not perform #! line transformation."""
        input = open(sourceFilePath)
        output = open(destFilePath, "w")
        self.streamWithSubstitution(input, output, overrideCfg);
        output.close()
        input.close()

    def streamWithSubstitution(self, input, output, overrideCfg=None):
        """go through the source file line by line looking for
           tokens.  Write the translated results to output"""
        offset = 0
        for line in input:
            offset += 1;
            try:
                line = self.substitute(line)
                if (overrideCfg is not None):
                   line = self.processOverride(line, overrideCfg)
            except Exception:
                sys.stderr.write("Unknown token in file '%s' on line %s : %d\n"
                                  % (input.name, line, offset))
            output.write(line)

    def processOverride(self, str, overrideCfg):

        procStr = str.rstrip()
        matcher = '^\s*([\w\.\-]+)\s*=\s*(.*)\s*$'
        mObj = re.search(r'%s'%matcher, procStr)
        if mObj is None:
           return str

        key = mObj.group(1)

        if (not overrideCfg.has_key(key)):
           return str
        return "%s=%s\n"%(key, overrideCfg[key])

    def substitute(self, str):
        def substituteMatch(matchobj):
            # matchobj should have the format {__NAME__} -- we want NAME
            name = matchobj.group(1)

            if name.startswith("OBFUSCATE:"):
                name = name[len("OBFUSCATE:"):]

                # If obfuscation is enabled, return the obfuscated value.
                # In all other cases, fall thru.
                if configObj.has_key(name):
                    return app.options.Obfuscator \
                        .createFromRuntimeOptions(configObj) \
                        .obfuscateToken(name, configObj.get(name))

            if configObj.has_key(name):
                return configObj.get(name)

            # Cannot be resolved.
            return matchobj.group(0)

        configObj = self.conf

        return self.pattern.sub(substituteMatch, str)

def copyScriptWithTransform(allOptions, srcPath, destPath):
    """Copies a script with the minimal #! line transformation
    (described by runtime-directory.html).  Does not perform site
    option token substitution."""

    destFile = open(destPath, 'w')
    srcFile = open(srcPath)

    header = "## #! automatically transformed for use in CollabNet.  " + \
             "The original line was:\n" + \
             "## "

    copyEntireFile = 0
    processFirstLine = 1
    processSecondLine = 1
    emptyLinesToRemove = 0
    tmpFirstLine = ""
    linesProcessed = 0

    # Note: this code does not process the shebangs of files with fewer than
    # two lines.  It should not be the case that you have a script with only
    # one line it - the shebang - so it's not a case we're checking for.  files
    # in a copy path that have fewer than 2 lines will be copied over without
    # transformation.
    for line in srcFile:
        linesProcessed = linesProcessed + 1
        if processFirstLine:
            tmpFirstLine = line
            processFirstLine = 0
            continue
        if processSecondLine:
            processSecondLine = 0
            # Swap the first and the second lines.  We're going to process
            # this pass as if we were processing the first line, while we try
            # to determine what to do about the second.
            secondLine = line
            line = tmpFirstLine

            execLine, extra = __transformExecLine(allOptions, line, srcPath)
            # Is this line a #! line which we can transform?
            if line != execLine:
                preHeader = ""
                postHeader = ""

                # python can declare the encoding for the file as the second
                # line using emacs (or similar) syntax.  Since we're processing
                # the first line, check if this is an encoding declaration, and
                # if so, preserve its position in the file.  See:
                # http://www.python.org/dev/peps/pep-0263
                if re.search("coding[:=]\s*([-\w.]+)", secondLine):
                    preHeader = secondLine
                else:
                    postHeader = secondLine
                line = (execLine + preHeader + header + line)
                if extra is not None:
                    line = line + extra
                # changes to the last block must be reflected in this
                # calculation -- the total number of lines to be printed,
                # minus one for the initial line (the old #! line), plus
                # two (currently) for the lines below.  the split function
                # gives one extra element with an empty string in it, so it's
                # the current number of lines + 1 -- which is correct.
                emptyLinesToRemove = len(line.split("\n"))
                line = (line +
                        "## Add %d lines of blank or empty comments to "
                             "maintain line numbers below this\n"
                        "## -- END MODIFIED SECTION --\n" % emptyLinesToRemove +
                        postHeader)
            else:
                # Copy the file wholesale instead of line by line.
                copyEntireFile = 1
                break
        elif emptyLinesToRemove > 0:
            if re.match("^#* *$", line):
	        emptyLinesToRemove = emptyLinesToRemove - 1
		line = ""
	    else:
	        # found something other than any empty comment; no
	        # more modifications
	        emptyLinesToRemove = 0

        destFile.write(line)
    # The file had fewer than 2 lines, so never got processed - copy the whole
    # file over unmolested.
    if linesProcessed < 2:
        copyEntireFile = 1
    if copyEntireFile:
        srcFile.seek(0)
        shutil.copyfileobj(srcFile, destFile)

    # Set file permissions to same as source location
    mode = stat.S_IMODE(os.fstat(srcFile.fileno())[stat.ST_MODE] | 0111)
    os.chmod(destFile.name, mode)


def copyScriptsWithTransform(allOptions, srcDir, destDir):
    """Copies scripts with the minimal #! line transformation
    described by runtime-directory.html"""
    if not os.path.isdir(destDir):
        os.makedirs(destDir)

    for item in os.listdir(srcDir):
        srcPath = os.path.join(srcDir, item);
        destPath = os.path.join(destDir, item);

        if os.path.isdir(srcPath):
            copyScriptsWithTransform(allOptions, srcPath, destPath) # recurse
        elif os.path.isfile(srcPath):
            copyScriptWithTransform(allOptions, srcPath, destPath)

        # if it's not a directory or a path, it's something we don't want
        # to touch

def __transformExecLine(allOptions, line, srcPath = ""):
    extra = None
    if line.startswith('#!'):
        # First we look for a second argument to store -- we may want to add
        # this later.
        additional = None
        list = line.split()
        if len(list) > 1:
            additional = list[1]
        # Many of our scripts have /usr/bin/env {interpreter name}.
        # In that case, we don't want the additional argument.
        if list[0].endswith('/env'):
            additional = None

        if re.search('python', line, re.IGNORECASE):
            line = "#!%s\n" % allOptions['PYTHON_WRAPPER']
        elif re.search('/bin/(ba)?sh', line) or re.search('SH_WRAPPER', line):
            line = "#!%s\n" % allOptions['SH_WRAPPER']
        elif re.search('perl', line, re.IGNORECASE):
            line = "#!%s\n" % allOptions['PERL_WRAPPER']
            extra = "%s\n" % allOptions['PERL_WRAPPER_EXTRA_LINE']

        # Consider adding the additional argument (if any).
        if additional:
            # If the additional argument pushes us over the length
            # limit, warn the user and avoid appending a truncated
            # argument.
            if len(line) + len(additional) + 1 > 128:
                sys.stderr.write("RUNTIME_DIR is a long path -- with args, "
                                 "some scripts' #! lines excede 128 bytes "
                                 "-- this can cause problems on some OSes. "
                                 "Arg %s could be not included for: %s\n"
                                 % (additional, srcPath))
                additional = None
            else:
                # strip the trailing newline and append a space,
                # the add'l argument, and the newline:
                line = line[0:-1] + " " + additional + "\n"
    return line, extra
