﻿
import app.util
import sys
import time
import app.secureconfig
import app.logutil

"""
The module that defines all base classes used for python hook-scripts execution
"""

class BaseScript:

    """The interface defining all CollabNet scripts."""

    scriptDescription = "Unnamed script"

    def __init__(self, log, cfg, version = None):
        self.cfg = cfg
        self.log = log
        self.version = version

    def logMesg(self, msgType, msgData, lineFeed=True):
        app.logutil.printLog(self.log, msgType, msgData, lineFeed)

    def run(self):
        """
        Main method for invoking the script.
        Should be overridden in the sub classes
        """
        raise AttributeError('Method not implemented')

class BaseOptionScript(BaseScript):
    """
    The base class for all options hook scripts.
    """
    def optionSnippet(self):
        """
        Main method which corresponds to the running the actual python option
        snippet
        """
        raise AttributeError('Method not implemented.')

    def run(self):
        return self.optionSnippet()



class BaseCreateRuntimeScript(BaseScript):
    """
    The base class for all create runtime hook scripts.
    """
    def createRuntime(self):
        """
        Main method which corresponds to the running the actual python option
        snippet
        """
        raise AttributeError('Method not implemented.')

    def run(self):

        self.logMesg('info', "  %s" % self.scriptDescription.ljust(58), False)

        try:
           sTime = time.time()
           self.createRuntime()
           elapsedTime = time.time() - sTime
           elapsedTime = "{0:.2f}".format(elapsedTime)
           self.logMesg('success', 'Done'.ljust(8), False)
           self.logMesg('info', '%s secs'%elapsedTime.rjust(6))
           return 0
        except Exception, msg:
           self.logMesg('error', 'Failed')
           self.logMesg('error', str(msg))
           self.log.error( str(msg) + ': ' + ''.join(app.util.getStackTraceAndException()).strip())
           raise msg

class BaseConfigScript(BaseScript):
    """
    The base class for all options hook scripts.
    """

    def configSnippet(self):
        """
        Main method which corresponds to the running the actual python option
        snippet
        """
        raise AttributeError('Method not implemented.')

    def run(self):
        return self.configSnippet()

class BaseBootstrapScript(BaseScript):
    """
    The base class for all bootstrap hook scripts.
    """
    def bootstrap(self):
        """
        Main method which corresponds to the running the actual python bootstrap
        snippet
        """
        raise AttributeError('Method not implemented.')

    def run(self):

        self.logMesg('info', "  %s"%self.scriptDescription.ljust(58), False)

        try:
           sTime = time.time()
           self.bootstrap()
           elapsedTime = time.time() - sTime
           elapsedTime = "{0:.2f}".format(elapsedTime)
           self.logMesg('success', 'Done'.ljust(8), False)
           self.logMesg('info', '%s secs'%elapsedTime.rjust(6))
           return 0
        except Exception, msg:
           self.logMesg('error', 'Failed')
           self.logMesg('error', str(msg))
           self.log.error( str(msg) + ': ' + ''.join(app.util.getStackTraceAndException()).strip())
           raise msg

class BaseMigrationScript(BaseScript):
    """
    The base class for all migration scripts. Each migration script should extend
    this class.
    """
    def migrate(self):
        """
        Main method which corresponds to the running the actual migration activity
        """
        raise AttributeError('Method not implemented. No migration code')

    def preMigrate(self):
        return True

    def postMigrate(self):
        pass

    def run(self):
        self.logMesg('info', "  %s"%self.scriptDescription.ljust(58), False)

        try:
            sTime = time.time()

            if self.preMigrate():
                self.migrate()
                self.postMigrate()
                self.logMesg('success', 'Done'.ljust(8), False)
            else:
                self.logMesg('success', 'Ignored'.ljust(8), False)

            elapsedTime = time.time() - sTime
            elapsedTime = "{0:.2f}".format(elapsedTime)
            self.logMesg('info', '%s secs'%elapsedTime.rjust(6))

            return 0
        except Exception, msg:
            self.logMesg('error', 'Failed')
            self.logMesg('error', str(msg))
            self.log.error( str(msg) + ': ' + ''.join(app.util.getStackTraceAndException()).strip())
            sys.exit(1)

class BaseCronScript(BaseScript):
    """
    The base class for all cron job scripts. Each cron job script should extend
    this class.
    """
    def cronjob(self):
        """
        Main method which corresponds to the running the actual cron job activity
        """
        raise AttributeError('Method not implemented. No cron job code')

    def run(self):
        self.logMesg('info', "  %s"%self.scriptDescription.ljust(58), False)

        try:
            sTime = time.time()

            self.cronjob()
            self.logMesg('success', 'Done'.ljust(8), False)

            elapsedTime = time.time() - sTime
            elapsedTime = "{0:.2f}".format(elapsedTime)
            self.logMesg('info', '%s secs'%elapsedTime.rjust(6))

            return 0
        except Exception, msg:
            self.logMesg('error', 'Failed')
            self.logMesg('error', str(msg))
            self.log.error( str(msg) + ': ' + ''.join(app.util.getStackTraceAndException()).strip())


class BaseVerificationScript(BaseScript):
    """
    The base class for all verification hook scripts.
    """
    def verificationSnippet(self):
        """
        Main method which corresponds to the running the actual python verification
        snippet
        """
        raise AttributeError('Method not implemented.')

    def run(self):

        self.logMesg('info', "  %s"%self.scriptDescription.ljust(58), False)

        try:
           sTime = time.time()
           self.verificationSnippet()
           elapsedTime = time.time() - sTime
           elapsedTime = "{0:.2f}".format(elapsedTime)
           self.logMesg('success', 'Done'.ljust(8), False)
           self.logMesg('info', '%s secs'%elapsedTime.rjust(6))
           return 0
        except Exception, msg:
           self.logMesg('error', 'Failed')
           self.logMesg('error', str(msg))
           self.log.error( str(msg) + ': ' + ''.join(app.util.getStackTraceAndException()).strip())
           raise msg

class BaseInitializationScript(BaseScript):
    """
    The base class for all initialization hook scripts.
    """
    def initialize(self):
        """
        Main method which corresponds to the running the actual python initialization
        snippet
        """
        raise AttributeError('Method not implemented.')

    def run(self):

        self.logMesg('info', "  %s"%self.scriptDescription.ljust(58), False)

        try:
           sTime = time.time()
           self.initialize()
           elapsedTime = time.time() - sTime
           elapsedTime = "{0:.2f}".format(elapsedTime)
           self.logMesg('success', 'Done'.ljust(8), False)
           self.logMesg('info', '%s secs'%elapsedTime.rjust(6))
           return 0
        except Exception, msg:
           self.logMesg('error', 'Failed: %s' % str(msg) )
           self.log.error(str(msg) + ': ' + ''.join(app.util.getStackTraceAndException()).strip())
           raise msg

class BaseFolderScript(BaseScript):
    """
    The base class for all folder hook scripts.
    """
    def deleteFolders(self):
        raise AttributeError('Method not implemented.')

    def createFolders(self):
        raise AttributeError('Method not implemented.')

    def applyFolderPermissions(self):
        pass

    def arePermissionsApplied(self):
        return True

    def areSELinuxPermissionsApplied(self):
        return True

    def checkAndApplyFolderPermissions(self):
        # There may be chance that a folder hook function may return NULL even in success case.So checking only
        # in case of False will call applyFolderPermissions function
        if self.arePermissionsApplied() == False:
           self.applyFolderPermissions()

    def applySELinuxPermissions(self):
        pass

    def checkAndApplySELinuxPermissions(self):
        if self.areSELinuxPermissionsApplied() == False:
            self.applySELinuxPermissions()

