import os
import app.util
import app.colors
import commands
import re
import selinux
import app.logutil

class File(object):
    @staticmethod
    def __run(command):
        exitCode = os.system(command)
        if exitCode != 0:
            raise SelinuxException("Command '%s' failed with exit code %d" % (command, exitCode))

    @staticmethod
    def __addTypeAndRestore(path, type, recursive):
        # It is always better to have absolute paths in
        # /etc/selinux/targeted/contexts/files/file_contexts.local
        path = os.path.abspath(path)

        # Clear recurive path if this is not a directory
        recursive = recursive and os.path.isdir(path)

        # Add file context.
        if type:
            # semanage command will work only when selinux mode either permissive or Enforce mode
            if File.isSELinuxEnabled():
                fileContext = File.getFileContext(path)
                if fileContext != 'nfs_t' and fileContext != type:
                    Policy.deleteLocalContextByPath(path)
                File.__run("semanage fcontext -a -t %s '%s%s'" % (
                    type,
                    path,
                    "(/.*)?" if recursive else ""))
            else:
                print app.colors.Color.amber("  SElinux is disabled, so SELinux context will not be applied for path %s" % path)

        # Run restorecon to have above change take effect.
        File.__run("/sbin/restorecon %s %s" % (
            "-R" if recursive else "",
            path))

    @staticmethod
    def getFileContext(path):
        contextOutput = selinux.getfilecon(path)
        if len(contextOutput) < 2:
            return None

        contextArr = contextOutput[1].split(":")
        if len(contextArr) < 3:
            return None

        return contextArr[2]

    @staticmethod
    def isFileContextMatchWithPolicy(path):
        cmd = "/usr/sbin/matchpathcon -q -V %s" % (path)
        if os.system(cmd):
            return False

        return True

    @staticmethod
    def isSELinuxEnabled():
        selinuxMode = app.util.executeCommand("/usr/sbin/getenforce")
        if selinuxMode.strip() == "Disabled":
            return False
        return True

    @staticmethod
    def setContext(path, type=None, recursive=True):
        if os.path.islink(path):
            # The symlink must be treated separately from the target
            File.__addTypeAndRestore(path, type, False)
            originalPath = os.readlink(path)
            if os.path.islink(originalPath):
                File.__addTypeAndRestore(originalPath, type, False)
            File.__addTypeAndRestore(os.path.realpath(path), type, recursive)
        else:
            File.__addTypeAndRestore(path, type, recursive)

class Policy(object):
    @staticmethod
    def __run(command):
        exitCode = os.system(command)
        if exitCode != 0:
            raise SelinuxException("Command '%s' failed with exit code %d" % (command, exitCode))
    @staticmethod
    def load(policyFile):
        cmd = "semodule -i %s" % (policyFile)
        Policy.__run(cmd)

    @staticmethod
    def unloadPolicy(moduleName):
        if File.isSELinuxEnabled():
            cmd = "semodule -r %s >/dev/null" % ( moduleName)
            Policy.__run(cmd)

    @staticmethod
    def listPolicy(policyName=None):
        modules = []

        if not File.isSELinuxEnabled():
            return modules

        cmd = "semodule -l"
        (retVal, output) = commands.getstatusoutput(cmd)
        if retVal:
            return modules
        for moduleWithVer in output.split("\n"):
            moduleName = moduleWithVer.split()[0]
            if policyName:
                if re.search(policyName, moduleName):
                    modules.append(moduleName)
            else:
                modules.append(moduleName)

        return modules

    @staticmethod
    def isSELinuxPermissionsApplied(path, type=None):
        if not File.isSELinuxEnabled() or not os.path.exists(path):
            return True
        fileContext = File.getFileContext(path)
        #If the path resides in NFS then SELinux context for the path would be an 'nfs_t' and return true always
        if fileContext == "nfs_t":
            return True
        if type:
            return type == fileContext
        else:
            return File.isFileContextMatchWithPolicy(path)

    @staticmethod
    def __setBoolValue(boolKey, state="on"):

        if not File.isSELinuxEnabled():
            return
        cmd = "setsebool %s %s" % (boolKey, state)
        Policy.__run(cmd)

    @staticmethod
    def enableNFSHomeDir():
        Policy.__setBoolValue("use_nfs_home_dirs")

    @staticmethod
    def __applyServicePolicy(logger, policyFile, osVersion):

        # fileName extension should be .pp
        if not re.search('\.pp$', policyFile):
            logger.error("Not valid policy file %s" % policyFile)
            return

        # By default it will load the policy file which are common for all
        # os. The sample policy file name is "tf_subversion.pp"
        # If the policy file name contain os distribution then appropriate
        # distribution policy file would be loaded.The sample os distribution policy file
        # would be <policyfile_rhel(6|7).pp

        if re.search('rhel[0-9]\.pp$', policyFile) and not re.search(('%s\.pp$' % osVersion), policyFile):
            return
        if os.path.isfile(policyFile):
            logger.info("Loading policy %s " % policyFile)
            Policy.load(policyFile)

    @staticmethod
    def deployServicePolicy(logger, service):

        if not app.selinuxutil.File.isSELinuxEnabled():
            logger.info("Skipping installation of SELinux policies as SELinux is disabled")
            return

        policyFolder = os.path.join(
            app.defaults.DEFAULT_SERVICE_DIR,
            service.getName(),
            "hooks",
            "policy")

        if not os.path.isdir(policyFolder):
            return

        if "redhat-6" in app.util.getPlatform():
            rhelVersion="rhel6"
        else:
            rhelVersion="rhel7"

        for fileOrFolder in sorted(os.listdir(policyFolder)):
            policyFileOrFolder = os.path.join(policyFolder, fileOrFolder)
            if os.path.isfile(policyFileOrFolder):
                Policy.__applyServicePolicy(logger, policyFileOrFolder, rhelVersion)
            elif os.path.isdir(policyFileOrFolder):
                for fileName in os.listdir(policyFileOrFolder):
                    policyFile=os.path.join(policyFileOrFolder, fileName)
                    if os.path.isfile(policyFile):
                        Policy.__applyServicePolicy(logger, policyFile, rhelVersion)

    @staticmethod
    def listLocalFileContext():
        localContextMappingList = []
        # List only local context which are added by semanage command
        (retVal, output) = commands.getstatusoutput('semanage fcontext -C -l -n')
        if retVal:
            return None
        for fileContextMapping in output.split("\n"):
            localContextMappingList.append(fileContextMapping.split())

        return localContextMappingList

    @staticmethod
    def getSELinuxModuleName(policyFile, moduleName, loadedModules, policyOSExten):
        #Custom module name. In redhat6 policy filename and module name may be different.
        if moduleName in loadedModules:
            return moduleName
        #In redhat7 the loaded module name and policy file name are same.
        moduleName = re.sub("\.pp$", "", os.path.basename(policyFile))
        if moduleName in loadedModules:
            return moduleName
        #In redhat6 module name and policy file name are same but policy file name may have os extension like
        #<tf_policyfile_rhel7.pp>
        moduleName = re.sub(("_%s.pp$" % policyOSExten), '', os.path.basename(policyFile))
        if moduleName in loadedModules:
            return moduleName

        return None

    @staticmethod
    def deleteLocalContextByPath(path):
        localContextList = Policy.listLocalFileContext()
        for fileContextMappingList in localContextList:
            if len(fileContextMappingList) < 3:
                return
            fileContextPath = fileContextMappingList[0]
            regExp = "^%s" % path
            if re.search(regExp, fileContextPath):
                Policy.__run('semanage fcontext -d "%s"' % fileContextPath)
                break;

    @staticmethod
    def deleteLocalContext(contexts, localContextList):
        for fileContextMappingList in localContextList:
            if len(fileContextMappingList) < 3 :
                return
            fileContext = fileContextMappingList[-1]
            for context in contexts:
                regExp = ":%s:" % context
                if re.search(regExp, fileContext):
                    path = fileContextMappingList[0]
                    Policy.__run('semanage fcontext -d "%s"' % path)
                    localContextList.remove(fileContextMappingList)
                    break

    @staticmethod
    def unloadSelinuxPolicy(logger, services):
        if not app.selinuxutil.File.isSELinuxEnabled():
            return
        policyOSExten = "rhel7"
        localContextsList = Policy.listLocalFileContext()
        selinuxModules =  Policy.listPolicy()
        if "redhat-6" in app.util.getPlatform():
            policyOSExten = "rhel6"
        for service in services:
            policyFolder = os.path.join(
                app.defaults.DEFAULT_SERVICE_DIR,
                service.getName(),
                "hooks",
                "policy")
            if not os.path.isdir(policyFolder):
                continue
            localContexts = service.getController(logger).getLocalContext()
            if localContexts:
                Policy.deleteLocalContext(localContexts, localContextsList)
            for fileOrFolder in reversed(sorted(os.listdir(policyFolder))):
                policyFileOrFolder = os.path.join(policyFolder, fileOrFolder)
                if os.path.isfile(os.path.join(policyFileOrFolder)):
                    # Some policy file has OS specfic and this will discord other OS policy files.
                    if re.search('_rhel[0-9]\.pp$', policyFileOrFolder) and not \
                            re.search(('_%s\.pp$' % policyOSExten), policyFileOrFolder):
                        continue
                    moduleName = Policy.getSELinuxModuleName(policyFileOrFolder,
                                                             service.getController(logger).getCustomModuleName(),
                                                             selinuxModules,
                                                             policyOSExten)
                    if not moduleName:
                        continue
                    app.logutil.printLog(logger, "info", "Unloading %s SELinux policy" % service.getName())
                    Policy.unloadPolicy(moduleName)
                    selinuxModules.remove(moduleName)
                elif os.path.isdir(policyFileOrFolder):
                    for fileName in os.listdir(policyFileOrFolder):
                        policyFile = os.path.join(policyFileOrFolder, fileName)
                        if os.path.isfile(policyFile):
                            # Some policy file has OS specfic and this will discord other OS policy files.
                            if re.search('_rhel[0-9]\.pp$', policyFileOrFolder) and not \
                                    re.search(('_%s\.pp$' % policyOSExten), policyFileOrFolder):
                                continue
                            moduleName = Policy.getSELinuxModuleName(policyFile,
                                                                     service.getController(logger).getCustomModuleName(),
                                                                     selinuxModules,
                                                                     policyOSExten)
                            if not moduleName:
                                continue
                            app.logutil.printLog(logger, "info", "Unloading %s SELinux policy" % service.getName())
                            Policy.unloadPolicy(moduleName)
                            selinuxModules.remove(moduleName)
    @staticmethod
    def applySELinuxPermissions(path, context=None, recursive=True):
        if not File.isSELinuxEnabled() or not os.path.exists(path):
            return
        print "Applying SELinux permissions to %s" % path
        return File.setContext(path, context, recursive)

class SelinuxException(Exception):
    pass
