﻿import sys
import os
import new
import signal
import subprocess
import time

import app.landscape
import app.logutil

class ServiceMetaStatus(object):
    OK = 0
    INCONCLUSIVE = 1
    BAD = 2

class ServiceStatus(object):
    def __init__(self, name, metastatus):
        self.__name = name
        self.__metastatus = metastatus

    def __repr__(self):
        return self.getName()

    def getName(self):
        return self.__name

    def getMetastatus(self):
        return self.__metastatus

ServiceStatus.ENV_VERIFICATION_FAILED =    ServiceStatus("environment verification failed", ServiceMetaStatus.BAD)
ServiceStatus.NOT_DEPLOYED =               ServiceStatus("not deployed", ServiceMetaStatus.BAD)
ServiceStatus.NOT_BOOTSTRAPPED =           ServiceStatus("not bootstrapped", ServiceMetaStatus.BAD)
ServiceStatus.NOT_MIGRATED =               ServiceStatus("not migrated", ServiceMetaStatus.BAD)
ServiceStatus.AVAILABLE =                  ServiceStatus("available", ServiceMetaStatus.OK)
ServiceStatus.UNHEALTHY =                  ServiceStatus("unhealthy", ServiceMetaStatus.BAD)
ServiceStatus.DEAD =                       ServiceStatus("dead", ServiceMetaStatus.BAD)
ServiceStatus.STOPPED_AUTOSTART =          ServiceStatus("stopped", ServiceMetaStatus.BAD)
ServiceStatus.STOPPED_DEMANDSTART =        ServiceStatus("inactive", ServiceMetaStatus.OK)
ServiceStatus.STARTING =                   ServiceStatus("starting", ServiceMetaStatus.INCONCLUSIVE)
ServiceStatus.STOPPING =                   ServiceStatus("stopping", ServiceMetaStatus.INCONCLUSIVE)
ServiceStatus.PORTS_BLOCKED =              ServiceStatus("ports blocked", ServiceMetaStatus.BAD)
ServiceStatus.DEPENDENCIES_UNAVAILABLE =   ServiceStatus("dependencies unavailable", ServiceMetaStatus.BAD)
ServiceStatus.DOOMED =                     ServiceStatus("doomed", ServiceMetaStatus.BAD)

class ServiceControlGlobalFlags(object):
    # Verify environment as part of checking the service status
    verificationEnabled = True

    # Prompt for confirmation before doing anything dangerous
    confirmDangerousActions = True

    # Be extra verbose about everything
    debugMode = False

class ServiceController(object):
    def __init__(self, service, logger):
        self.__service = service
        self.logger = logger

    def getService(self):
        return self.__service

    def getName(self):
        return self.__service.getName()

    def __raiseStatusException(self, action):
        raise ServiceControllerException("Cannot %s service '%s' while in status '%s'" %
                                (action, self.__service.getName(), self.getStatus()))


    #
    # Status callbacks - can be overridden.
    #

    def isDeployed(self):
        """ Check if runtime configuration is in place and data is up to date"""
        return DeploymentManager.isServiceMarkedAsDeployed(self.__service.getName())

    def getLocalContext(self):
        return []

    def getCustomModuleName(self):
        return None

    def isBootstrapped(self):
        """ Check if service data has been bootstrapped """
        return True

    def isMigrated(self):
        """ Check if service data has been migrated to latest schema """
        return True

    def printSnapshot(self, label):
        pass

    def isEnvironmentVerified(self):
        """ Check if environment passes verification """

        if not ServiceControlGlobalFlags.verificationEnabled:
            return True

        return self.createEnvironmentVerificationCheckAction(True).execute()

    def areRequiredEndpointsListening(self):
        """ Check if all required endpoints are listening """
        return self.createRequiredEndpointsListeningCheckAction(True).execute()

    def areProvidedEndpointsStillAvailable(self):
        """ Check if all required endpoints are listening """
        return self.createProvidedEndpointsStillAvailableCheckAction(True).execute()

    def areProvidedEndpointsListening(self):
        """ Check if all provided endpoints are listening """
        return self.createProvidedEndpointsListeningCheckAction(True).execute()

    def areDeploymentDependenciesAvailable(self):
        """ Check if all underlying deployments are available (only relevant for passive services) """
        for dependency in self.__service.getDeploymentDependencies():
            if dependency.getController(self.logger).getStatus().getMetastatus() != ServiceMetaStatus.OK:
                return False

        return True

    def areDeploymentDependenciesDeployed(self):
        """ Check if all underlying deployments are deployed """
        for dependency in self.__service.getDeploymentDependencies():
            status = dependency.getController(self.logger).getStatus()
            if status in (ServiceStatus.ENV_VERIFICATION_FAILED,
                          ServiceStatus.NOT_DEPLOYED,
                          ServiceStatus.NOT_MIGRATED,
                          ServiceStatus.NOT_BOOTSTRAPPED):
                return False

        return True

    def getPidFile(self):
        """ Return path to PID file """
        return None

    def getPid(self):
        """ Return PID or None if not running """
        pidFile = self.getPidFile()

        if pidFile is not None and os.path.isfile(pidFile):
            with open(pidFile, 'r') as file:
                pid = file.read().strip()
                if len(pid):
                    return int(pid)

        return None

    def getKillSignal(self):
        """ Return approprite signal to terminate """
        return signal.SIGKILL

    def isSupposedToBeRunning(self):
        return self.getPid() is not None

    def isStarting(self):
        """ Check if service is currently starting up """
        return False

    def isStopping(self):
        """ Check if service is currently stopping """
        return False

    def isDoomed(self):
        """ Check if service is in a limbo state and should be killed """
        return False

    def isProcessRunning(self):
        pid = self.getPid()
        if not pid:
            return False
        else:
            # Check if there really is a process with this PID
            try:
                os.kill(pid, 0)
                return True
            except OSError:
                return False

    def isDemandStart(self):
        """ For a daemon service, determine if the service is supposed to be
            handled as demand-start service """
        return False

    def isHealthy(self):
        # This method should be overridden. By default assume that
        # if the provided ports are not available anymore (i.e.,
        # somebody is listening), that everything is fine.

        return self.areProvidedEndpointsListening()

    def isBadlyBehavedServiceWhichCannotStartBecausePostInstallHasNotRunYet(self):
        return False

    def getDataStatus(self):
        """ This method can be overriden to check both the bootstrapped
        and migrated status in one go """

        if not self.isBootstrapped():
            return ServiceStatus.NOT_BOOTSTRAPPED
        elif not self.isMigrated():
            return ServiceStatus.NOT_MIGRATED
        else:
            return ServiceStatus.AVAILABLE

    def getLogPaths(self):
        """ Return list of paths of relevant logs for this service. Paths may be globs. """
        return []

    def startDaemonAsync(self):
        """ Start the daemon in background """
        raise ServiceControllerException("Do not know how to start '%s'" % self.__service.getName())

    def stopDaemonAsync(self):
        """ Stop the daemon in background """
        raise ServiceControllerException("Do not know how to stop '%s'" % self.__service.getName())

    def killDaemon(self):
        pid = self.getPid()
        if pid:
            try:
                # The PID may actually refer to some wrapper script - therfore, kill
                # the entire process group.
                os.killpg(os.getpgid(pid), self.getKillSignal())
            except Exception as err:
                self.logger.error("Killing %d failed: %s" % (pid, str(err)))

                # If the kill failed, odds are the process is gone already. Still,
                # it makes sense to clean up any leftover PID files.

        # Remove PID file so that it does not interfere with the next
        # startup. As termination might take a few seconds, we do
        # multiple attempts.
        pidFile = self.getPidFile()
        if pidFile:
            for x in range(0, 6):
                if os.path.isfile(pidFile):
                    try:
                        time.sleep(2)
                        os.remove(pidFile)
                        self.logger.info("Deleted PID file %s" % (pidFile))

                        break
                    except Exception as err:
                        self.logger.error("Deleting PID file %s failed: %s" % (pidFile, str(err)))

    #
    # Status transition callbacks - can be overridden.
    #

    def canDeploy(self):
        return self.getStatus() in [
            ServiceStatus.NOT_DEPLOYED,
            ServiceStatus.NOT_MIGRATED,
            ServiceStatus.NOT_BOOTSTRAPPED,
            ServiceStatus.AVAILABLE,
            ServiceStatus.DEPENDENCIES_UNAVAILABLE,
            ServiceStatus.DEAD,
            ServiceStatus.STOPPED_AUTOSTART,
            ServiceStatus.STOPPED_DEMANDSTART]

    def canUndeploy(self):
        if self.__service.isDaemon():
            return self.getStatus() not in [
                ServiceStatus.STARTING,
                ServiceStatus.STOPPING,
                ServiceStatus.AVAILABLE,
                ServiceStatus.UNHEALTHY,
                ServiceStatus.DOOMED]
        else:
            return True

    def canBootstrap(self):
        if self.__service.isDaemon():
            return self.getStatus() in [
                ServiceStatus.NOT_BOOTSTRAPPED,
                ServiceStatus.NOT_MIGRATED,
                ServiceStatus.DEPENDENCIES_UNAVAILABLE,
                ServiceStatus.DEAD,
                ServiceStatus.STOPPED_AUTOSTART,
                ServiceStatus.STOPPED_DEMANDSTART,
                ServiceStatus.PORTS_BLOCKED]
        else:
            return self.getStatus() in [
                ServiceStatus.NOT_BOOTSTRAPPED,
                ServiceStatus.NOT_MIGRATED,
                ServiceStatus.AVAILABLE,
                ServiceStatus.DEPENDENCIES_UNAVAILABLE]

    def canMigrate(self):
        return self.getStatus() in [
            ServiceStatus.AVAILABLE,
            ServiceStatus.DEPENDENCIES_UNAVAILABLE,
            ServiceStatus.NOT_MIGRATED,
            ServiceStatus.DEAD,
            ServiceStatus.STOPPED_AUTOSTART,
            ServiceStatus.STOPPED_DEMANDSTART]

    def canPostInstall(self):
        return self.getStatus() in [ServiceStatus.AVAILABLE]

    def canStart(self):
        return self.getStatus() in [
            ServiceStatus.STOPPED_AUTOSTART,
            ServiceStatus.STOPPED_DEMANDSTART,
            ServiceStatus.DEAD,

            # Dependencies might come available when other services are started as well.
            ServiceStatus.DEPENDENCIES_UNAVAILABLE]

    def canStop(self):
        return self.getStatus() in [
            ServiceStatus.AVAILABLE,
            ServiceStatus.UNHEALTHY,
            ServiceStatus.STOPPED_AUTOSTART,
            ServiceStatus.DOOMED]

    def canReinitialize(self):
        return self.getStatus() in [ServiceStatus.AVAILABLE]

    def canApplySelinuxPolicy(self):
        return self.getStatus() not in [
            ServiceStatus.NOT_DEPLOYED,
            ServiceStatus.NOT_BOOTSTRAPPED]

    def canSnapshot(self):
        return self.getStatus() not in [ServiceStatus.NOT_DEPLOYED]
    #
    # Status determination and actions - should not be overridden.
    #

    def getStatus(self):
        if not self.isEnvironmentVerified():
            return ServiceStatus.ENV_VERIFICATION_FAILED
        elif not self.isDeployed():
            return ServiceStatus.NOT_DEPLOYED

        dataStatus = self.getDataStatus()
        if dataStatus != ServiceStatus.AVAILABLE:
            return dataStatus
        elif not self.__service.isDaemon():
            if self.areDeploymentDependenciesAvailable():
                return ServiceStatus.AVAILABLE
            else:
                return ServiceStatus.DEPENDENCIES_UNAVAILABLE
        elif self.isStarting():
            return ServiceStatus.STARTING
        elif self.isStopping():
            return ServiceStatus.STOPPING
        elif self.isSupposedToBeRunning():
            if self.isProcessRunning():
                if self.isDoomed():
                    return ServiceStatus.DOOMED
                elif self.isHealthy():
                    return ServiceStatus.AVAILABLE
                else:
                    return ServiceStatus.UNHEALTHY
            else:
                return ServiceStatus.DEAD
        elif not self.areProvidedEndpointsStillAvailable():
            return ServiceStatus.PORTS_BLOCKED
        elif not self.areDeploymentDependenciesDeployed():
            return ServiceStatus.DEPENDENCIES_UNAVAILABLE
        elif not self.areRequiredEndpointsListening():
            return ServiceStatus.DEPENDENCIES_UNAVAILABLE
        elif self.isDemandStart():
            return ServiceStatus.STOPPED_DEMANDSTART
        else:
            return ServiceStatus.STOPPED_AUTOSTART

    def createPrintStatusAction(self):
        import app.actions.machine
        return app.actions.machine.PrintStatusServiceAction([self.__service], self.logger)

    def createEnvironmentVerificationCheckAction(self, silent=False):
        import app.actions.service
        return app.actions.service.EnvironmentVerificationCheckAction(
            [self.__service],
            app.logutil.getNullLogger()
                if silent and not ServiceControlGlobalFlags.debugMode
                else self.logger)

    def createProvidedEndpointsStillAvailableCheckAction(self, silent=False):
        import app.actions.service
        return app.actions.service.ProvidedEndpointAvailableCheckServiceAction(
            [self.__service],
            app.logutil.getNullLogger()
                if silent and not ServiceControlGlobalFlags.debugMode
                else self.logger)

    def createProvidedEndpointsListeningCheckAction(self, silent=False):
        import app.actions.service
        return app.actions.service.ProvidedEndpointListeningCheckServiceAction(
            [self.__service],
            app.logutil.getNullLogger()
                if silent and not ServiceControlGlobalFlags.debugMode
                else self.logger)

    def createRequiredEndpointsListeningCheckAction(self, silent=False):
        import app.actions.service
        return app.actions.service.RequiredEndpointListeningCheckServiceAction(
            [self.__service],
            app.logutil.getNullLogger()
                if silent and not ServiceControlGlobalFlags.debugMode
                else self.logger)

    def createPostDeployAction(self):
        """ Return action (if any) to run after deployment """
        return None

    def createDeployAction(self):
        # The Undeploy action will take care of stopping services if
        # necessary - so we do not need to call canUndeploy here.

        import app.actions.deploy
        return app.actions.deploy.DeployServiceAction(
            [self.__service],
            self.logger)

    def createUndeployAction(self):
        # The Undeploy action will take care of stopping services if
        # necessary - so we do not need to call canUndeploy here.

        import app.actions.deploy
        return app.actions.deploy.UndeployServiceAction(
            [self.__service],
            self.logger)

    def createBootstrapAction(self):
        if not self.canBootstrap():
            self.__raiseStatusException("bootstrap")

        import app.actions.bootstrap
        return app.actions.bootstrap.BootstrapServiceAction(
            [self.__service],
            self.logger)

    def createMigrateAction(self):
        if not self.canMigrate():
            self.__raiseStatusException("migrate")

        import app.actions.migrate
        return app.actions.migrate.MigrateServiceAction(
            [self.__service],
            self.logger)

    def createStartAction(self):
        if not self.__service.isDaemon():
            raise ServiceControllerException("Service '%s' is not a daemon and cannot be started" %
                                             self.getName())

        if not self.canStart():
            self.__raiseStatusException("start")

        import app.actions.start
        return app.actions.start.StartServiceAction(
            [self.__service],
            self.logger)

    def createStopAction(self):
        if not self.__service.isDaemon():
            raise ServiceControllerException("Service '%s' is not a daemon and cannot be stopped" %
                                             self.getName())

        if not self.canStop():
            self.__raiseStatusException("stop")

        import app.actions.stop
        return app.actions.stop.StopServiceAction(
            [self.__service],
            self.logger)

    def createKillAction(self):
        if not self.__service.isDaemon():
            raise ServiceControllerException("Service '%s' is not a daemon and cannot be killed" %
                                             self.getName())

        if not self.canStop():
            self.__raiseStatusException("stop")

        import app.actions.kill
        return app.actions.kill.KillServiceAction(
            [self.__service],
            self.logger)

    def createReinitializeAction(self):
        if not self.__service.isDaemon():
            raise ServiceControllerException("Service '%s' is not a daemon and cannot be reinitialized" %
                                             self.getName())

        if not self.canReinitialize():
            self.__raiseStatusException("reinitialize")

        import app.actions.initialize
        return app.actions.initialize.InitializeServiceAction(
            [self.__service],
            self.logger)

    def createSnapshotAction(self):
        if not self.canSnapshot():
            self.__raiseStatusException("snapshot")

        import app.actions.machine
        return app.actions.machine.SnapshotAction(
            [self.__service], self.logger)

    def createAwaitStartupDependenciesAction(self):
        import app.actions.await
        return app.actions.await.AwaitStartupDependenciesAction(
            [self.__service],
            self.logger)

    def createApplyPermissionsAction(self):
        import app.actions.folders
        return app.actions.folders.ApplyPermissionsAction(
            [self.__service],
            self.logger)

    def createCheckDataIntegrityAction(self):

        import app.actions.checkdataintegrity
        return app.actions.checkdataintegrity.CheckDataIntegrityAction(
            [self.__service],
            self.logger)

    def createApplySelinuxPolicyAction(self):
        if not self.canApplySelinuxPolicy():
            self.__raiseStatusException("apply-selinux")

        import app.actions.folders
        return app.actions.folders.ApplySelinuxPolicyAction(
            [self.__service],
            self.logger)

    def createUnloadSelinuxPolicyAction(self):
        import app.actions.folders
        return app.actions.folders.UnloadSelinuxPolicyAction(
            [self.__service],
            self.logger)

    def createTailLogsAction(self):
        import app.actions.logs
        return app.actions.logs.TailLogsAction(
            [self.__service],
            self.logger)

    @staticmethod
    def loadServiceControllerFromFile(service, filePath, logger):
        # Make a python-safe module name
        moduleName = "__%s__tmp" % filter(str.isalnum, str(service.getName()))

        scriptContent = open(filePath).read()

        sys.modules[moduleName] = mod = new.module(moduleName)
        mod.__file__ = filePath
        try:
            exec scriptContent in mod.__dict__

            exec """
from %s import ServiceController
obj = ServiceController(service, logger)
__file__ = '%s'
obj.__file__ = '%s'
mod.__dict__['result'] = obj
""" % (moduleName, filePath, filePath)
            return mod.__dict__['result']
        except Exception as e:
            raise ServiceControllerException("Script %s failed: %s" % (filePath, str(e)))


    @staticmethod
    def forService(service, logger):
        libDir = os.path.join(app.defaults.DEFAULT_SERVICE_DIR,
            service.getName(),
            "lib")

        sharedLibDir = os.path.join(libDir, "shared")
        if os.path.isdir(sharedLibDir):
            # Add this dir to Python search path
            sys.path.append(sharedLibDir)

        overrideScript = os.path.join(libDir, "service.py")
        if os.path.isfile(overrideScript):
            return ServiceController.loadServiceControllerFromFile(service, overrideScript, logger)
        else:
            return ServiceController(service, logger)

class SimpleDaemonController(ServiceController):
    """
        Controller for managing a program as a daemon that does not provide any
        daemon capabilities on its own.
    """
    def __init__(self, service, logger):
        super(SimpleDaemonController, self).__init__(service, logger)

    def getCommand(self):
        raise ServiceControllerException("Method needs to be overridden")

    def getPidFile(self):
        return os.path.join(app.defaults.DEFAULT_RUN_DIR, "%s.pid" % self.getService().getName())

    def getServiceLogPath(self):
        logFolder = os.path.join(app.defaults.DEFAULT_LOG_DIR, self.getService().getName())
        if not os.path.exists(logFolder):
            # Lazily create folder structure (mind you, we are running as root).
            os.makedirs(logFolder)

            # Set ownership so that the daemon can actually access the log file.
            startupInfo = self.getService().getDaemonStartupInfo()
            app.util.runSystem("chown -R %s:%s %s" % (startupInfo.getUser(), startupInfo.getGroup(), logFolder), self.logger)

        return os.path.join(logFolder, "service.log")

    def startDaemonAsync(self):
        user = self.getService().getDaemonStartupInfo().getUser()

        self.logger.info("Launching '%s' as %s" % (self.getCommand(), user))

        # Prepare command to run the command as daemon:
        # * switch user to <daemon user> while preserving the environment,
        # * use nohup -c to ensure that the process keeps running when the current shell is terminated,
        # * run the actual program,
        # * remove the pid file after regular termination,
        # * capture the PID and save it to the pid file.
        daemonizedCommand = """su -m %s -c 'nohup sh -c "%s >> %s 2>&1 ; rm -f %s" >/dev/null 2>&1 & echo $! > %s'  """ % \
            (user, self.getCommand().replace('"', '\\"'), self.getServiceLogPath(), self.getPidFile(), self.getPidFile())

        # Note that the PID refers to a shell, which in turn has at least one child. When we kill, we want
        # to make sure that we not only kill the shell, but the all children as well. Therefore,
        # spawn a new process group.
        self.logger.info("Running %s" % daemonizedCommand)
        subprocess.Popen(daemonizedCommand, preexec_fn=os.setsid, shell=True)

        return True

    def stopDaemonAsync(self):
        # Kill process group.
        pid = self.getPid()
        if pid:
            try:
                 os.killpg(os.getpgid(self.getPid()), self.getKillSignal())
            except Exception as err:
                self.logger.error("Killing process grouo pf %d failed: %s" % (pid, str(err)))
                return False

        os.remove(self.getPidFile())
        return True

    def getLogPaths(self):
        return [self.getServiceLogPath()]

class MachineController(object):
    """ Controlls all services on the machine """

    def __init__(self, machine, logger):
        self.__machine = machine
        self.__logger = logger

    def __purgeAliases(self, services):
        # For actions, we should avoid considering any aliases because
        # it may result in redundant activity.
        return [s for s in services if not s.isAlias()]

    def createPrintStatusAction(self):
        import app.actions.machine
        return app.actions.machine.PrintStatusServiceAction(
            self.__purgeAliases(self.__machine.getServicesInAlphabeticOrder()),
            self.__logger)

    def createSnapshotAction(self):
        chainedAction = None
        for service in reversed(self.__purgeAliases(self.__machine.getServicesInDeploymentOrder())):
            action = service.getController(self.__logger).createSnapshotAction()
            chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    def createEnvironmentVerificationCheckAction(self):
        # Create action for each service (so that the status checks are run) and
        # then chain them all together.

        chainedAction = None
        for service in self.__purgeAliases(self.__machine.getServices()):
            action = service.getController(self.__logger).createEnvironmentVerificationCheckAction()
            chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    def createProvidedPortAvailabilityCheckAction(self):
        # Create action for each service (so that the status checks are run) and
        # then chain them all together.

        chainedAction = None
        for service in self.__purgeAliases(self.__machine.getServices()):
            action = service.getController(self.__logger).createProvidedPortAvailabilityCheckAction()
            chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    def createRequiredPortAvailabilityCheckAction(self):
        # Create action for each service (so that the status checks are run) and
        # then chain them all together.

        chainedAction = None
        for service in self.__purgeAliases(self.__machine.getServices()):
            action = service.getController(self.__logger).createRequiredPortAvailabilityCheckAction()
            chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    def createDeployAction(self):
        # Create action for each service (so that the status checks are run) and
        # then chain them all together.

        # Deployment needs to happen in deploy order.

        chainedAction = None
        for service in self.__purgeAliases(self.__machine.getServicesInDeploymentOrder()):
            action = service.getController(self.__logger).createDeployAction()
            chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    def createUndeployAction(self):
        # Create action for each service (so that the status checks are run) and
        # then chain them all together.

        # Undeployment needs to happen in reverse deploy order.

        chainedAction = None
        for service in reversed(self.__purgeAliases(self.__machine.getServicesInDeploymentOrder())):
            action = service.getController(self.__logger).createUndeployAction()
            chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    def createBootstrapAction(self, services=None):
        # Bootstrapping needs to happen in startup order. This is important for
        # those scripts which require other services to be running already
        # (ex: create-database scripts require the DB server to be up and
        # bootstrapped already).
        servicesToBootstrap = self.__machine.getServicesInStartOrder()

        if services:
            # Intersect lists
            servicesToBootstrap = [s for s in servicesToBootstrap if s in services]

        # Create action for each service (so that the status checks are run) and
        # then chain them all together.
        chainedAction = None
        for service in self.__purgeAliases(servicesToBootstrap):
            action = service.getController(self.__logger).createBootstrapAction()
            chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    def createMigrateAction(self, services=None):
        # Migration also needs to happen in startup order. This is important for
        # those scripts which require other services to be running already
        # (ex: scripts accessing the database will assume the database has been
        # migrated already).
        servicesToMigrate = self.__machine.getServicesInStartOrder()

        if services:
            # Intersect lists
            servicesToMigrate = [s for s in servicesToMigrate if s in services]


        # Old CTF releases did not maintain a dedicated schema version for the datamart.
        # To be compatible with these releases, ensure that the datamart is always upgraded
        # *before* the database is upgraded. See migratorator for details

        servicesToMigrate = self.__purgeAliases(servicesToMigrate)
        if any (service for service in servicesToMigrate if service.getName() == "ctfcore-datamart-admin-postgres"):
            servicesToMigrate = [svc for svc in servicesToMigrate if svc.getName() == "ctfcore-datamart-postgres-cohosted"
                                                                  or svc.getName() == "ctfcore-datamart-postgres"
                                                                  or svc.getName() == "ctfcore-datamart-admin-postgres"] + \
                                [svc for svc in servicesToMigrate if svc.getName() != "ctfcore-datamart-postgres"
                                                                  and svc.getName() != "ctfcore-datamart-postgres-cohosted"
                                                                  and svc.getName() != "ctfcore-datamart-admin-postgres"]


        # Create action for each service (so that the status checks are run) and
        # then chain them all together.

        chainedAction = None
        for service in servicesToMigrate:
            action = service.getController(self.__logger).createMigrateAction()
            chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    def createStartAction(self):
        # Create action for each service (so that the status checks are run) and
        # then chain them all together.

        # Naturaly, starting up services must be done in startup order.

        chainedAction = None
        for service in self.__purgeAliases(self.__machine.getServicesInStartOrder()):
            controller = service.getController(self.__logger)
            if service.isDaemon() and \
                not controller.isDemandStart() and \
                not controller.isProcessRunning() and \
                not controller.isBadlyBehavedServiceWhichCannotStartBecausePostInstallHasNotRunYet():

                action = controller.createStartAction()
                chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    def createStopAction(self, services=None):
        # Stopping services should happen in reverse startup order.
        servicesToStop = self.__machine.getServicesInStartOrder()

        if services:
            # Intersect lists
            servicesToStop = [s for s in servicesToStop if s in services]

        # Create action for each service (so that the status checks are run) and
        # then chain them all together.

        chainedAction = None
        for service in reversed(self.__purgeAliases(servicesToStop)):
            if service.isDaemon() and service.getController(self.__logger).isProcessRunning():
                action = service.getController(self.__logger).createStopAction()
                chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    def createKillAction(self, services=None):
        # The order in which services are killed does not really matter, we are using
        # force here anyway. But for consistency, use same order as for stopping.
        servicesToKill = self.__machine.getServicesInStartOrder()

        if services:
            # Intersect lists
            servicesToKill = [s for s in servicesToKill if s in services]

        # Create action for each service (so that the status checks are run) and
        # then chain them all together.

        chainedAction = None
        for service in reversed(self.__purgeAliases(servicesToKill)):
            if service.isDaemon() and service.getController(self.__logger).isProcessRunning():
                action = service.getController(self.__logger).createKillAction()
                chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    def createReinitializeAction(self, services=None):
        servicesToReinitialize = self.__purgeAliases(self.__machine.getServicesInStartOrder())
        chainedAction = None

        for service in servicesToReinitialize:
            if service.isDaemon() and service.getController(self.__logger).canReinitialize():
                action = service.getController(self.__logger).createReinitializeAction()
                chainedAction = chainedAction.chain(action) if chainedAction else action

        return  chainedAction

    def createApplyPermissionsAction(self):
        # Create action for each service (so that the status checks are run) and
        # then chain them all together.

        # Order does not really matter, but use deployment order for consistency's sake.
        chainedAction = None
        for service in reversed(self.__purgeAliases(self.__machine.getServicesInDeploymentOrder())):
            action = service.getController(self.__logger).createApplyPermissionsAction()
            chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    def createCheckDataIntegrityAction(self):

        chainedAction = None
        for service in self.__purgeAliases(self.__machine.getServices()):
            action = service.getController(self.__logger).createCheckDataIntegrityAction()
            chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    def createAwaitStartupDependenciesAction(self):
        chainedAction = None
        for service in self.__purgeAliases(self.__machine.getServices()):
            action = service.getController(self.__logger).createAwaitStartupDependenciesAction()
            chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    def createApplySelinuxPolicyAction(self):
        chainedAction = None
        for service in self.__purgeAliases(self.__machine.getServicesInDeploymentOrder()):
            action = service.getController(self.__logger).createApplySelinuxPolicyAction()
            chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    def createUnloadSelinuxPolicyAction(self):
        chainedAction = None
        for service in reversed(self.__purgeAliases(self.__machine.getServicesInDeploymentOrder())):
            action = service.getController(self.__logger).createUnloadSelinuxPolicyAction()
            chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    def createTailLogsAction(self):
        chainedAction = None
        for service in reversed(self.__purgeAliases(self.__machine.getServicesInDeploymentOrder())):
            action = service.getController(self.__logger).createTailLogsAction()
            chainedAction = chainedAction.chain(action) if chainedAction else action

        return chainedAction

    @staticmethod
    def forMachine(machine, logger):
        return MachineController(machine, logger)

class SiteController(object):
    """ Controlls all services on the machine """

    def __init__(self, site, logger):
        self.__site = site
        self.__logger = logger

    def createDumpEndpointsAction(self):
        import app.actions.site
        return app.actions.site.DumpEndpointsSiteAction(self.__site, self.__logger)

    def createDumpDeploymentDependenciesAction(self):
        import app.actions.site
        return app.actions.site.DumpDeploymentDependenciesSiteAction(self.__site, self.__logger)

    @staticmethod
    def forSite(site, logger):
        return SiteController(site, logger)

class ControllerAction(object):
    def __init__(self, logger):
        self.logger = logger

    def printLog(self, logType, logMsg, lineFeed=True):
        app.logutil.printLog(self.logger, logType, logMsg, lineFeed)

    def display(self, logMsg, lineFeed=True):
        sys.stdout.write(logMsg)
        if (lineFeed):
           sys.stdout.write("\n")
        sys.stdout.flush()

    def getInvolvedServices(self):
        raise Exception("needs to be overridden")

    def execute(self):
        raise Exception("needs to be overridden")

class DeploymentManager(object):
    """
        Helper managing the state of services. The state is maintained as marker files on disk.

        To maintain the old idea of "deleting the runtime folder wipes the slate clean", we
        store these markers in the runtime directory.
    """
    @staticmethod
    def getDeploymentFolder(serviceName, createIfNotExist=True):
        folder = os.path.join(
                app.defaults.DEFAULT_RUNTIME_DIR,
                "service",
                serviceName)
        if createIfNotExist and not os.path.isdir(folder):
            os.makedirs(folder)

        return folder

    @staticmethod
    def __getMarkerFilePath(serviceName, createIfNotExist=True):
        return os.path.join(
            DeploymentManager.getDeploymentFolder(serviceName, True),
            ".deployed")

    @staticmethod
    def markServiceAsDeployed(serviceName):
        open(DeploymentManager.__getMarkerFilePath(serviceName), 'a').close()

    @staticmethod
    def markServiceAsUndeployed(serviceName):
        file = DeploymentManager.__getMarkerFilePath(serviceName)
        if os.path.isfile(file):
            os.remove(file)

    @staticmethod
    def isServiceMarkedAsDeployed(serviceName):
        if not os.path.isdir(DeploymentManager.getDeploymentFolder(serviceName, False)):
            # Deployment folder does not even exist.
            return False

        return os.path.isfile(DeploymentManager.__getMarkerFilePath(serviceName))

    @staticmethod
    def getAllServicesMarkedAsDeployed():
        parentFolder = os.path.join(app.defaults.DEFAULT_RUNTIME_DIR, "service")

        deployedServices = []
        if os.path.isdir(parentFolder):
            for serviceName in os.listdir(parentFolder):
                if DeploymentManager.isServiceMarkedAsDeployed(serviceName):
                    deployedServices.append(serviceName)

        return deployedServices

class ServiceControllerException(Exception):
    pass
