﻿#!/usr/bin/env python2

""" This module is used for validating Teamforge site option token values.
    It is a framework developed exclusively to define the property and
    validations if required for site option tokens.

    Though it is not mandatory for a developer to define the every Teamforge
    site option token property here, but it atleast helps in cases where a
    developer need to do some validations for some specific tokens.

    How to use this:
        An XML file viz teamforge_tokens_def.xml should be available in the
    same place where this module is located. A typical property of a Teamforge
    site option should contain the following xml block.

    <token key="SSL_KEY_FILE">
      <name>Apache SSL Certificate Key File</name>
      <description>Apache SSL Certificate Key File.</description>
      <validations>
        <maxlength></maxlength> <!-- Specify Maximum length -->
        <minlength></minlength> <!-- Specify Minimum length -->
        <fixlength></fixlength> <!-- Specify Fixed length -->
        <regex></regex>         <!-- Any Regular expression to match -->
        <eval>isFile</eval>     <!-- External validation functions. Refer validations.py -->
        <exec></exec>           <!-- External executable -->
      </validations>
    </token>

    Additional understanding:
       (*) None of the validation types are mandatory. A Token can be defined
    even without <validations> block. At the same time, if a validation type is
    declared with no value in it (for ex: <regex></regex>) then this validation
    will not be considered by this framework, IOW it is omitted or discarded.

       (*) One can have multiple external validation functions, external
    executables in a comma separated fashion.
    For ex: <eval>isFile, isFileExecutable</eval>.
    In the above example the framework will validate the token value using both
    the validation functions.

    Example:
      If a developer needs a validation for a Token value such as it should have
    length of atleast 10 byte and not more than 24 byte, should contain only
    alpha-numeric and should begin with a numeric character then the xml definition
    of the token should be declare something like this.

    <token key="MYPHRASEKEY">
      <name>My Phrase Key</name>
      <description>My phrase key used by teamforge licensed module.</description>
      <validations>
        <minlength>10</minlength>
        <maxlength>24</maxlength>
        <regex>^(\d).*</regex>
        <eval>isAlphaNumeric</eval>
      </validations>
    </token>

"""

__author__ = 'Moorthy Rajendran <rmoorthy@collab.net>'

import os
import re
import sys
import commands
from lxml import etree
import app.tokens.constants
import app.tokens.validations
import app.defaults

class TokenDef:

    def __init__(self, xmlFile = None):

        self._tokenDefXMLFile = "teamforge_tokens_def.xml"
        if (xmlFile is None):
           self._xmlFile = self._getTokenDefinitionFile()
        else:
           self._xmlFile = xmlFile

        self._tokensAll        = self.parseTokenXML()
        self._tokenValidations = app.tokens.validations.TokenValidations()
        self._componentList    = [ 'app', 'database', 'cvs', 'subversion' ]

    def parseTokenXML(self):
        """ Function that parses teamforge_tokens_def.xml and creates a token dictionary """

        xmlData = self._readXMLFile()
        rootDoc = etree.XML(xmlData)
        context = etree.iterwalk(rootDoc, events=("start", "end"))

        tokenDict = {}
        tokensAll = {}
        validationDict = {}
        autoGenDict = {}
        subElementProc = False
        tokenKey = None

        for action, elem in context:
            if (action == "start"):
                if elem.tag == "catalog":
                    pass
                elif elem.tag == "token":
                    tokenDict = {}
                    tokenKey  = elem.attrib['key']
                elif elem.tag == "validations":
                    validationDict = {}
                    subElementProc = True
                    for child in elem:
                        if type(child.tag) is str:
                            validationDict[child.tag] = [child.text, child.attrib]
                elif elem.tag == "notexist":
                    notexistDict = {}
                    subElementProc = True
                    for child in elem:
                        if type(child.tag) is str:
                            notexistDict[child.tag] = child.text
                elif elem.tag == "autogen":
                    autoGenDict = {}
                    subElementProc = True
                    for child in elem:
                        if type(child.tag) is str:
                            autoGenDict[child.tag] = child.text
                else:
                    if not subElementProc:
                        text = elem.text
                        tokenDict[elem.tag] = text
            elif (action == "end"):
                if elem.tag == "token":
                    tokenDict["token_verified"]= False
                    tokensAll[tokenKey] = tokenDict
                elif elem.tag == "validations":
                    tokenDict[elem.tag] = validationDict
                    subElementProc = False
                elif elem.tag == "notexist":
                    tokenDict[elem.tag] = notexistDict
                    subElementProc = False
                elif elem.tag == "autogen":
                    tokenDict[elem.tag] = autoGenDict
                    subElementProc = False

        return tokensAll

    def _readXMLFile(self):

        fileHandler = open(self._xmlFile, 'r')
        xmlOut = fileHandler.read()
        fileHandler.close()

        return xmlOut

    def _getTokenDefinitionFile(self):

        modFile = os.path.abspath( __file__ )
        modPath = os.path.dirname(modFile)
        xmlFile = os.path.join(modPath, self._tokenDefXMLFile)
        return xmlFile

    def setComponentList(self, cmpList = None):

        # _componentList should not be empty.
        if cmpList is not None:
           self._componentList = cmpList

    def getComponentList(self):

        return self._componentList

    def validateConfig(self, cfgHash):

        for key, value in cfgHash.iteritems():
            self.validateToken(key, value)

        self.validateMissingToken(cfgHash)

        return True

    def getProperty(self, key, property):
        if not self._tokensAll.has_key(key):
           return None

        defDict = self._tokensAll[key]

        if not defDict.has_key(property):
           return None

        return defDict[property]

    def checkComponentExists(self,tokenDist):

        tokenComponents    = tokenDist['components'].replace(' ','')
        tokenComponentList = tokenComponents.split(',')
        componentFound = False
        for thisComponent in tokenComponentList:
            try:
                self._componentList.index(thisComponent)
                componentFound = True
            except Exception:
                pass
        return componentFound

    def validateToken(self, key, value):
        if not self._tokensAll.has_key(key):
            return True

        defDict = self._tokensAll[key]

        # Validate token only if applications in 'components' is enabled in the local box
        if (defDict.has_key('components') and len(defDict['components']) > 0):
            if self.checkComponentExists(defDict) == False:
                defDict["token_verified"] = True
                return True
        if (value == app.tokens.constants.AUTO_KEYWORD):
            self.validateAutoToken(key, value, defDict);
            defDict["token_verified"] = True
            return

        if not defDict.has_key('validations'):
            defDict["token_verified"] = True
            return True

        validationDict = defDict['validations']

        for validationType in validationDict.keys():
            validationFunc = "_" + validationType
            moduleName     = self
            if hasattr(moduleName, validationFunc):
               functionCall = getattr(moduleName, validationFunc)
               functionCall(key, value,
                            validationDict[validationType][0],
                            validationDict[validationType][1])

        defDict["token_verified"] = True
        return True

    def validateMissingToken(self,cfgHash):
        """Validates the tokens which are not enabled in the site_options.conf"""

        for key,tokenDist in self._tokensAll.iteritems():
            if tokenDist["token_verified"] == True:
                continue
            if (tokenDist.has_key('components') and len(tokenDist['components']) > 0):
                if self.checkComponentExists(tokenDist) == False:
                    continue
            if not tokenDist.has_key('notexist'):
                continue

            notexistDist = tokenDist['notexist']
            if notexistDist.has_key('error'):
                message= "The token %s %s" % (key ,notexistDist['error'])
                raise Exception (message)
            elif notexistDist.has_key('warning'):
                BEGIN_WARNING = '\033[93m'
                END_WARNING = '\033[0m'
                message = "\n  The token %s %s" % (key ,notexistDist['warning'])
                sys.stdout.write(BEGIN_WARNING + "\n  " + message + "\n" + END_WARNING)

    def validateAutoToken(self, key, value, defDict):
        """Validates the whether the token can be configured for auto value generation"""

        if (not defDict.has_key('autogen')):
           raise Exception ("AutoGen Exception: The token [%s] cannot be auto-generated. Instead set a valid value for this token."%(key))

        autoGenDict = defDict['autogen']
        if (not autoGenDict.has_key('enabled')):
           raise Exception ("AutoGen Exception: The token [%s] cannot be auto-generated. Instead set a valid value for this token."%(key))

        autoGenEnabled = autoGenDict['enabled']
        if (autoGenEnabled != 'true'):
           raise Exception ("AutoGen Exception: The token [%s] cannot be auto-generated. Instead set a valid value for this token."%(key))

        return

    def _exec(self, key, value, execScripts, execAttr = None):
        """Validates the token value by evaluating an external executable"""

        if execScripts is None:
           return

        allExecScripts = execScripts.split(',')

        for execScript in allExecScripts:
            if execScript.find('%s') == -1:
               continue
            runCommand = execScript%(value)
            (ret, output) = commands.getstatusoutput(runCommand)
            if (ret != 0):
               raise Exception ("Exec Exception: Received while executing script [%s] for validating token %s. Error: %s"
                          %(execScript, key, output))

        return

    def _eval(self, key, value, evalFunctions, evalAttr = None):
        """Validates the token value by evaluating an external function"""

        if evalFunctions is None:
           return

        allEvalFunctions = evalFunctions.split(',')

        for evalFunction in allEvalFunctions:
           evalFunction = evalFunction.replace(' ','')
           if hasattr(self._tokenValidations, evalFunction):
              functionCall = getattr(self._tokenValidations, evalFunction)
              evalReturns  = functionCall(key, value)
              if (evalReturns[0] != 0):
                 if (evalAttr and evalAttr.has_key('errormsg')):
                    raise Exception ("Eval Exception: Received while executing function [%s]. Error: %s" %(evalFunction, evalAttr['errormsg']))
                 else:
                    raise Exception ("Eval Exception: Received while executing function [%s]. Error: %s" %(evalFunction, evalReturns[1]))

        return

    def _maxlength(self, key, value, expLength, expAttr = None):
        """Validates if the token value length is not more than the length value
           defined in 'maxlength' validation type"""

        if expLength is None:
           return

        expLength = int(expLength)
        calcLength = len(value)
        if (calcLength > expLength):
           raise Exception ("Maximum Length Exception: Token %s has value of length %s, but should not be more than %s"
                      %(key, calcLength, expLength))

    def _minlength(self, key, value, expLength, expAttr = None):
        """Validates if the token value length not less than the length value
           defined in 'maxlength' validation type"""

        if expLength is None:
           return

        expLength = int(expLength)
        calcLength = len(value)
        if (calcLength < expLength):
           raise Exception ("Minimum Length Exception: Token %s has value of length %s, but should not be less than %s"
                      %(key, calcLength, expLength))

    def _fixlength(self, key, value, expLength, expAttr = None):
        """Validates if the token value length equals to length value defined in
           'fixlength' validation type"""

        if expLength is None:
           return

        expLength = int(expLength)
        calcLength = len(value)
        if (calcLength != expLength):
           raise Exception ("Invalid Length Exception: Token %s has value of length %s, but should be %s"
                      %(key, calcLength, expLength))

    def _regex(self, key, value, regExp, regAttr = None):
        """Validates if the token value matches with the regular expression
           pattern defined in 'regex' validation type"""

        if regExp is None:
           return

        try:
           match = re.match(regExp, value)
           if match is None:
              raise Exception ("Invalid Regexp Exception: The value of token %s doesnot match regular expression %s"
                        %(key, regExp))
        except Exception as e:
           raise Exception ("Regular Expression Exception: While validating token %s. Reason: %s"
                       %(key, str(e)))

    def _usageForbidden(self, key, value, defaultValue, attr = None):
        """This token must not be used in a site-options.conf as it has been deprecated
           and the functionality has been removed"""

        if value is not None and value != defaultValue:
            raise Exception("The token %s is not supported anymore and related functionality has been removed. %s" % (key, self._tokensAll[key]["description"]))

    def _deprecated(self, key, value, defaultValue, attr = None):
        """This token should not be used in a site-options.conf as it has been deprecated
           and the functionality will be removed soon"""

        # By comparing with the default value, we avoid false warnings caused by default
        # values.

        if value is not None and value != defaultValue:
            BEGIN_WARNING = '\033[93m'
            END_WARNING = '\033[0m'

            message = "\n  The token %s has been deprecated and support for it will be dropped soon.\n  %s\n\n" % (key, self._tokensAll[key]["description"])
            sys.stdout.write(BEGIN_WARNING + message + END_WARNING)

    def _validJavaHome(self, key, value, defaultValue, attr = None):

        if os.path.isdir(value) and not os.path.islink(value):
            regex = r"([jre]+)(\d+\.)(\d+\.)([0]+\_\d+)"

            # Parsing the ctf's supprted JRE
            defaultJre = os.path.basename(app.defaults.DEFAULT_JAVA_HOME).split('.')
            defaultJreVersion = int(defaultJre[1])
            defaultJreBuild = int(defaultJre[2].split('_')[1])

            # Check if site-options is really pointing to an JRE, not a JDK.
            if os.path.isdir(os.path.join(value, "jre")):
                raise Exception("%s is pointing to a JDK installation, not a JRE" % key)

            # Validate if site-options has old JRE
            siteOptJreFolderName = os.path.basename(value)

            if re.match(regex, siteOptJreFolderName):
                siteOptJreVersion = int(siteOptJreFolderName.split('.')[1])
                siteOptJreBuild = int((siteOptJreFolderName.split('.')[2]).split('_')[1])

                if siteOptJreVersion == defaultJreVersion:
                    if siteOptJreBuild < defaultJreBuild:
                        raise Exception("Please remove the token %s from site-options.conf. TeamForge will then automatically use a supported JRE version. If you want to use a higher version than supported JRE version, only then enable the token." %key)

                elif siteOptJreVersion < defaultJreVersion:
                    raise Exception("Please remove the token %s from site-options.conf. TeamForge will then automatically use a supported JRE version. If you want to use a higher version than supported JRE version, only then enable the token." %key)
