﻿#!/usr/bin/env python2

""" Contains all the validations functions required for validating Teamforge
    Site Options.

    All methods in this module should adhere strict input and output singatures
    as these functions are called internally and not by developers explicitly.

    Input parameters: No of Arguments = 3
                        1st Arg : Is a blessed class object
                        2nd Arg : Teamforge Token name
                        3rd Arg : Teamforge Token value

    Return parameters: No of Response variables = 2
                        1st variable : Should have 0/1. 0=>Success, 1=>Failure
                        2nd variable : Should have None for success / Error
                                       String for failure

"""
__author__ = 'Moorthy Rajendran <rmoorthy@collab.net>'

from datetime import datetime

class TokenValidations(object):

    def isAlphabets (self, key, value):
        """Checks if value has only Alphabets"""
        if value.isalpha():
           return (0, None)
        else:
           return (1, 'Token %s should contain only alphabets'%key)

    def isNumeric(self, key, value):
        """Checks if value has only Numeric data"""
        if value.isdigit():
           return (0, None)
        else:
           return (1, 'Token %s should contain only numeric value'%key)

    def isTrueFalse(self, key, value):
        """Checks if value has either "true"/"false" """
        if (value == "true" or value == "false"):
           return (0, None)
        else:
           return (1, 'Token %s should have value either "true" or "false"'%key)

    def isTrue(self, key, value):
        """Checks if value has "true" """
        if (value == "true"):
           return (0, None)
        else:
           return (1, 'Token %s should have value "true"'%key)

    def isFalse(self, key, value):
        """Checks if value has "false" """
        if (value == "false"):
           return (0, None)
        else:
           return (1, 'Token %s should have value "false"'%key)

    def isAlphaNumeric(self, key, value):
        """Checks if value is a Alpha-Numeric data"""
        if value.isalnum():
           return (0, None)
        else:
           return (1, 'Token %s should contain only alpha-numeric value'%key)

    def isFile(self, key, value):
        """Checks if value is a valid existing file"""
        import os.path

        if not os.path.isfile(value):
           return (1, 'File not exists [%s]. Token %s should contain a valid filename'%(value, key))

        return (0, None)

    def isDir(self, key, value):
        """Checks if value is a valid existing directory"""
        import os.path

        if not os.path.isdir(value):
           return (1, 'Directory not exists [%s]. Token %s should contain a valid directory name'%(value, key))

        return (0, None)

    def isNotEmpty(self, key, value):
        """Checks if value is not empty"""
        if len(value):
           return (0, None)
        else:
           return (1, 'Token %s should not be empty'%key)

    def isDateMDY(self, key, value):
        """Checks if value is a valid date in MM/DD/YYYY format"""
        if not len(value):
           return (0, None)

        try:
           datetime.strptime(value, "%m/%d/%Y")
           return (0, None)
        except:
           return (1, 'Token %s should contain valid date in MM/dd/yyyy format'%key)

    def isValidTFPassword(self, key, value):
       """Checks if value has ampersand"""
       if '&' in value:
          return (1, "Token %s should not have '&' character" %key)
       else:
          return (0, None)

    def isValidDomainName(self, key, value):
       """Checks if value is a valid domain name (checks for underscore)"""
       if '_' in value:
          return (1, "Token %s should not have '_' character." %key)
       else:
          return (0, None)

    def isValidSessionTimeout(self, key, value):
       """Checks if value is between 0 and 24*60 (one day)"""
       if not (0 < int(value) <= 1440):
          return (1, "Token %s cannot be greater than 1440 or less than or equal to 0."%key)
       else:
          return (0, None)
