﻿#!/usr/bin/env python2
# unixusers.py: -*- Python -*-


""" Contains utilities related to unix users accounts -- creating, updating,
    and locking.
"""

import os
import pwd
import grp
import re
import shutil

import app.util

def config():
    return app.util.getRuntimeConfiguration()

def _assertCmdAllowed(cfg, failure_msg):
    if cfg.get('USE_USERADD', 'true') == 'false':
        raise Exception(failure_msg)

def _canRemoveAccounts(cfg):
    if cfg.get('REMOVE_CONFLICTING_ACCOUNTS', 'false') == 'true':
        return 1
    else:
        return 0

def enableGroup(group, preferred_gid):
    # if we're not root, we might as well return -- we're not doing anything.
    if not app.util.invokedBySuperUser():
        return
    # TODO : correct the group if it exists w/ a different gid?
    # first, ensure that the group exists; if not, create it.
    try:
        grp.getgrnam(group)
    except KeyError:
        _assertCmdAllowed(config(), 'Site configuration prevents use of groupadd '
                          'to create group %s' % group)
        # If there is a error with groupadd, the group(id) might exist already
        # Delete the existing group and recreate it if
        # REMOVE_CONFLICTING_ACCOUNTS is set to true.
        command = "groupadd -g %s %s" % (preferred_gid, group)
        try:
            gid_record = grp.getgrgid(preferred_gid)
        except KeyError:
            # The 'preferred_gid' is not in use.
            gid_conflict = 0
        else:
            # Another group has 'preferred_gid'.
            gid_conflict = 1

        if gid_conflict:
            # Resolve case where 'preferred_gid' conflicts with a group on
            # the system.
            if _canRemoveAccounts(config()) or re.search('sf-admin', gid_record[0]):
                # If REMOVE_CONFLICTING_ACCOUNTS is true, do so.
                try:
                    _runCommandWithCheck("groupdel %s > /dev/null 2>&1" % gid_record[0])
                except:
                    # Maybe the Group is the primary group. Try with the
                    # next gid and move this groupid to that.
                    new_gid = gid_record[2] + 1
                    _runCommandWithCheck("groupmod -g %s %s"
                                  % (new_gid, gid_record[0]))
            else:
                command = "groupadd %s" % (group)

        # Now attempt to add the group
        try:
            _runCommandWithCheck(command)
        except Exception as err:
            raise Exception("groupadd for %s failed: %s" % (group, err))

def enableUserAccount(login, comment, uid, group, home, shell):
    # if we're not root, we might as well return -- we're not doing anything.
    if not app.util.invokedBySuperUser():
        return

    # first, check that the user exists; if not, create it.
    try:
        user_record = pwd.getpwnam(login)
    except KeyError:
        # this means the user doesn't exist
        user_record = None


    # if the group doesn't exist, we'll throw an error.  that should
    # never happen -- enableGroup should have been called first;
    # throwing an exception sounds OK.
    gid = grp.getgrnam(group)[2]
    if user_record is None:
        _assertCmdAllowed(config(), 'Site configuration prevents use of useradd '
                          'to create user %s' % login)
        # avoid creating the home dir at this point, but create the
        # account
        if config().get('PLATFORM').find('redhat') != -1:
            args = '-M'
        else:
            args = ''
        #command = "useradd %s -c '%s' -d '%s' -s '%s' -u %s -g %s %s" % \
        #          (args, comment, home, shell, uid, gid, login)
        command = "useradd %s -c '%s' -M -s '%s' -u %s -g %s %s" % \
                    (args, comment, shell, uid, gid, login)
        #If there is a error with useradd, the user(id) might exist already
        # Delete the existing user and recreate it if
        # REMOVE_CONFLICTING_ACCOUNTS is set to true.

        try:
            uid_record = pwd.getpwuid(uid)
        except KeyError:
            # 'uid' is not in use.
            uid_conflict = 0
        else:
            # Another user account has 'uid'.
            uid_conflict = 1

        if uid_conflict:
            # Resolve case where 'uid' conflicts with that of a user on
            # the system.
            if _canRemoveAccounts(config()) or re.search('sf-admin', uid_record[0]):
                # If REMOVE_CONFLICTING_ACCOUNTS is true, do so.
                _runCommandWithCheck("userdel %s" % uid_record[0])
            else:
                # Otherwise, raise an Exception so the user can fix things.
                raise Exception("This UID Already exists. Please remove "
                                "the account manually. UID: %s " %uid)
        try:
            _runCommandWithCheck(command)
        except Exception as err:
            raise Exception("useradd for %s failed: %s" % (uid_record[0], err))

        # now we carry on
        user_record = pwd.getpwnam(login)

    # note that home directory creation is generally a separate step
    # here we just do a simple chown iff the directory exists.
    if os.path.isdir(home):
        # avoid recursive chown here -- it's probably not needed, and
        # there might be lots of files there (which could take a while)
        # os.chown reports errors with an exception; no need to check return
        os.chown(home, user_record[2], gid)

    # third, we ensure that the user record has the correct shell and home
    # directory -- regardless of whether the home directory exists or not.
    if user_record[5] != home or user_record[6] != shell:
        _assertCmdAllowed(config(), 'Site configuration prevents use of usermod '
                          'to modify user %s' % login)
	result = os.system('/bin/grep -q -E "^%s:" /etc/passwd' % login)
	if result == 0:
            home = os.path.normpath(home)
            _runCommandWithCheck("usermod -d %s -s %s %s -p %s" % (home, shell, login, login))


def setupUserHomedir(login, group, home):

    # if we're not root, we might as well return -- we're not doing anything.
    if not app.util.invokedBySuperUser():
        return
    # if the group or user doesn't exist, we'll throw an error.  that should
    # never happen -- enableGroup/enableUserAccount must be called first.
    gid = grp.getgrnam(group)[2]
    user_record = pwd.getpwnam(login)

    # second we determine if we should create the home directory.  we do this
    # by checking if the data directory exists -- that's to allow this
    # to set up user accounts without touching the data directory when
    # create-runtime is run on, for example, disaster recovery boxes or
    # boxes which are about to receive migrated data -- this is
    # typically before the data is installed.
    if not os.path.isdir(home):
        # create the parent directory if it does not already exist
        if not os.path.isdir(os.path.realpath(os.path.split(home)[0])):
            os.makedirs(os.path.realpath(os.path.split(home)[0]), 0755)

        # look for seed data; copy if found
        initial_data = os.path.join(config().get('DISTRIBUTION_DIR'),
                                    'initial-data', 'homedirs', login)
        if (os.path.isdir(initial_data)):
            shutil.copytree(initial_data, home)
        else:
            # create an empty home directory
            os.makedirs(home, 0755)
        # recursive chown here -- we created all the files in the
        # home dir
        chownHomedir(login, group, home)
    else:
        # avoid recursive chown here -- it's probably not needed, and
        # there might be lots of files there (which could take a while)
        # os.chown reports errors with an exception; no need to check return
        os.chown(home, user_record[2], gid)

def chownHomedir(login, group, home, subdir=None):
    if subdir is None:
        dir = home
    else:
        dir = os.path.join(home, subdir)

    if (os.path.exists(dir)):
        _runCommandWithCheck("chown -R %s:%s %s" % (login,group,dir))

def disableUserAccount(login, shell = None):
    shell = shell or config().get('DISABLED_USER_SHELL')
    # if we're not root, we might as well return -- we're not doing anything.
    if not app.util.invokedBySuperUser():
        return

    # if the user doesn't exist, we don't need to lock the account.  otherwise,
    # we'll do so.
    try:
        user_record = pwd.getpwnam(login)
    except KeyError:
        user_record = None

    if user_record is not None:
        _assertCmdAllowed(config(), 'Site configuration prevents use of '
                          'passwd and usermod to modify user %s' % login)
        _runCommandWithCheck("passwd -l %s" % (login,))
        _runCommandWithCheck("usermod -s %s %s" % (shell, login))


def _runCommandWithCheck(command):
    _setupEnvironment()

    result = os.system(command)
    if (result != 0):
        raise Exception("command '%s' failed: returned %d" % (command,
                                                              result >>8,))

def _setupEnvironment():
    # ensure the path is consistent, and includes the dirs we need
    os.putenv('PATH', config().get('UNIX_COMMANDS_PATH'))

