﻿#
# This module contains items necessary to support general Version Control
# operations (CVS, SVN, etc).
#
# Legacy code
# pylint: skip-file

import os
import re
import sys
import shutil
import string
import types
import xmlrpclib
import threading


# ------------------------  XMLRPC utilities  ------------------------

# this config value is used to authentication access to the XMLRPC server
AUTH_KEY = 'DATABASE_READ_ONLY_PASSWORD'


def assertValidCommit(cfg, projectName, userName, msg):
  """If invalid, throws an exception describing why the failure may
  have occurred."""
  if not _invoke(cfg, 'isCommitValid', projectName, userName, msg):
    raise Exception("Invalid change log message or author: Message "
                    "must identify an artifact, and the author (%s) "
                    "may need to own the referenced artifact" %
                    userName or "none")

def invalidateCachedVCBackendForID(cfg, projectID):
  _server(cfg).cache.invalidateCache(cfg[AUTH_KEY],
                                     'org.tigris.helm.om.AttributeValue',
                                     ['component-vc', projectID])

def postCommit(cfg, projectName, userName, msg, id='n/a', branch='n/a',
               paths_changed=''):
  return _invoke(cfg, 'logCommitInfo', projectName, userName, branch, id, msg,
                 paths_changed)

def getVCBackendForID(cfg, projectID):
  return _server(cfg).attribute.getAttributeValue(cfg[AUTH_KEY],
                                                  'component-vc',
                                                  projectID)

def _invoke(cfg, methname, *args):
  return getattr(_server(cfg).vc, methname)(cfg[AUTH_KEY],
                                            cfg['DOMAIN'],
                                            *args)

def _server(cfg):
  return xmlrpclib.Server('http://%s:%s/RPC2' % (cfg['XMLRPC_HOST'],
                                                 cfg['XMLRPC_PORT']))


# ------------------------  External scripting -------------
def run_cmd(cmd):
    """Run command CMD in the current working direcotry, print any
    stderr output from the child process, and return a tuple of two
    lists: the lines of stdout output from CMD, and the lines of
    stderr output."""

    stdout_lines = []
    stderr_lines = []

    # Run the command, getting pipes for reading/writing from/to the child.
    try:
       from subprocess import Popen, PIPE
       p = Popen(cmd, shell=True, stdin=PIPE, stdout=PIPE, stderr=PIPE)
       child_in, child_out, child_err = (p.stdin, p.stdout, p.stderr)
    except ImportError:
       child_in, child_out, child_err = os.popen3(cmd)

    # Close up the child's input stream.  We aren't using it.
    child_in.close()

    def _append_lines(stream, ary):
        ### replace the contents of ARY with the lines from STREAM.
        ary[:] = stream.readlines()

    # Create threads for reading from the child's stdout and stderr.
    outthread = threading.Thread(target=_append_lines,
                                 name="StdoutReader",
                                 args=(child_out, stdout_lines))
    errthread = threading.Thread(target=_append_lines,
                                 name="StderrReader",
                                 args=(child_err, stderr_lines))

    # Start and join those threads.
    outthread.start()
    errthread.start()
    outthread.join()
    errthread.join()

    # All done.  Return our findings.
    if stderr_lines:
        print "----- STDERR FROM '",
        if isinstance(cmd, types.StringTypes):
          print cmd,
        else:
          print string.join(cmd),
        print "':"
        map(sys.stderr.write, stderr_lines)
        print "----- END STDERR -----"

    return stdout_lines, stderr_lines


# ------------------------  Usage note formatting  ------------------------

def usage(before, long_opts, after):
    """Print BEFORE, then format and print LONG_OPTS, then print AFTER.

LONG_OPTS is a list of lists, each sublist contains three elements: the
option name, the placeholder name of its argument if any (else None),
and a description string.  If the option name ends with '=', the
option takes an argument, so the placeholder arg name will be used
(defaulting to 'FOO'); else if the option does not take an argument,
the placeholder argument name is ignored.

If BEFORE or AFTER is None, it is not printed."""
    if before:
        print before
        print ""

    for x in long_opts:
        if x[0][-1:] == '=':
            print "   --" + x[0] + (x[1] or "FOO")
        else:
            print "   --" + x[0]
        print "     ", x[2]

    print ""
    if after:
        print after

# -----------  Manipulating specialized file/directory trees  -----------

# Usage:
# app.vc.prunedcopytree('^(CVS|\.svn|lost\+found)$', src, dst, symlinks)
#     - or -
# copier = app.vc.prunedcopytree_compile('^(CVS|\.svn|lost\+found)$')
# copier.copytree(src, dst, symlinks)
#     - or -
# regexp = re.compile('^(CVS|\.svn|lost\+found)$'[, flags])
# use either above form, passing regexp instead of pattern

def prunedcopytree_compile(reorpat):
  return PrunedCopyTree(reorpat)

def prunedcopytree(reorpat, src, dst, symlinks=0):
  return PrunedCopyTree(reorpat).copytree(src, dst, symlinks)

class PrunedCopyTree:
  def __init__(self, reorpat="^$"):
    STRING_TYPES = [type("")]
    try:
      STRING_TYPES.append(type(unicode("")))
    except NameError:
      pass
    if type(reorpat) in STRING_TYPES:
      self.ignores = re.compile(reorpat)
    else:
      self.ignores = reorpat

  def copytree(self, src, dst, symlinks=0):
    """Recursively copy a directory tree, pruning some stuff.

  The destination directory may exist or not.
  Things matching the regexp / pattern provided at construction
  are skipped..
  Errors are reported to standard output.

  If the optional symlinks flag is true, symbolic links in the
  source tree result in symbolic links in the destination tree; if
  it is false, the contents of the files pointed to by symbolic
  links are copied.

  XXX Based on shutil.copytree.

  """
    names = os.listdir(src)
    if not os.path.isdir(dst):
      os.mkdir(dst)
    for name in names:
      if not self.ignores.search(name):
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)
        if not self.ignores.search(srcname):
          try:
            if symlinks and os.path.islink(srcname):
              linkto = os.readlink(srcname)
              os.symlink(linkto, dstname)
            elif os.path.isdir(srcname):
              self.copytree(srcname, dstname, symlinks)
            else:
              shutil.copy2(srcname, dstname)
              # XXX What about devices, sockets etc.?
          except (IOError, os.error), why:
            print "Can't copy %s to %s: %s" % (`srcname`, `dstname`, str(why))

