﻿import socket
import fcntl
import os

g_recursiveAcquisitionCount = 0

class GlobalLock(object):
    """ A machine-global lock """
    def __init__(self):
        self.__socket = None

    def __enter__(self):
        """
            Acquire a system-wide lock for this process. It will be held until the
            process terminates or the 'with' block ends.
        """
        global g_recursiveAcquisitionCount

        if g_recursiveAcquisitionCount > 0:
            # Lock acquired already
            g_recursiveAcquisitionCount += 1

        else:
            # Without holding a reference to our socket somewhere it gets garbage
            # collected when the function exits
            self.__socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)

            # Do not inherit this socket to child processes
            flags = fcntl.fcntl(self.__socket, fcntl.F_GETFD)
            fcntl.fcntl(self.__socket, fcntl.F_SETFD, flags | fcntl.FD_CLOEXEC)

            try:
                self.__socket.bind('\0TeamForgeRuntime')

                # We bound successfully, that means we own the lock now.
                g_recursiveAcquisitionCount += 1

            except socket.error:
                raise LockOwnedByOtherProcess()

    def __exit__(self, type, value, traceback):
        global g_recursiveAcquisitionCount

        g_recursiveAcquisitionCount -= 1
        if self.__socket and g_recursiveAcquisitionCount == 0:
            self.__socket.close()

class NetworkFolderLock(object):
    """ A lock on a a shared folder. The lock is non-recursive, but adopts dangling locks. """
    def __init__(self, path, ownerName):
        self.__ownerName = ownerName
        self.__path = path

    def __getOwner(self):
        try:
            with open(self.__getLinkFilePath(), "r") as f:
                return f.read()
        except Exception:
            return "unknown"

    def __getLinkFilePath(self):
        return os.path.join(self.__path, ".lock")

    def __getOwnerFilePath(self):
        return "%s.%s" % (self.__getLinkFilePath(), self.__ownerName)

    def __enter__(self):
        if self.__getOwner() == self.__ownerName:
            # Reacquiring dangling lock.
            return

        # Create a file that carries the owner name. Existing files can just be overwritten.
        with open(self.__getOwnerFilePath(), "w") as f:
            f.write(self.__ownerName)

        # Now try to acquire the lock by attempting to create a hard link to the file.
        # Note that link() is atomic, even on NFS.
        try:
            os.link(self.__getOwnerFilePath(), self.__getLinkFilePath())

            # Yay, we own the lock.
            return

        except:
            os.unlink(self.__getOwnerFilePath())

            # Somebody else has the lock. Let see if we can find out who (although this
            # is prone to a race condition and therefore not necessarily accurate).
            owner = self.__getOwner()

            raise NetworkFolderOwnedByOtherPeer(
                "Folder %s is locked (probably by %s)" % (self.__path, owner), owner)

    def __exit__(self, type, value, traceback):
        if self.__getOwner() == self.__ownerName:
            os.unlink(self.__getLinkFilePath())

        if os.path.exists(self.__getOwnerFilePath()):
            os.unlink(self.__getOwnerFilePath())


class NetworkFolderOwnedByOtherPeer(Exception):
    def __init__(self, message, peer):
        super(NetworkFolderOwnedByOtherPeer, self).__init__(message)
        self.__peer = peer

    def getPeer(self):
        return self.__peer

    def __str__(self):
        return "Lock owned by %s" % self.__peer

class LockOwnedByOtherProcess(Exception):
    pass
