# TeamForge Components

TeamForge comprises a set of components -- CTF Core, SVN, James, etc. Some components are always
required to be installed, others are optional. By spreading these components over multiple RPMs,
we can make sure that users do not have to install everything all the time. While this is valuable,
RPMs alone prove insufficient to manage the components and their inter-dependencies:

* Some components do not have a physical representation but are configuration-only.
* RPM dependencies are restricted to the local machine only -- however, in a multi-box install, 
  dependencies between components must be tracked across machines.

In addition to the _physical_ componentization (using RPMs), there is thus also a need for a
_logical_ componentization of TeamForge.

## Services   

Services represents a _logical_ component of TeamForge. Such a _logical_ component may either be a 
feature which the user explicitly opts to install (ex: _cvs_) or a technical component (ex: Apache, Logrotate) which other 
services depend on.

Services come with additional metadata which makes it possible to track and manage dependencies
on a more fine-grained level than with RPMs alone:

 * Deployment dependencies specify which other services need to be deployed locally.
 * Provided Endpoints specify which network endpoints the service offers.
 * Required Endpoints specify which network endpoints the service depends on.

As such services are more fine-grained than RPMs and it is common to have a single RPM contain 
multiple services.

## Service Lifecycle

* *Uninstalled*: A service is uninstalled if the RPM that contains it is not installed. Uninstalled 
  services do not exist from the perspective of TeamForge.
* *Undeployed*: The RPM containing the service is installed, but the service has not been _deployed_ 
  yet. _Deployment_ is what used to be called _creating the runtime_ pre-16.7, but is specific to
  one service.
* *Dependencies unavailable*: The service itself might be available, but at least one of its deployment
  dependencies is not in _Available_ state.
* *Available*: Service is fully functional.

Services that mnage data additional states:

* *Not bootstrapped*: Data structures have bot been initialized yet.
* *Not migrated*: Data structures are there, but data needs to be migrated.

Services that have a daemon have the following additional states:

* *Dependencies unavailable*: The service itself might be available, but at least one of its deployment
  dependencies is not in _Available_ state.
* *Ports Blocked*: The service is impeded from starting up because at least one of the ports it needs
  is in use by a different process.
* *Stopped*: Service is an autos tart-type service, yet is stopped.
* *Inactive*: Service is a demand start-type service and is stopped.
* *Starting*: Service is in the process of starting up.
* *Available*: Service is running and healthy according to its health check.
* *Unhealthy*: Service is running but unhealthy according to its health check.
* *Dead*: Service is supposed to be running, but the process disappeared.
* *Doomed*: Service is technically running, but it will never work properly because some part of it 
  failed to initialize properly
* *Stopping*: Service is in the process of stopping.

## Auto start vs demand start

By default, services are considered _auto start_. That means that the service are started by default
when the machine starts and it is considered a problem if the service is stopped.

A service can opt to be considered _demand start_ though. Such services are not started by default
and it is considered okay for them to be stopped. Starting a _demand start_ service either has 
to be done manually or thru some other means such as cron.

## Service Files

All files owned by a service should be installed to `/opt/collabnet/teamforge/service/{service-name}/`. 
For backward compatibility, this is not the case for many services yet though.

The conventions for placing files are:

* `manifest.json`       Service manifest (mandatory)
* `etc/`:               Installer-generated configuration, empty intermediate folder
  * `v.v.v/`:           Installer-generated configuration for version v.v.v (ex: 16.7.321)
* `bin/`:               Scripts: Init script, tools
* `lib/service.py`:     Custom Service Controller
* `lib/shared/`         Folder for shared Python modules, will be added to Python path
* `hooks/`: 	
   * `option-snippets/` Option Snippets for Service X
   * `migrate/`: 	    Migration scripts for Service X
   * `deploy/`: 	    Deployment scripts (formerly called create-runtime scripts) for Service X
   * `undeploy/`: 	    Undeployment scripts 
   * `bootstrap/`: 	    Bootstrap scripts for Service X
   * `initialize/`: 	Initialization scripts (run after startup)
   * `policy/`:         SElinux policy files, will be installed automatically
   * `folder/`:         Folder hooks to customize folder creation and folder permission management.
   * `cron/`: 	        Cron scripts for Service X
      * `5mins/`
      * `hourly/`
      * `daily/`
      * `weekly/`
      * `monthly/`

### Manifest

The only mandatory file for a service is the _manifest_, which contains the metadata describing the service.

The header of the file contains some basic information such as name and description. The `name` attribute
must contain the name of the folder, the `description` can be anything and is not evaluated.


```
{
    "name": "myservice",
    "description": "This service does something useful",
```

If the service has a daemon process, then it needs to include a section describing the properties of that daemon.

For a custom daemon, `user` and `group` objects must be provided to specify the credentials the daemon should be 
started as. For each `user` and `group`, at least a `default` value needs to be provided. Optionally, `option` 
can be used to specify the site option which can be used to override the default.

Finally, `startupTimeout` defines how long the daemon startup is allowed to take at maximum before startup is 
considered to have failed.

```
     "daemon": {
        "user": {
            "default": "sf-admin",
            "option": "APP_USER"
        },
        "group": {
            "default": "sf-admin",
            "option": "APP_GROUP"
        },
        "startupTimeout": 300
    },
```

In addition to these settings, a daemon must also have a `lib/service.py` file containing a class deriving from 
`app.servicecontrol.ServiceController`. A daemon is expected to at least override the `getPidFile`, `startDaemonAsync` 
and `stopDaemonAsync`methods, although many services will override additional methods to determine the service status 
like `isBootstrapped` as well.

To run simple console applications (which cannot daemonize themselves) as daemon, there is the 
`SimpleDaemonController` class, which can be used as base class. It only expects `getCommand` to be overridden 
and return the command line to run.

The next section defines the _deployment_ dependencies. Any service that must be _deployed_ before this service can
be deployed locally is considered a deployment dependency. If the service represents an application that is hosted
in a container which is represented by a different service (example: an EAR hosted in Jboss), the "container" attribute 
refers to the container service. The container service has to be a deployment dependency.

```    
    "deployment": {
        "dependencies": ["apache", "cvs", "jboss"],
        "container": "jboss"
    },
```

The next section contains the list of endpoints (provided, required). 

* A _provided endpoint_ is an endpoint that is exposed for use by other services. Whether there actually is a 
  service that needs this endpoint or not is irrelevant. Provided endpoints have the following attributes:
  * `id`: Identifier for the endpoint. It should be componsed of the protocol (like `http`) and its purpose,
    ex: `http-soap`. 
  * `scope`: Defines the visibility of the endpoint. It can have either of two values:
    * `internal`: Endpoint is for TeamForge-internal cross-service communication only, but not for end user clients.
    * `public`: Endpoint is intended to be used by end user clients.
  * `port`: Defines the port number used. 
  * `via`: In some cases, the endpoint is not supposed to be accessed directly but _via_ some other service
    (like Apache). In this case, `via.service` specifies the service which should be used to access the endpoint
    and `via.id` defines the endpoint of the service to use.
   
* A _required endpoint_ is an endpoint from another service that this services consumes. It does not matter
  where this service is deployed (locally or on a different box), but it needs to be provided somewhere.
  Required endpoints have the following attributes:
  * `service`: Specifies the service whose endpoint is required.
  * `id`: Identifies the endpoint of this service.
  * `scope`: Defines what the endpoint is needed for. Allowed values are:
    * `vital`: Endpoint must be available to even start up the service.
    * `default`: The endpoint is required for the service to operate properly, but the service can start up while the 
       endpoint is unavailable-
    * `optional`: The endpoint is not required for startup and not required at runtime -- it is strictly optional.
  
The _required_ and _provided_ endpoints are wired up in `app.landscape.Site.resolveEndpointDependencies`. For this,
a few rules apply:

* Provided endpoints must be unique -- it is considered an error if there is more than one instance of a service.
* For intra-machine communication, endpoints with _internal_ scope are preferred. This avoids unnecessary overhead
  caused by Apache reverse-proxying or SSL encryption.
* For intra-machine communication, _localhost_ is used as hostname. This ensures that communication goes via
  the loopback interface, which may be necessary because some _internal_ scoped endpoints may only be bound
  to 127.0.0.1, but not to a "real" IP.
* For cross-machine communication, public endpoints are preferred. The rationale here is that public endpoints
  are usually better secured (SSL) and chances are better that there is no firewall rule in the way. If a no
  _public_ scoped endpoint is available, an _internal_ scoped endpoint is used.

  
```
    "endpoints": {
        "provided" : [ 
            {
                "id": "http-web",
                "scope": "internal",
                
                "description": "Web UI",
                
                "port": {
                    "default": 8080,
                    "option": "MYSERVICE_LISTEN_PORT"
                }
            },
            {
                "id": "http-web",
                "scope": "public",
                "via": {
                    "service": "apache",
                    "id": "http"
                }
            }
        ],
        "required": [ 
            {
                "service": "ctfcore-database",
                "id": "jdbc",
                "scope": "vital"
            },
            {
                "service": "search",
                "id": "rmi" 
            },
            {
                "service": "etl",
                "id": "http-soap",
                "scope": "optional"
            }
        ]
    }
```

The difference between a _deployment dependency_ and an _endpoint dependency_ might be confusing at first, however
the two types of dependencies serve different purposes:

* A _deployment dependency_ states that some other service must be deployed *on the same machine* -- 
  usually because a service depends on the files or background processes this service puts in place.
* An _endpoint dependency_ states that some other service may or must be deployed *somewhere* because 
  a service intends to talk to its endpoint -- it really does not matter on which machine, all you care about is that 
  there is a machine name and port to talk to. In fact, all endpoints should be derived from endpoint dependencies –- 
  services should never assume that something is installed locally and should not juggle with host names and ports 
  themselves -– instead, they should declare and endpoint dependency and let the runtime figure out what the right 
  hostname and port number is. This becomes particularly important in HA environments, where there may be 
  load balancers in place.


If a service manages any files and folders which should be created during the installation, then the `folders` section
can be used to describe these. The installer will create these folders and apply ownership/permissions and Selinux 
context during the _bootstrap_ phase. During  the _migrate_ phase, it will re-create any missing folders, but 
(for performance reasons) not reapply file ownership/permissions. 
 * Folder ownership/permissions can be reapplied using `teamforge apply-permissions`.
 * Folder selinux context can be reapplied using `teamforge apply-selinux`.
 * All Service selinux policies can be unloaded using `teamforge unload-selinux` 

* Each entry in the `folders` map corresponds to a folder. If the path is relative, then it is assumed to be relative
to `/opt/collabnet/teamforge/var`. 

* The `state` attribute defines whether the folder should be `present` (default) or `absent`.
* `mode` defines the file mode (in octal).
* `user` and `group` (both optional) define the ownership. If the user is not static, a site option can be referenced
  to determine the user/group name. If the attributes are not defined, the daemon user/group settings are used. 
  If that is not defined either, the processes effective user and group are used.
* `context` (optional) defines the SElinux context. 
* `recursive` (optional) attributes defines whether the permissions should be applied recursively (default) or not.

Notes regarding SElinux:
* It is generally preferred to use SElinux policies over using the `context` attribute. If both are present, the
  `context` attribute value prevails.
* The runtime will run `restorecon` for all folders listed in the `folders` section, regardless of whether `context`
  has been specified or not. This ensures that any existing, manually specified settings are being reset.
  
```
    "folders": {
        "some/relative/path": {
            "mode": "0751",
            "owner" : {
                "default" : "apache",
                "option" : "HTTPD_USER"
            },
            "group" : {
                "default" : "apache",
                "option" : "HTTPD_GROUP"
            },
            "context": "system_u:object_r:tf_cvs_t:s0",
	    "recursive": true
        },
        
        "/some/absolute/path": {
            "state": "present"
            "mode": "0751", 
	    "recursive": false
        },
        
        "/some/obsolete/path": {
            "state": "absent"
        }
    }

```

Finally, a service should specify how it behaves in an HA environment.

* The `compatible` flag determines if the service can be deployed in an environment where there
  is at least one (other) service that is deployed redundantly. The flag defaults to false.
* The `clusterable` flag determines if the service itself is capable of being deployed 
  redundantly. Whether the service operates in an active/active or active/passive fashion is
  irrelevant in this context. The flag defaults to false.

```
    "ha": {
        "compatible": true,
        "clusterable": true
    }
```


### Using endpoints in configuration files

The endpoint information derived from the manifest files is made available for use in configuration snippets etc.
via special runtime options:

```
{service}/bind/{endpoint-id}/host=%s
{service}/bind/{endpoint-id}/port=%s
```

These runtime option describe which host and port the service _{service}_ should bind to for endpoint _{endpoint-id}_. 
Services other than _{service}_ *MUST NOT* use this runtime option. They can be used in snippets as usual using 
the syntax `{__someservice/bind/http/port__}`

```
{service}/dependency/{provider-service}/{endpoint-id}/host
{service}/dependency/{provider-service}/{endpoint-id}/port
```

These runtime option are intended to be used only for _{service}_ and describe which host/port the endpoint _{endpoint-id}_ 
of service _{provider-service}_ is available on.

For all these options, the runtime will make sure that_host_ contains the right host name -- whether that is _localhost_,
the physical machine nane, or the domain name -- based on the `scope` information in the manifest files. Users of these 
runtime option therefore do not need to care about whether it is a single- or multibox deployment and whether the service
is publicly exposed or not.

## Criteria to split services

Everything that could potentially be installed on a separate server should be implemented as a separate service. Beyond that,
there are no hard and fast rules on when it is okay to implement a component as a single service vs. when a service
should be split. However, common reasons to split services include:

* Some components may be optional (i.e., should only be deployed when mentioned in the HOST/SERVICES token).
* Having more than 1 daemon process to control.
* Avoiding cyclic endpoint dependencies.
* Making it more manageable.
