#!/opt/collabnet/teamforge/runtime/bin/wrapper/python

import getopt
import os
import sys
import time
import getpass
from suds.client import Client
from suds.cache import ObjectCache

import app.util
import app.db

# Global Parameters
STDOUT  = True

def main():
    """ Main Function """

    global STDOUT

    try:
       opts, args = getopt.getopt(sys.argv[1:], 'hsu:p:', ['help', 'silent','tfuser=', 'tfpasswd='])
    except:
       printUsageAndExit('Invalid option supplied.')
    tf_username = ''
    tf_password = ''
    for opt, value in opts:
        if (opt in ('-h', '--help')):
            printUsageAndExit()
        elif (opt in ('-s', '--silent')):
            STDOUT = False
        elif (opt in ("-u", "--tfuser")):
           tf_username = value
        elif (opt in ("-p", "--tfpasswd")):
            tf_password= value
        else:
            printUsageAndExit('Invalid option supplied ' + opt )

    # Check if this script can run now
    checkRunProgram()

    # Get TeamForge username and password
    if tf_username == '':
        tf_username = raw_input("Enter TeamForge Site-admin Username:")
    if tf_password == '':
        tf_password = getpass.getpass("Enter TeamForge Site-admin Password:")

    enable_eventq_component(tf_username,tf_password)

def __connectToDatabase():
    config = app.util.getRuntimeConfiguration()
    return app.db.Database.connect(
        config.get('DATABASE_TYPE'),
        config.get('ctfcore-app/dependency/ctfcore-database/jdbc/host'),
        config.get('ctfcore-app/dependency/ctfcore-database/jdbc/port'),
        config.get('DATABASE_NAME'),
        config.get('DATABASE_USERNAME'),
        config.get('DATABASE_PASSWORD'))

def enable_eventq_component(username,password):
    """Adds EventQ integrated application to all projects"""

    config = app.util.getRawRuntimeConfiguration()

    plug_id, prefix = get_plug_id_and_prefix()
    soap_url = "%s/ce-soap60/services" % config['API_SOAP']
    sudsCacheDir = os.path.join(config['TEMPORARY_DIR'], 'suds')
    sudsCache = ObjectCache(sudsCacheDir)
    cnapp = Client("%s/CollabNet?wsdl" % soap_url, cache=sudsCache, location="%s/CollabNet?wsdl" % soap_url)
    session_id = cnapp.service.login(username, password)
    plugapp = Client("%s/PluggableApp?wsdl" % soap_url, cache=sudsCache, location="%s/PluggableApp?wsdl" % soap_url)

    for proj_id in get_project_ids():
        try:
            named_values = plugapp.service.enablePluggableComponent(session_id, proj_id[0], plug_id, None, prefix)
            names = named_values['names']
            if len(names) == 0:
                printLog("EventQ component is added to project id: %s" % proj_id)
            else:
                values = named_values['values']
                printLog("Error: Failed to add EventQ to project id: %s" % proj_id)
                for i in range(len(names)):
                    printLog("%s: %s" % (names[i], values[i]))
                sys.exit(1)
        except Exception as e:
            printLog("Error: Failed to add EventQ to project id: %s" % proj_id)
            printLog("Reason: %s" % e)
            sys.exit(1)

    printLog ("Done.")

def get_plug_id_and_prefix():
    """Get id and prefix of EventQ integrated application"""

    try:
        with __connectToDatabase() as db:
            result = db.query("select id, prefix, is_enabled from integrated_application where name = 'EventQ'")

    except Exception as e:
        printLog("Error: Unable to execute sql query")
        printLog("Reason: %s" %(e))
        sys.exit(1)

    if len(result) == 0:
        printLog("Warn: EventQ is not added to projects as it is not integrated to the site")
        sys.exit(1)
    elif result[0][2] == False:
        printLog("Warn: EventQ is not added to projects as it is disabled")
        sys.exit(1)

    return (result[0][0], result[0][1])

def get_project_ids():
    """Get all project ids that do not have EventQ"""

    printLog ("Fetching project ids ...")
    try:
        with __connectToDatabase() as db:
            return db.query("""select id from project where is_deleted = '%s' and is_locked = '%s' and path like 'projects.%s'
                and id not in (select iapa.project_id from integrated_application ia
                inner join ia_project_association iapa on (ia.id = iapa.integrated_application_id)
                where ia.name = 'EventQ')""" % (db.boolValue(False), db.boolValue(False), '%'))

    except Exception as e:
       printLog("Error: Unable to execute sql query")
       printLog("Reason: %s" %(e))
       sys.exit(1)

def printLog (logMessage, forceStdOut=False):
    """Logs the output to the log file in append mode. Also prints the output to stdout if
       not running in silent mode"""

    global STDOUT

    config = app.util.getRawRuntimeConfiguration()
    logFile = os.path.join(config['RUNTIME_LOG_DIR'], "add_eventq_to_projects.log")

    timeStr = time.ctime(time.time())
    logOutput = "%s: %s" % (timeStr, logMessage)

    wFH = open(logFile, 'a')
    wFH.write("%s\n" % logOutput)
    wFH.close()

    if STDOUT or forceStdOut:
       print("%s" % logOutput)


def checkRunProgram():
    """ Checks whether the script can run in the current environment """

    # Check if script invoked by root user
    if (getpass.os.getuid() != 0):
       printUsageAndExit("This program should run as root user. Exiting!!\n", 1)
       sys.exit(1)

def printUsageAndExit(error=None, exit_code=0):
    scriptName = __file__
    print "\n Usage: %s [[--tfuser=<username>] [--tfpasswd=<password>]] [--silent]\n" % scriptName
    print '''
 When to run:
     This script need to be run after Teamforge 7.2 migration.
 What it does:
     Adds EventQ integrated application to existing projects.
 Prerequisites:
     * Should run as OS root user
     * Should run only in Teamforge App box
     * Require your Teamforge to be up and running

 Arguments:
     -u, --tfuser=    Specify teamforge site-admin username
     -p, --tfpasswd=  Specify password for teamforge user
     -h, --help       Prints this help information
     -s, --silent     Donot print progress message on standard-out
'''
    if (error is not None):
       print "\n ERROR MESSAGE:\t%s" % error
       sys.exit(1)
    else:
       sys.exit(0)


if __name__ == '__main__':
    main()

