#!/usr/bin/env python

import getopt
import os
import sys
import threading
import time

import app.util
import app.defaults
import app.landscape

# Global Parameters
cfg = app.util.getRuntimeConfiguration()
HOSTNAME = cfg['NODE_HOSTNAME']
dbType = cfg['DATABASE_TYPE']
logFile = os.path.join(cfg['RUNTIME_LOG_DIR'], "domain_change_db.log")
stdOut  = True
consoleOut = ''
oldDomain = ''
newDomain = ''
threadLimit = 5
threadLock = None

class DBThread (threading.Thread):

    def __init__(self, cfg, threadName, runQuery):

        self.cfg = cfg
        self.threadName = threadName
        self.runQuery = runQuery
        self.rowCount = 0

        printLog ("Thread for event [%s] is invoked."%(self.threadName))
        # DB Connection Initialisation
        self.dbType = self.cfg['DATABASE_TYPE']
        try:
            if self.dbType == 'postgresql':
                import pgdb
                self.dbConnection = pgdb.connect(user = self.cfg['DATABASE_USERNAME'],
                                   password = self.cfg['DATABASE_PASSWORD'],
                                   database = self.cfg['DATABASE_NAME'],
                                   host = self.cfg['ctfcore/dependency/ctfcore-database/jdbc/host'] + ':' + self.cfg['ctfcore/dependency/ctfcore-database/jdbc/port'])
            elif self.dbType == 'oracle':
                import cx_Oracle
                dsn = cx_Oracle.makedsn(self.cfg['ctfcore/dependency/ctfcore-database/jdbc/host'],
                                        self.cfg['ctfcore/dependency/ctfcore-database/jdbc/port'],
                                        self.cfg['DATABASE_NAME'])
                self.dbConnection = cx_Oracle.connect(u"%s/%s@%s" \
                              % (self.cfg['DATABASE_USERNAME'], self.cfg['DATABASE_PASSWORD'], dsn))
            self.cursor = self.dbConnection.cursor()
        except StandardError, e:
            printLog ("Unable to connect to %s database."%self.dbType)
            printLog ("Reason: %s"%e)
            printLog ("Please start your database service and then run this script!!")
            return None
        threading.Thread.__init__(self)

    def run(self):
        printLog ("Executing query %s"%self.runQuery)
        if self.dbType == 'postgresql':
            self.executePgsqlQuery()
        elif dbType == 'oracle':
            self.executeOracleQuery()

        self.rowCount = self.cursor.rowcount
        printLog ("Completed running event [%s]. Total records modified: %s"%(self.threadName, self.cursor.rowcount))
        self.dbConnection.commit()
        self.cursor.close()
        self.dbConnection.close()

        # Require this sleep because oracle server takes some time to release the connection.
        if self.dbType == 'oracle':
           time.sleep(3)

    def executeOracleQuery(self):
        return(self.cursor.execute(u"%s" % self.runQuery))

    def executePgsqlQuery(self):
        return(self.cursor.execute(self.runQuery))

def main():

    opts, args = getopt.getopt(sys.argv[1:], 'hso:n:t', ['help', 'silent', 'old=', 'new=', 'threadlimit='])
    global oldDomain
    global newDomain
    global threadLimit
    global stdOut
    global threadLock
    global consoleOut

    threadLock = threading.Lock()
    for opt, value in opts:
        if (opt in ('-h', '--help')):
            printUsageAndExit()
        elif (opt in ('-o', '--old')):
            oldDomain = value
        elif (opt in ('-n', '--new')):
            newDomain = value
        elif (opt in ('-l', '--threadlimit')):
            threadLimit = int(value)
            if (threadLimit > 5 and dbType == 'oracle'):
               threadLimit = 5
               printLog ("WARN: Maximum threadlimit for oracle db should not be greater than 5. So using default '5'")
        elif (opt in ('-s', '--silent')):
            stdOut = False
            consoleOut = '> /dev/null'
        else:
            printUsageAndExit('Unknown option : ' + opt )

    if not ( oldDomain and newDomain ):
       printUsageAndExit('Old & New Domain names are needed')

    printLog ("============ Begin Domain Change in Database ============")

    RUNTIME_DIR = cfg['RUNTIME_DIR']
    if not os.path.exists(RUNTIME_DIR):
       printLog ("Invalid Configuration: Runtime directory - %s is missing" %RUNTIME_DIR)
       sys.exit(1)

    validateDomains(oldDomain, newDomain)

    printLog ("Ensure collabnet services are shut down")
    os.system("%s/scripts/collabnet stop | tee -a %s %s" %(RUNTIME_DIR, logFile, consoleOut))

    printLog ("Ensure postgres is running")
    os.system("%s/scripts/collabnet start postgres | tee -a %s %s" %(RUNTIME_DIR, logFile, consoleOut))

    updateRecords(oldDomain, newDomain)

    #As per artf145098, there are some changes in reviewboard.properties from TF6.2 to TF7.0.
    #So for 7.0 it is advised not to run domain_change_rb script as a part of this script
    #This can be uncommented out in the post 7.0 releases.
    #if (doesRBExists()):
    #    p = Popen(['%s/scripts/domain_change_rb.py' % RUNTIME_DIR, oldDomain, newDomain], stdout = PIPE, stderr = PIPE)
    #    stdout, stderr = p.communicate()
    #    returncode = p.returncode
    #    if returncode != 0:
    #        if len(stderr) == 0:
    #            stderr = stdout
    #        printLog("ERROR: When running domain change on Review Board database: %s" % stderr)
    #        sys.exit(1)
    #    printLog(stdout)

    waitForThreadCompletion()
    printLog ("============ End Domain Change in Database ============")


def printUsageAndExit(error=None, exit_code=0):
    print "\n Usage: domain_change_db.py --old={old_domain_name} --new={new_domain_name} --threadlimit={max_thread_limit} \n"
    print '''
This script is designed to make necessary 'domain change' updates in the CTF database tables.
It will change the below database contents.
 ---------------------------------------------------------------------------------------------------------
|  table                      : column                                                                      |
 ---------------------------------------------------------------------------------------------------------
|  discussion_forum            - footer, reply_to
|  discussion_post             - content
|  document                    - description
|  document_review             - description
|  document_version            - version_comment, stored_file_url
|  external_system             - title, description
|  external_system_parameter   - value
|  folder                      - title, description
|  frs_file                    - description
|  integration_data_namespace  - name
|  integration_data_value      - value
|  item                        - title
|  linked_application          - application_url
|  mess_template               - description, body_template
|  news_post                   - body
|  ia_message_resource         - value
|  ia_config_param             - description
|  ia_config_param_option      - value
|  integrated_application      - description, base_url, go_url, end_point, admin_url
|  project                     - description
|  relationship                - forward_description, back_description
|  search_or_filter            - name,type
|  sfcomment                   - description
|  sfgroup                     - description
|  sfuser                      - detail, email, alternate_email_1, alternate_email_2, alternate_email_3
|  taskmgr_task                - description
|  wiki_page_version           - version_comment
|  artifact                    - description, text_0 to text_29
----------------------------------------------------------------------------------------------------------
'''
    if error is not None:
       print "\n\t ERROR MESSAGE\t\n%s"%error
    sys.exit(1)

## Validate the new domain is the one expected.
def validateDomains(oldDomain, newDomain):

    domain = None

    app.landscape.Landscape.initializeFromSiteOptionsPath(app.defaults.RUNTIME_OPTIONS_CONF)
    domain = app.landscape.Landscape.getSite().getLocalMachine().getPublicAddress().getName()

    if domain:
        if newDomain != domain:
            printLog ("\n New domain %s does not match with the configured domain %s\n" %(newDomain, domain))
            sys.exit(1)


def constructPgsqlQuery(columnName):
    global oldDomain
    query = " position('%s' in %s) > 0 " %(oldDomain,columnName)
    return query

def constructOracleQuery(columnName):
    global oldDomain
    query = " INSTR(%s,'%s') > 0 " %(columnName,oldDomain)
    return query

def isThreadPoolFree():
    global threadLimit
    threadCount = threading.activeCount()

    if threadCount <= threadLimit:
       return True
    else:
       return False

def waitForThreadCompletion():

    while (threading.activeCount() > 1):
        time.sleep(0.25)

def waitForThreadPool():

    while (not isThreadPoolFree()):
       time.sleep(0.25)

    return

def printLog (logMessage):
    """Logs the output to the log file in append mode. Also prints the output to stdout if
       not running in silent mode"""

    global stdOut
    global logFile

    timeStr = time.ctime(time.time())
    logOutput = "%s: %s"%(timeStr, logMessage)

    threadLock.acquire()
    wFH = open(logFile, 'a')
    wFH.write("%s\n"% logOutput)
    wFH.close()

    if stdOut:
       print("%s"% logOutput)
    threadLock.release()

def executeQuery(queryFunction, querySql):

    waitForThreadPool()
    queryThread = DBThread(cfg, queryFunction, querySql)
    queryThread.start()

    return queryThread

def updateRecords(oldDomain, newDomain):

    pgsql_regexp = "E'(^|\\\\s|\\\\W)" + oldDomain + "', E'\\\\1" + newDomain + "', 'g'"
    oracle_regexp = "'(^|\\s|\\W)" + oldDomain + "', '\\1" + newDomain + "'"
    if dbType == 'postgresql':
        regexpStr = pgsql_regexp
        positionStr = constructPgsqlQuery
    elif dbType == 'oracle':
        regexpStr = oracle_regexp
        positionStr = constructOracleQuery

    dsforumQuery = "update discussion_forum set reply_to = regexp_replace(reply_to, %s), footer = regexp_replace(footer, %s) where %s or %s" %(regexpStr, regexpStr, positionStr('reply_to'), positionStr('footer'))
    dsforumQuery_affected = executeQuery("Discussion Forum", dsforumQuery)

    dspostQuery = "update discussion_post set content = regexp_replace(content, %s) where %s " %(regexpStr, positionStr('content'))
    dspostQuery_affected = executeQuery("Discussion Post", dspostQuery)

    documentQuery = "update document set description = regexp_replace(description, %s) where %s " %(regexpStr, positionStr('description'))
    documentQuery_affected = executeQuery("Document", documentQuery)

    doc_reviewQuery = "update document_review set description = regexp_replace(description, %s) where %s " %(regexpStr, positionStr('description'))
    doc_reviewQuery_affected = executeQuery("Document Review", doc_reviewQuery)

    doc_versionQuery = "update document_version set version_comment = regexp_replace(version_comment, %s), stored_file_url = regexp_replace(stored_file_url, %s) where %s or %s " %(regexpStr, regexpStr, positionStr('version_comment'), positionStr('stored_file_url'))
    doc_versionQuery_affected = executeQuery("Document Version", doc_versionQuery)

    ext_systemQuery = "update external_system set title = regexp_replace(title, %s), description = regexp_replace(description, %s) where %s or %s " %(regexpStr, regexpStr, positionStr('title'), positionStr('description'))
    ext_systemQuery_affected = executeQuery("External System", ext_systemQuery)

    frs_fileQuery = "update frs_file set description = regexp_replace(description, %s) where %s " %(regexpStr, positionStr('description'))
    frs_fileQuery_affected = executeQuery("FRS File", frs_fileQuery)

    esparameterQuery = "update external_system_parameter set value = regexp_replace(value, %s) where %s " %(regexpStr, positionStr('value'))
    esparameterQuery_affected = executeQuery("External System Parameter", esparameterQuery)

    folderQuery = "update folder set title = regexp_replace(title, %s), description = regexp_replace(description, %s) where %s or %s " %(regexpStr, regexpStr, positionStr('title'), positionStr('description'))
    folderQuery_affected = executeQuery("Folder", folderQuery)

    int_data_namespaceQuery = "update integration_data_namespace set name = regexp_replace(name, %s) where %s " %(regexpStr, positionStr('name'))
    int_data_namespaceQuery_affected = executeQuery("Integration Data Namespace", int_data_namespaceQuery)

    int_data_valueQuery = "update integration_data_value set value = regexp_replace(value, %s) where %s " %(regexpStr, positionStr('value'))
    int_data_valueQuery_affected = executeQuery("Integration Data Value", int_data_valueQuery)

    itemQuery = "update item set title = regexp_replace(title, %s) where %s " %(regexpStr, positionStr('title'))
    itemQuery_affected = executeQuery("Item", itemQuery)

    lkapplicationQuery = "update linked_application set application_url = regexp_replace(application_url, %s) where %s " %(regexpStr, positionStr('application_url'))
    lkapplicationQuery_affected = executeQuery("Linked Application", lkapplicationQuery)

    mess_tempQuery = "update mess_template set body_template = regexp_replace(body_template, %s), description = regexp_replace(description, %s) where %s or %s " %(regexpStr, regexpStr, positionStr('body_template'), positionStr('description'))
    mess_tempQuery_affected = executeQuery("Mess Template", mess_tempQuery)

    news_postQuery = "update news_post set body = regexp_replace(body, %s) where %s " %(regexpStr, positionStr('body'))
    news_postQuery_affected = executeQuery("New Post", news_postQuery)

    plug_app_message_resourceQuery = "update ia_message_resource set value = regexp_replace(value, %s) where %s " %(regexpStr, positionStr('value'))
    plug_app_message_resourceQuery_affected = executeQuery("IA Message Resource", plug_app_message_resourceQuery)

    plug_config_paramQuery = "update ia_config_param set description = regexp_replace(description, %s) where %s "  %(regexpStr, positionStr('description'))
    plug_config_paramQuery_affected = executeQuery("IA Config Param", plug_config_paramQuery)

    plug_config_param_valueQuery = "update ia_config_param_option set value = regexp_replace(value, %s) where %s "  %(regexpStr, positionStr('value'))
    plug_config_param_valueQuery_affected = executeQuery("IA Config Param Options", plug_config_param_valueQuery)

    plcomponentQuery = "update integrated_application set base_url = regexp_replace(base_url, %s), go_url = regexp_replace(go_url, %s), end_point = regexp_replace(end_point, %s), admin_url = regexp_replace(admin_url, %s) where %s or %s or %s or %s "  %(regexpStr, regexpStr, regexpStr, regexpStr, positionStr('base_url'), positionStr('go_url'), positionStr('end_point'), positionStr('admin_url'))
    plcomponentQuery_affected = executeQuery("Integrated Application", plcomponentQuery)

    project_Query = "update project set description = regexp_replace(description, %s) where %s "  %(regexpStr, positionStr('description'))
    project_Query_affected = executeQuery("Project Set", project_Query)

    relationship_Query = "update relationship set forward_description = regexp_replace(forward_description, %s), back_description = regexp_replace(back_description, %s) where %s or %s"  %(regexpStr, regexpStr, positionStr('forward_description'), positionStr('back_description'))
    relationship_Query_affected = executeQuery("RelationShip", relationship_Query)

    search_or_filter_Query = "update search_or_filter set name = regexp_replace(name, %s), type = regexp_replace(type, %s) where %s or %s " %(regexpStr, regexpStr, positionStr('name'), positionStr('type'))
    search_or_filter_Query_affected = executeQuery("Search OR Filter", search_or_filter_Query)

    sfcomment_Query = "update sfcomment set description = regexp_replace(description, %s) where %s " %(regexpStr, positionStr('description'))
    sfcomment_Query_affected = executeQuery("SF Comment", sfcomment_Query)

    sfgroup_Query = "update sfgroup set description = regexp_replace(description, %s) where %s " %(regexpStr, positionStr('description'))
    sfgroup_Query_affected = executeQuery("SF Group", sfgroup_Query)

    taskmgr_task_Query = "update taskmgr_task set description = regexp_replace(description, %s) where %s "  %(regexpStr, positionStr('description'))
    taskmgr_task_Query_affected = executeQuery("TaskMgr Task", taskmgr_task_Query)

    wiki_page_version_Query = "update wiki_page_version set version_comment = regexp_replace(version_comment, %s) where %s "  %(regexpStr, positionStr('version_comment'))
    wiki_page_version_Query_affected = executeQuery("Wiki Page Version", wiki_page_version_Query)

    sfuserQuery = "update sfuser set detail = regexp_replace(detail, %s), email = regexp_replace(email, %s), alternate_email_1 = regexp_replace(alternate_email_1, %s), alternate_email_2 = regexp_replace(alternate_email_2, %s), alternate_email_3 = regexp_replace(alternate_email_3, %s)  where %s or %s or %s or %s or %s "  %(regexpStr, regexpStr, regexpStr, regexpStr, regexpStr, positionStr('detail'), positionStr('email'), positionStr('alternate_email_1'), positionStr('alternate_email_2'), positionStr('alternate_email_3'))
    sfuserQuery_affected = executeQuery("SF User", sfuserQuery)

    artifact_1_Query = "update artifact set description = regexp_replace(description, %s) where %s " %(regexpStr, positionStr('description'))
    artifact_1_Query_affected = executeQuery("Artifact 1 Query", artifact_1_Query)

    for x in range(0,29):
        textStr = "text_%s" %x
        artifact_loop_Query = "update artifact set %s= regexp_replace(%s, %s) where %s " %(textStr, textStr, regexpStr, positionStr(textStr))
        artifact_loop_Query_affected = executeQuery("Artifact %s Query"%textStr, artifact_loop_Query)

# Check if Review Board is available
def doesRBExists():
    return execute_ia_query('Review Board')

def execute_ia_query(name):
    query = "select name from integrated_application where name = '%s'" % name
    thrObj = executeQuery("%s Check" % name, query)
    while (thrObj.is_alive()):
      time.sleep(0.25)
    return thrObj.rowCount


if __name__ == '__main__':
    main()
