#!/usr/bin/env python

import os
import os.path
import sys
import shutil
import getopt
import re

# Helper Class that walks through a directory and returns a list of all files that match
# the required pattern
class GlobDirectoryWalker:

    def __init__(self, directory, pattern="*"):
        self.stack = [directory]
        self.pattern = pattern
        self.files = []
        self.index = 0

    def __getitem__(self, index):
        while 1:
            try:
                file = self.files[self.index]
                self.index = self.index + 1
            except IndexError:
                # pop next directory from stack
                self.directory = self.stack.pop()
                self.files = os.listdir(self.directory)
                self.index = 0
            else:
                # got a filename
                fullname = os.path.join(self.directory, file)
                if os.path.isdir(fullname) and not os.path.islink(fullname) and not file == '.svn':
                    self.stack.append(fullname)
                if re.match(self.pattern, file):
                    return fullname

# Sync class copies resource bundle files from a source directory to the target directory. This can be
# used to copy files from forge to partnerproject and viceversa.
# TODO: Doesnt take care of Help files for now
class Sync:
    def __init__(self, srcDir, targetDir, parentDirForFiles='./src/java/com/vasoftware/sf/i18n', language = 'en'):
        self.srcDir = srcDir
        self.targetDir = targetDir
        self.parentDirForFiles = parentDirForFiles
        self.language = language
        self.patterns = { 'en': '(application|common).properties',
                          'ja': '(application|common)_ja.properties(\.eucjp)?',
                          'ko' : '(application|common)_ko.properties(\.euckr)?',
                          'zh_CN': '(application|common)_zh_CN.properties(\.big5)?' }

    def validateParams(self):
        msg = None
        if not os.path.isdir(self.srcDir):
            msg = 'Source Directory not found'
        elif not os.path.isdir(self.targetDir):
            msg = 'Target Directory not found'
        elif not self.patterns.has_key(self.language):
            msg = 'Language not supported'

        return msg

    def copyFilesToTargetDir(self):
        oldDir = os.getcwd()
        os.chdir(self.srcDir)
        globWalker = GlobDirectoryWalker(self.parentDirForFiles, self.patterns[self.language])
        for file in globWalker:
            dirName = os.path.dirname(file)
            targetBundleDir = os.path.join(self.targetDir, dirName)
            if not os.path.exists(targetBundleDir):
                os.makedirs(targetBundleDir)
            shutil.copy(file, os.path.join(targetBundleDir))

        os.chdir(oldDir)

def usage(msg = None):
    if msg:
        stream = sys.stderr
    else:
        stream = sys.stdout

    stream.write("\n"
                 "Usage: %s --srcDir=<Source Directory> --destDir=<Target Directory> [--prefixPath=<Path for Search>] [--language=<en|ja|ko|zh_CN] [-h] \n"
                            "Checkout both the source and destination directories. Point them using the --srcDir and --destDir. \n"
                            "Prefixpath indicates the base path from where the search for bundle files need to be done. \n"
                            "Language indicates the language for which the sync needs to be done. Defaults to english \n"
                            %(sys.argv[0],))
    if msg:
        stream.write("\nERROR: %s\n" % msg)
        exit_code = 1
    else:
        exit_code = 0

    sys.exit(exit_code)

if __name__ == '__main__':
    if len(sys.argv) < 2:
        usage('Invalid usage')

    try:
        opts,args = getopt.getopt(sys.argv[1:], "hs:d:p:l:", ["help","srcDir=","destDir=","prefixPath=","language="])
    except getopt.GetoptError:
        usage('Invalid usage')

    srcDir = destDir = None
    prefixPath = './src/java/com/vasoftware/sf/i18n'
    language = 'en'

    for opt,val in opts:
        if opt in ('--help', '-h'):
            usage()
        elif opt in ('--srcDir', '-s'):
            srcDir = val
        elif opt in ('--destDir', '-d'):
            destDir = val
        elif opt in ('--prefixPath', '-p'):
            prefixPath = val
        elif opt in ('--language', '-l'):
            language = val
        else:
            usage('Invalid usage')

    sync = Sync(srcDir, destDir, prefixPath, language)
    errMsg = sync.validateParams()
    if errMsg:
        usage(errMsg)

    sync.copyFilesToTargetDir()
    sys.stdout.write("Synched Files for the %s language\n" %(language))

