#!/usr/bin/env python
"""ETL client for CollabNet TeamForge
"""
__license__ = "Commercial/Restrictive"
__version__ = "2.0"

from suds.client import Client
from suds.cache import ObjectCache
import sys
import os
import getopt

import app.util

# ETL service
verbose = False
etl_service = None
check_status = False
check_status_all = False
jobName = None
cfg = app.util.getRuntimeConfiguration()

def main():
    global verbose
    global site_url
    global check_status

    opts = ""
    etl_port = cfg.get("etl/bind/http-soap/port")
    site_url = "http://localhost:%s" % etl_port
    _handleParams()

    #set status flag as there is no other operation available now
    if not check_status:
        check_status = True

    # begin main program logic
    _setupServices()
    if jobName:
        _triggerJob()
    elif check_status_all:
        _status(None)
    else:
        _status('HISTORICAL_LOAD')
    # end main program logic

    # completed successfully
    _exit(0)

def usage(msg):
    """ Prints the tools usage to stdout
    """
    print "\nNAME"
    print "    etl-client.py - Client script to access ETL services. "
    print "\nSYNOPSIS"
    print "    etl-client.py [--status|-s]"
    print "    etl-client.py [--status-all|-a]"
    print "    etl-client.py [--help|-h]"
    print "    etl-client.py [--run=|-r]"
    print "\nDESCRIPTION"
    print "    etl-client.py allows you to access ETL scheduler and allows to check the "
    print "    status of the jobs configured. Also supports triggering jobs manually."
    print ""
    print "\n  Options"
    print "    Not providing any switches to this script will assume that you want "
    print "    to check the status of all the jobs."
    print ""
    print "    --status , -s"
    print "      Prints the status of Incremental jobs configured in ETL service."
    print ""
    print "    --status-all , -a"
    print "      Prints the status of Incremental and Historical jobs configured in ETL service."
    print ""
    print "    --verbose , -v"
    print "      Chronicles the process of requested operation a bit more."
    print ""
    print "    --run= , -r"
    print "      triggers a job manually for the given job."
    print ""
    print "OUTPUT"
    print "    Depending on the nature of your request, you may see a few lines "
    print "    or many lines outputted by this script."
    print ""

    _exit(0)

def _setupServices():
    """ Establishes connections to the CollabNet Web Services.
    """
    global etl_service

    etlwsdlfile = site_url + '/etl/services/ETLApp?wsdl'
    #etlwsdlfile = site_url + '/etl/services/Mock?wsdl'
    sudsCacheDir = os.path.join("%s" %cfg.get('TEMPORARY_DIR'), "suds")

    try:
        sudsCache = ObjectCache(sudsCacheDir)
        etl_service = Client(etlwsdlfile, cache=sudsCache, location=etlwsdlfile)
    except:
        print "Unable to grab the WSDL: %s\n" % etlwsdlfile
        print "Please ensure that the hostname is entered correctly, and "
        print "that you are accessing the SOAP services directly and not "
        print "through apache."
        print
        print "For help, please visit the CSFE Forum at:"
        print " http://forums.open.collab.net/ds/viewForumSummary.do?dsForumId=738"
        print " or open a support case via your usual channels."
        print
        _exit(1)

def _status(groupName):

    if verbose:
        print "Invoking web service : getStatusOfJobs"

    shared_secret = cfg.get("ETL_SOAP_SHARED_SECRET")
    try:
        job_status_array = etl_service.service.getStatusOfJobs(shared_secret, groupName)
    except:
        print
        print "Unable to invoke ETL job status service"
        print
        print "For help, please open a support case via your usual channels."
        print
        _exit(1)
    print ""
    for job_status in job_status_array:
        print job_status
    print ""

    if verbose:
        print "Completed web service invocation."

def _triggerJob():
    shared_secret = cfg.get("ETL_SOAP_SHARED_SECRET")
    try:
        job_start_status = etl_service.service.triggerJob(shared_secret, jobName)
    except:
        _exit(1)
    print job_start_status

def _handleParams():
    global jobName
    global check_status_all
    try:
        opts, cmd_args = getopt.getopt(
            sys.argv[1:], 'hsavr:',
                          ['help',
                           'status',
                           'status-all',
                           'verbose',
                           'run='])
    except getopt.GetoptError, e:
        usage('ERROR: %s' % e.msg)

    for opt, value in opts:
        if (opt in ('-s', '--status')):
            check_status = True
        elif (opt in ('-a', '--status-all')):
            check_status_all = True
        elif (opt in ('-h','--help')):
            usage('')
        elif (opt in ('-v', '--verbose')):
            verbose = True
        elif (opt in ('-r', '--run')):
            jobName = value
        else:
            print value

def _exit(code):
    """ Performs proper clean up before exiting the program
    """
    sys.exit(code)

if __name__ == "__main__":
    main()


