﻿import os
import glob
import time

import app.servicecontrol
import app.defaults
import app.health
import app.threaddump

START_TIMEOUT = 30

class ServiceController(app.servicecontrol.ServiceController):
    def __init__(self, service, logger):
        super(ServiceController, self).__init__(service, logger)

        self.__lastStartupTime = 0

    def getPidFile(self):
        return os.path.join(app.defaults.DEFAULT_RUN_DIR, "jboss.pid")


    def __runInit(self, verb):
        config = app.util.getRuntimeConfiguration()

        logFile = os.path.join(config['APPLICATION_LOG_DIR'], 'service.log')
        initScript = os.path.join(app.defaults.DEFAULT_SERVICE_DIR, "jboss", "bin", "init-jboss.sh")

        return os.system("%s %s < /dev/null >> %s 2>&1 &" % (initScript, verb, logFile)) == 0

    def __areDeploymentMarkerFilesPresent(self, marker):
        config = app.util.getRuntimeConfiguration()

        markerFiles = glob.glob("%s/*.%s" % (config['JBOSS_SERVER_DEPLOY_DIR'], marker))

        return markerFiles and len(markerFiles) > 0

    def isStarting(self):
        if (time.time() - self.__lastStartupTime) < START_TIMEOUT:
            # JBoss is pretty quick to open ports - it takes a a bit longer for it to publish its pid file and
            # a pretty long time to deploy all the apps though. However, there is a delay between (a) ports
            # are opened and (b) deployment markers shot up.
            self.logger.info("Jboss startup was initialized less than %d secs ago - assuming it is still starting" % START_TIMEOUT)
            return True

        if not self.getPid():
            return False

        return self.__areDeploymentMarkerFilesPresent("isdeploying")

    def isDoomed(self):
        # Once one application has failed to deploy, things are over.
        return self.__areDeploymentMarkerFilesPresent("failed")

    def startDaemonAsync(self):
        if self.__runInit("start"):
            self.__lastStartupTime = time.time()
            return True
        else:
            return False

    def stopDaemonAsync(self):
        return self.__runInit("stop")

    def printSnapshot(self, label):
        config = app.util.getRuntimeConfiguration()
        jbossFile = os.path.join(config['APPLICATION_LOG_DIR'], 'jboss_console.log')
        jbossPid = self.getPid()
        app.threaddump.ThreadDump.dumpJVMthread(jbossPid, self.logger, jbossFile, "jboss", label)
        if jbossPid is not None:
            print "jboss JVM threads dumped to %s." %jbossFile
            print "Find thread dump between lines from log demarked by - %s"%label

    def isHealthy(self):
        #
        # We directly invoke the JBoss endpoint (not via Apache). This is okay
        # as Apache has its own health check. Also, it avoids trouble with self-
        # signed SSL certificates.
        #
        jbossHttpEndpoint = self.getService().getProvidedEndpoint("http")

        url = "http://localhost:%d/ctfrest/foundation/v1/server/health?verbose=true&thorough=false" % \
            (jbossHttpEndpoint.getPort())

        healthy = app.health.HealthCheck("Jboss", url, self.logger).invoke()

        return healthy
    
    def getLogPaths(self):
        return [os.path.join(app.util.getRuntimeConfiguration().get("APPLICATION_LOG_DIR"), "*")]
