﻿import os
import glob
import time
import signal
import sys

import app.servicecontrol
import app.util
import app.defaults

START_TIMEOUT = 30

class ServiceController(app.servicecontrol.ServiceController):
    def __init__(self, service, logger):
        super(ServiceController, self).__init__(service, logger)

        self.__lastStartupTime = 0

    def getPidFile(self):
        port = self.getService().getProvidedEndpoint("jdbc").getPort()

        # Note: If there are multiple Postgres instances running in parallel,
        # we need to take the PID that corresponds to the instance's port number.

        pidFiles = glob.glob("/var/run/postmaster*.%d.pid" % port)
        if pidFiles:
            return pidFiles[0]
        else:
            return None

    def getKillSignal(self):
        # Postgres really, really does not like SIGKILLs - it will leave
        # child processes behind and may corrupt data. Therefore, be
        # a little more gentle.
        return signal.SIGQUIT

    def __isRunningUncontrolledWithoutDeployment(self):
        # If somebody wiped the runtime without stopping Postgres first,
        # we are in an odd state.
        return self.isProcessRunning() and \
            not super(ServiceController, self).isDeployed() and \
            not os.path.isfile(app.defaults.RUNTIME_OPTIONS_CONF)

    def isDeployed(self):
        if self.__isRunningUncontrolledWithoutDeployment():
            return True
        else:
            return super(ServiceController, self).isDeployed()

    def isDoomed(self):
        if self.__isRunningUncontrolledWithoutDeployment():
            return True
        else:
            return super(ServiceController, self).isDoomed()

    def canBootstrap(self):
        # Bootstrapping is a no-op, so allow it always.
        return True

    def isStarting(self):
        # Postgres does not immediately publish its PID file - but that does not mean
        # it is in trouble. So we need to give it some extra time before making any status
        # decisions.
        return not self.getPid() and (time.time() - self.__lastStartupTime) < START_TIMEOUT

    def __runPostgres(self, verb):
        return app.util.runSystem("%s %s &> /dev/null" % (
                os.path.join(app.defaults.DEFAULT_RUNTIME_DIR, "scripts", "postgresql"),
                verb),
            self.logger,
            skip_error=True)

    def startDaemonAsync(self):
        if self.__runPostgres("start"):
            self.__lastStartupTime = time.time()
            return True
        else:
            return False

    def stopDaemonAsync(self):
        return self.__runPostgres("stop")

    def createPostDeployAction(self):
        # Start the service right after deployment
        return self.createStartAction()

    def printSnapshot(self, label):
        PG_STAT_SQLS_TF = {
                            'pg_stat_activity': "select * from pg_catalog.pg_stat_activity",
                            'pg_stat_database': "select * from pg_catalog.pg_stat_database",
                            'pg_locks': "select * from pg_catalog.pg_locks",
                            'pg_locks_stats'  : "select bl.pid as blocked_pid, bl.mode as blocked_mode, a.usename as blocked_user, ka.query as blocking_statement, now() - ka.query_start as blocking_duration, kl.pid as blocking_pid, kl.mode as blocking_mode, ka.usename as blocking_user, a.query as blocked_statement, now() - a.query_start as blocked_duration from pg_catalog.pg_locks bl join pg_catalog.pg_stat_activity a on bl.pid = a.pid join pg_catalog.pg_locks kl join pg_catalog.pg_stat_activity ka on kl.pid = ka.pid on bl.transactionid = kl.transactionid and bl.pid != kl.pid where not bl.granted"
                          }

        for sqlKey in PG_STAT_SQLS_TF:
            cmdStr = "su postgres -c \"psql -c '%s'\"" % PG_STAT_SQLS_TF[sqlKey]
            try:
                print "-------------------------------------------------------------"
                print "Start of database thread dump of [%s]" %sqlKey
                print "-------------------------------------------------------------"
                sys.stdout.write(str(app.util.executeCommand(cmdStr)) + "\n")
            except:
                self.logger.error("Command failed %s" %cmdStr)
                
    def getLogPaths(self):
        return [
            os.path.join(app.defaults.NEW_PGCLUSTER, "pgstartup.log"),
            os.path.join(app.util.getRuntimeConfiguration().get("PGSQL_LOG_DIR"), "*")
            ]

    def getLocalContext(self):
        return ['tf_postgresql_rw_t', 'tf_postgresql_exec_t']
