#!/usr/bin/env python
# CollabNet TeamForge re-indexing utility.
"""Search re-indexing script for CollabNet TeamForge 5
"""
__license__ = "Commercial/Restrictive"
__version__ = "2.1"

from suds.client import Client
from suds.cache import ObjectCache
import getpass as gp
import sys
import os
import time
import getopt
import datetime

import app.util

# globals
dry_run = False
verbose = False
artifacts_only = False
trackers_only = False
documents_only = False
document_folders_only = False
posts_only = False
forums_only = False
topics_only = False
wikis_only = False
news_only = False
packages_only = False
releases_only = False
frs_files_only = False
commits_only = False
project_pages_only = False


force_index = False
verify_indexed = False
item_id = ''
loginhash = None

# CollabNet SOAP Services
collabnet_service = None
tracker_service = None
discussion_service = None
document_service = None
frs_servce = None
wiki_service = None
scm_service = None
news_service = None
project_pages_service = None

# statics
ARTIFACT_TYPE = "artifact"
TRACKER_TYPE = "tracker"
DOCUMENT_FOLDER_TYPE = "document_folder"
DOCUMENT_TYPE = "document"
POST_TYPE = "post"
TOPIC_TYPE = "topic"
FORUM_TYPE = "forum"
WIKI_TYPE = "wiki"
RELEASE_TYPE = "release"
PACKAGE_TYPE = "package"
FRS_FILE_TYPE = "frs_file"
SCM_COMMIT_TYPE = "commit"
NEWS_POST_TYPE = "news"
PROJECT_PAGE_TYPE = "page"
SUPPORTED_TYPES = [ARTIFACT_TYPE, TRACKER_TYPE, DOCUMENT_TYPE, DOCUMENT_FOLDER_TYPE, POST_TYPE, TOPIC_TYPE,
FORUM_TYPE, WIKI_TYPE, RELEASE_TYPE, PACKAGE_TYPE, FRS_FILE_TYPE, SCM_COMMIT_TYPE, NEWS_POST_TYPE,
PROJECT_PAGE_TYPE]
LOG_INFO = "[INFO]"
LOG_WARN = "[WARN]"
LOG_SUCCESS = "[SUCCESS]"
LOG_ATTENTION = "[ATTENTION]"
LOG_ERROR = "[ERROR]"
LOG_DEBUG = "[DEBUG]"

# The following MUST, MUST, MUST be a floating point value!
REQUESTS_PER_SECOND_THROTTLE = 2.0

def main():
    global verbose
    global item_id
    global site_url
    global dry_run
    global artifacts_only
    global documents_only
    global posts_only
    global topics_only
    global forums_only
    global news_only
    global commits_only
    global releases_only
    global packages_only
    global frs_files_only
    global project_pages_only
    global wikis_only
    global trackers_only
    global document_folders_only
    global force_index
    global verify_indexed
    global log_file
    global project_id

    opts = ""
    item_id = ""
    project_id = None
    log_file = None
    site_url = "http://localhost:8080"

    try:
        opts, cmd_args = getopt.getopt(
            sys.argv[1:], 'hradfxvi:o:p:',
                          ['help',
                           'dry-run',
                           'artifacts-only',
                           'documents-only',
                           'posts-only',
                           'trackers-only',
                           'document_folders-only',
                           'releases-only',
                           'packages-only',
                           'forums-only',
                           'topics-only',
                           'frs_files-only',
                           'project_pages-only',
                           'commits-only',
                           'news-only',
                           'wikis-only',
                           'force-index',
                           'verify',
                           'single-item=',
                           'output-file=',
                           'project-id=',
                           'verbose'])
    except getopt.GetoptError, e:
        usage('ERROR: %s' % e.msg)

    for opt, value in opts:
        if (opt in ('-h','--help')):
            usage('')
        elif (opt in ('-r', '--dry-run')):
            dry_run = True
        elif (opt == '--posts-only'):
            posts_only = True
        elif (opt in ('-d', '--documents-only')):
            documents_only = True
        elif (opt in ('-a', '--artifacts-only')):
            artifacts_only = True
        elif (opt == '--trackers-only'):
            trackers_only = True
        elif (opt == '--document_folders-only'):
            document_folders_only = True
        elif (opt == '--commits-only'):
            commits_only = True
        elif (opt == '--news-only'):
            news_only = True
        elif (opt == '--topics-only'):
            topics_only = True
        elif (opt == '--forums-only'):
            forums_only = True
        elif (opt == '--project_pages-only'):
            project_pages_only = True
        elif (opt == '--wikis-only'):
            wikis_only = True
        elif (opt == '--packages-only'):
            packages_only = True
        elif (opt == '--releases-only'):
            releases_only = True
        elif (opt == '--frs_files-only'):
            frs_files_only = True

        elif (opt in ('--force-index', '-f')):
            force_index = True
        elif (opt in ('-x', '--verify')):
            verify_indexed = True
        elif (opt in ('-o', '--output-file')):
            log_file = open(str(value), 'a')
            log_file.write("STARTING at %s\n" % datetime.datetime.now())
        elif (opt in ('-v', '--verbose')):
            verbose = True
        elif (opt in ('-i', '--single-item')):
            item_id = str(value)
        elif (opt in ('-p','--project-id')):
            project_id = str(value)
        else:
            print value

    # begin main program logic
    _printWorkProposal()

    if not item_id and not artifacts_only and not documents_only and not posts_only and not trackers_only \
    and not document_folders_only and not topics_only and not forums_only and not wikis_only and not news_only \
    and not releases_only and not packages_only and not project_pages_only and not commits_only and not frs_files_only:
        _promptForFullReindex()

    _setupServices()

    if item_id:
        if not _reIndexItemByID(item_id):
            _exit(2)
    elif artifacts_only or documents_only or posts_only or trackers_only \
    or document_folders_only or topics_only or forums_only or wikis_only or news_only \
    or releases_only or packages_only or project_pages_only or commits_only or frs_files_only:
        if artifacts_only:
            _processItems(ARTIFACT_TYPE, force_index)
        if documents_only:
            _processItems(DOCUMENT_TYPE, force_index)
        if posts_only:
             _processItems(POST_TYPE, force_index)
        if trackers_only:
            _processItems(TRACKER_TYPE, True)
        if document_folders_only:
            _processItems(DOCUMENT_FOLDER_TYPE, True)
        if forums_only:
             _processItems(FORUM_TYPE, True)
        if topics_only:
            _processItems(TOPIC_TYPE, True)
        if commits_only:
            _processItems(SCM_COMMIT_TYPE, True)
        if news_only:
             _processItems(NEWS_POST_TYPE, force_index)
        if wikis_only:
            _processItems(WIKI_TYPE, True)
        if frs_files_only:
            _processItems(FRS_FILE_TYPE, True)
        if project_pages_only:
             _processItems(PROJECT_PAGE_TYPE, True)
        if packages_only:
            _processItems(PACKAGE_TYPE, True)
        if releases_only:
             _processItems(RELEASE_TYPE, True)
    else:
        _reIndexSite()
    # end main program logic

    # completed successfully
    _exit(0)

def _promptForFullReindex():
    """ Verifies that the user in fact wants to re-index the entire site
    """
    prompt = "\nRequesting that the entire site be re-indexed can introduce " \
             "a substantial amount of load on the server. Every item of every " \
             "type will be scheduled for re-indexing. Would you like to re-index " \
             "all TeamForge data? [y/n]:"
    while True:
        ans = raw_input(prompt)
        if ans.lower() not in ['y', 'n', 'no', 'yes']:
            print 'please enter y or n.'
            continue
        if ans.lower() in ['y','yes']:
            break
        if ans.lower() in ['n','no']:
            _exit(0)

    print ""

def _printWorkProposal():
    """ A quick summary of what we're about to attempt, given the parameters
    """
    print "\nYou are about to do the following:\n"

    # only want a single item reindexed
    if item_id:
        if dry_run:
            print "  THERE IS NO DRY-RUN MODE FOR THIS FUNCTION!"
        print "  Re-index a single item: %s" % item_id

        if verify_indexed:
            print "  Will verify item is searchable after re-indexing."
        else:
            print "  Will NOT verify item is searchable after re-indexing."
        if artifacts_only:
           print "    Skipping --artifacts-only option"
        if documents_only:
           print "    Skipping --documents-only option"
        if posts_only:
           print "    Skipping --posts-only option"
        if trackers_only:
           print "    Skipping --trackers-only option"
        if document_folders_only:
           print "    Skipping --document_folders-only option"
        if topics_only:
           print "    Skipping --topics-only option"
        if forums_only:
           print "    Skipping --forums-only option"
        if releases_only:
           print "    Skipping --releases-only option"
        if packages_only:
           print "    Skipping --packages-only option"
        if news_only:
           print "    Skipping --news-only option"
        if frs_files_only:
           print "    Skipping --frs_files-only option"
        if project_pages_only:
           print "    Skipping --project_pages-only option"
        if commits_only:
           print "    Skipping --commits-only option"
        if wikis_only:
           print "    Skipping --wikis-only option"

        print ""

        # DONE. That's all they need to know
        return

    # artifacts and/or documents and/or posts and/or trackers and/or document folders and/or forums and/or topics
    # and/or commits and/or news and/or wikis and/or project pages and/or packages and/or releases and/or frs files
    #only mode
    if artifacts_only or documents_only or posts_only or trackers_only or document_folders_only or \
    topics_only or forums_only or packages_only or frs_files_only or commits_only or wikis_only or \
    news_only or project_pages_only or releases_only :
        # warn if running in dry run mode
        if dry_run:
            print "\n**** THIS IS JUST A DRY-RUN ****"

        if artifacts_only:
            if force_index:
              print "Re-indexing all artifacts."
            else:
              print "  Re-index un-searchable artifacts."
        if documents_only:
            if force_index:
              print "Re-indexing all documents."
            else:
              print "  Re-index un-searchable documents."
        if project_id is not None:
            print "  Limit re-indexing to project: %s" % project_id
        if posts_only:
            if force_index:
              print "Re-indexing all posts."
            else:
              print "  Re-index un-searchable posts."
        if trackers_only:
              print "Re-indexing all trackers."
        if document_folders_only:
              print "Re-indexing all document folders."
        if topics_only:
              print "Re-indexing all topics."
        if forums_only:
              print "Re-indexing all forums."
        if packages_only:
              print "Re-indexing all packages."
        if frs_files_only:
              print "Re-indexing all frs files."
        if commits_only:
              print "Re-indexing all commits."
        if news_only:
              if force_index:
                print "Re-indexing all news posts."
              else:
                print "  Re-index un-searchable news posts."
        if releases_only:
              print "Re-indexing all releases."
        if wikis_only:
              print "Re-indexing all wikis."
        if project_pages_only:
              print "Re-indexing all project pages."


        if verify_indexed and not dry_run:
            print "  Will verify items are searchable after re-indexing."
        else:
            print "  Will NOT verify items are searchable after re-indexing."
        print ""

        # DONE. That's all they need to know
        return

    # warn if running in dry run mode
    if not project_id:
        if dry_run:
            print "  THERE IS NO DRY-RUN MODE FOR THIS FUNCTION!"
        print "  Re-index all of your TeamForge data"
    else:
        print "Option (--project-id) is only supported in artifacts-only, " \
              "documents-only, posts-only trackers_only, document_folders_only " \
              "topics_only, forums_only, packages_only, frs_files_only, commits_only, "\
               "or wikis_only, news_only, project_pages_only, releases_only mode."
        print ""
        print "If you want to re-index a single project use the (-i|--single-item) switch)."
        _exit(1)

def usage(msg):
    """ Prints the tools usage to stdout
    """
    print "\nNAME"
    print "    SearchReindex.py - Schedule CollabNet TeamForge data to be re-indexed " \
          "for searching."
    print "\nSYNOPSIS"
    print "    SearchReindex.py [--artifacts-only|-a] [--documents-only|-d] [--posts-only]" \
          "[--topics-only] [--forums-only] [--trackers-only]" \
          "[--wikis-only] [--news-only] [--project_pages-only]" \
          "[--commits-only] [--file_releases-only] [--packages-only]" \
          "[--releases-only] [--file_releases-only]" \
          "[--force-index|-f]"\
          "[--verify|-x] [--project-id projectId|-p projectId] [--dry-run|-r] [--output-file " \
          "filePath|-o filePath]"
    print "    SearchReindex.py [--single-item|-i itemId] [--verify|-x] [--output-file|-o]"
    print "    SearchReindex.py [--help|-h]"
    print "\nDESCRIPTION"
    print "    SearchReindex.py allows you to either re-index your entire CollabNet "
    print "    TeamForge data or do so selectively. Note that the main function of this "
    print "    tool is to schedule re-index events with the server. It's up to the TeamForge "
    print "    server to process those events in due time."
    print ""
    print "    If you select to re-index only a subset of data types, this script will "
    print "    first check if each item is already searchable before scheduling an "
    print "    unnecessary re-index event. You can force reindexing using the option force-index"
    print "\n  Options"
    print "    Not providing any switches to this script will assume that you want "
    print "    all of your data re-indexed. It will prompt you for confirmation prior "
    print "    to doing so."
    print ""
    print "    --single-item itemId, -i itemId"
    print "      Schedules a re-index for just the given item. If the item id is for a project "
    print "      the scheduling results in the server re-indexing all of the project data."
    print ""
    print "    --artifacts-only , -a"
    print "      Re-index all artifacts on the site that are currently not searchable or all artifacts if option f "
    print "      is selected."
    print ""
    print "    --documents-only , -d"
    print "      Re-index all documents on the site that are currently not searchable or all documents if option f "
    print "      is selected."
    print ""
    print "    ---posts-only"
    print "      Re-index all posts on the site that are currently not searchable or all posts if option f "
    print "      is selected."
    print ""
    print "    ---trackers-only"
    print "      Re-index all trackers on the site."
    print ""
    print "    ---document_folders-only"
    print "      Re-index all document folders."
    print ""
    print "    ---topics-only"
    print "      Re-index all topics."
    print ""
    print "    ---forums-only"
    print "      Re-index all forums on the site."
    print ""
    print "    ---news-only"
    print "      Re-index all news."
    print ""
    print "    ---project_pages-only"
    print "      Re-index all project pages."
    print ""
    print "    ---packages"
    print "      Re-index all packages on the site."
    print ""
    print "    ---commits-only"
    print "      Re-index all commits."
    print ""
    print "    ---frs_files-only"
    print "      Re-index all frs files."
    print ""
    print "    ---releases-only"
    print "      Re-index all releases on the site."
    print ""
    print "    ---wikis-only"
    print "      Re-index all wikis on the site."
    print ""
    print "    ---force-index , -f"
    print "      Force indexing (doesn't check if item is searchable already)."
    print ""
    print "    --project-id projectId, -p projectId"
    print ""
    print "      Limit the re-indexing to data for single project when re-indexing only "
    print "      artifacts and/or documents."
    print ""
    print "    --verify , -x"
    print "      Searches for each item that is scheduled for re-indexing. There is a "
    print "      one minute wait limit for each item to be re-indexed by the server. "
    print "      IMPORTANT: although this is functionally convenient this may introduce a "
    print "      very significant delay in scheduling a large number of your items to be "
    print "      re-indexed. The delay is subject to the amount of items being re-indexed, "
    print "      server's load and pre-existing search indexing queue size."
    print ""
    print "    --dry-run , -r"
    print "      Executes all of the steps for scheduling a re-index without actually sending "
    print "      any re-index requests to the server. This provides a list of items that are "
    print "      found to be requiring re-indexing."
    print ""
    print "    --output-file filePath, -o filePath"
    print "      Prints output to the given file."
    print ""
    print "    --verbose , -v"
    print "      Chronicles the process of scheduling the re-index a bit more."
    print ""
    print "OUTPUT"
    print "    Depending on the nature of your re-indexing request, you may see a few lines "
    print "    or many lines outputted by this script. Each line has a prefix marker. Here is "
    print "    the key for understanding the output:"
    print ""
    print "    [SUCCESS] - as the name suggests, which ever item ID included in this line "
    print "    represents an item that is searchable on the site."
    print ""
    print "    [ATTENTION] - This requires your attention. The associated item id was not"
    print "    found to be searchable. Now, given that this script provides an option to"
    print "    follow up with a re-index and search again feature, you want to pay close"
    print "    attention to the output that follows this line. The item may or may not "
    print "    become searchable after a follow-up re-index is scheduled."
    print ""
    print "    [INFO], [ERROR], [DEBUG] - these are simply informative."

    _exit(0)

def _setupServices():
    """ Establishes connections to the CollabNet Web Services.
    """
    global collabnet_service
    global tracker_service
    global document_service
    global discussion_service
    global wiki_service
    global frs_service
    global news_service
    global scm_service
    global page_service

    sitewsdlfile = site_url + '/ce-soap60/services/CollabNet?wsdl'
    trackerwsdlfile = site_url + '/ce-soap60/services/TrackerApp?wsdl'
    documentwsdlfile = site_url + '/ce-soap60/services/DocumentApp?wsdl'
    discussionwsdlfile = site_url + '/ce-soap60/services/DiscussionApp?wsdl'
    scmwsdlfile = site_url + '/ce-soap60/services/ScmApp?wsdl'
    wikiwsdlfile = site_url + '/ce-soap60/services/WikiApp?wsdl'
    frswsdlfile = site_url + '/ce-soap60/services/FrsApp?wsdl'
    pagewsdlfile = site_url + '/ce-soap60/services/PageApp?wsdl'
    newswsdlfile = site_url + '/ce-soap60/services/NewsApp?wsdl'

    cfg = app.util.getRuntimeConfiguration()
    sudsCacheDir = os.path.join("%s" %cfg.get('TEMPORARY_DIR'), "suds")
    sudsCache = ObjectCache(sudsCacheDir)

    try:
        collabnet_service = Client(sitewsdlfile, cache=sudsCache, location=sitewsdlfile)
    except:
        print "Unable to grab the WSDL: %s\n" % sitewsdlfile
        print "Please ensure that the hostname is entered correctly, and "
        print "that you are accessing the SOAP services directly and not "
        print "through apache."
        print
        print "For help, please visit the CSFE Forum at:"
        print " http://forums.open.collab.net/ds/viewForumSummary.do?dsForumId=738"
        print " or open a support case via your usual channels."
        print
        _exit(1)

    # Establish a valid session
    global username
    global loginhash
    print "Please provide login information"
    username = raw_input("Username: ")
    password = gp.getpass('Password: ')

    # log in
    try:
        loginhash = collabnet_service.service.login(username, password)
    except:
        _log(LOG_ERROR, " Unable to login to the server with the provided credentials")
        _log(LOG_ERROR, " Please check that you entered your Username and Password " \
              "correctly, and try again.")
        _exit(1)

    # get access to the necessary services
    try:
        tracker_service = Client(trackerwsdlfile, cache=sudsCache, location=trackerwsdlfile)
    except:
        _log(LOG_ERROR, " Unexpected error in grabbing Tracker WSDL File.")
        _log(LOG_ERROR, " It might help to try this manually: %s" % trackerwsdlfile)
        _exit(2)
    try:
        document_service = Client(documentwsdlfile, cache=sudsCache, location=documentwsdlfile)
    except:
        _log(LOG_ERROR, " Unexpected error in grabbing Document WSDL File.")
        _log(LOG_ERROR, " It might help to try this manually: %s" % documentwsdlfile)
        _exit(2)
    try:
        discussion_service = Client(discussionwsdlfile, cache=sudsCache, location=discussionwsdlfile)
    except:
        _log(LOG_ERROR, " Unexpected error in grabbing Discussion WSDL File.")
        _log(LOG_ERROR, " It might help to try this manually: %s" % discussionwsdlfile)
        _exit(2)
    try:
        wiki_service = Client(wikiwsdlfile, cache=sudsCache, location=wikiwsdlfile)
    except:
        _log(LOG_ERROR, " Unexpected error in grabbing Wiki WSDL File.")
        _log(LOG_ERROR, " It might help to try this manually: %s" % wikiwsdlfile)
        _exit(2)
    try:
        frs_service = Client(frswsdlfile, cache=sudsCache, location=frswsdlfile)
    except:
        _log(LOG_ERROR, " Unexpected error in grabbing FRS WSDL File.")
        _log(LOG_ERROR, " It might help to try this manually: %s" % frswsdlfile)
        _exit(2)
    try:
        scm_service = Client(scmwsdlfile, cache=sudsCache, location=scmwsdlfile)
    except:
        _log(LOG_ERROR, " Unexpected error in grabbing SCM WSDL File.")
        _log(LOG_ERROR, " It might help to try this manually: %s" % scmwsdlfile)
        _exit(2)
    try:
        page_service = Client(pagewsdlfile, cache=sudsCache, location=pagewsdlfile)
    except:
        _log(LOG_ERROR, " Unexpected error in grabbing Page WSDL File.")
        _log(LOG_ERROR, " It might help to try this manually: %s" % pagewsdlfile)
        _exit(2)

    try:
        news_service = Client(newswsdlfile, cache=sudsCache, location=newswsdlfile)
    except:
        _log(LOG_ERROR, " Unexpected error in grabbing News WSDL File.")
        _log(LOG_ERROR, " It might help to try this manually: %s" % newswsdlfile)
        _exit(2)

    if verbose:
        _log(LOG_INFO, "")
        _log(LOG_INFO, "SOAP services are now available.\n")

def _processItems(type, force):
    """ This will issue a request for re-indexing. For artifacts, news, posts and news, it identifies the items
    in need of re-indexing.  and issue a request only against items that the server
        isn't already making available through search. (Unless the option --force-index is used)

        Supports type = artifact, tracker, forum, topic, post, news, commit, package, frs file, release, document
        , document folder, wiki and project page.
    """
    #verify that type is supported
    if type not in SUPPORTED_TYPES:
        _log(LOG_ERROR, " %s is not a supported type for processing" % type)
        _exit(2)

    _log(LOG_INFO, "")
    _log(LOG_INFO, " BEGIN %s indexing" % type.capitalize())
    checklist = []
    item = None

    # either process all projects on the site or just the one asked for
    if project_id:
        project_list = [collabnet_service.service.getProjectData(loginhash, project_id)]
    else:
        projects = collabnet_service.service.getProjectList(loginhash, False)
        project_list = projects.dataRows

    # get items by project and process them one by one
    for project in project_list:
        item_list = None
        try:
            _log(LOG_INFO, " Processing %ss for project (%s) %s" % (type, project.id, project.title))
            if (type == ARTIFACT_TYPE):
                item_list = tracker_service.service.getArtifactList(loginhash, project.id, None)
            elif (type == TRACKER_TYPE):
                item_list = tracker_service.service.getTrackerList(loginhash, project.id)
            elif (type == DOCUMENT_TYPE):
                # get all folders starting with top-level folder
                folder_list = document_service.service.getDocumentFolderList(loginhash, project.id, True)
                # aggregate all documents stored in each of the project folders
                for folder in folder_list.dataRows:
                    curr_list = document_service.service.getDocumentList(loginhash, folder.id, None)

                    if item_list:
                        item_list["dataRows"].extend(curr_list["dataRows"]) # pylint: disable=unsubscriptable-object
                    else:
                        item_list = curr_list
            if (type == DOCUMENT_FOLDER_TYPE):
                item_list = document_service.service.getDocumentFolderList(loginhash, project.id, True)
            elif (type == FORUM_TYPE):
                 # get all forum
                item_list = discussion_service.service.getForumList(loginhash, project.id)
            elif (type == POST_TYPE):
                 # get all forum
                forum_list = discussion_service.service.getForumList(loginhash, project.id)
                # get all topics
                for forum in forum_list.dataRows:
                    topic_list = discussion_service.service.getTopicList(loginhash,forum.id)
                    #get all posts
                    for topic in topic_list.dataRows:
                      curr_list = discussion_service.service.getPostList(loginhash, topic.id)

                      if item_list:
                          item_list["dataRows"].extend(curr_list["dataRows"])
                      else:
                          item_list = curr_list
            elif (type == TOPIC_TYPE):
                 # get all forum
                forum_list = discussion_service.service.getForumList(loginhash, project.id)
                # get all topics
                for forum in forum_list.dataRows:
                    curr_list = discussion_service.service.getTopicList(loginhash,forum.id)
                    if item_list:
                          item_list["dataRows"].extend(curr_list["dataRows"])
                    else:
                          item_list = curr_list
            elif (type == PACKAGE_TYPE):
                 # get all packages
                item_list = frs_service.service.getPackageList(loginhash, project.id)
            elif (type == FRS_FILE_TYPE):
                 # get all packages
                package_list = frs_service.service.getPackageList(loginhash, project.id)
                # get all releases
                for package in package_list.dataRows:
                    release_list = frs_service.service.getReleaseList(loginhash,package.id)
                    #get all frs files
                    for release in release_list.dataRows:
                      curr_list = frs_service.service.getFrsFileList(loginhash, release.id)

                      if item_list:
                          item_list["dataRows"].extend(curr_list["dataRows"])
                      else:
                          item_list = curr_list
            elif (type == RELEASE_TYPE):
                # get all packages
                package_list = frs_service.service.getPackageList(loginhash, project.id)
                # get all releases
                for package in package_list.dataRows:
                    curr_list = frs_service.service.getReleaseList(loginhash,package.id)
                    if item_list:
                          item_list["dataRows"].extend(curr_list["dataRows"])
                    else:
                          item_list = curr_list
            elif (type == SCM_COMMIT_TYPE):
                 # get all repositories
                repository_list = scm_service.service.getRepositoryList(loginhash, project.id)
                 # get all commits
                for repository in repository_list.dataRows:
                    curr_list = scm_service.service.getCommitList(loginhash,repository.id)
                    if item_list:
                          item_list["dataRows"].extend(curr_list["dataRows"])
                    else:
                          item_list = curr_list
            elif (type == WIKI_TYPE):
                 # get all wiki pages
                item_list = wiki_service.service.getWikiPageList(loginhash, project.id)
            elif (type == PROJECT_PAGE_TYPE):
                 # get all project pages
                item_list = page_service.service.getPageList(loginhash, project.id)
            elif (type == NEWS_POST_TYPE):
                 # get all news post pages
                item_list = news_service.service.getNewsPostList(loginhash, project.id)

            # now search and re-index, when necessary, every item for the current project
            if (item_list is not None):
              for item in item_list.dataRows:
                start_time = time.time()
                if verbose:
                    _log(LOG_INFO, " Processing %s" % item.id)

                # now find the item by its title. we're not searching for it by ID because
                # it may be the case that searching by ID doesn't actually go to the search
                # index but instead to some in-memory cache.
                if (not force and type == ARTIFACT_TYPE):
                    checklist = tracker_service.service.findArtifacts(loginhash, "\"%s\"" % item.title,
                                                              '', False)
                elif (not force and type == DOCUMENT_TYPE):
                    checklist = document_service.service.findDocuments(loginhash, "\"%s\"" % item.title,
                                                               project.id, True, True)
                elif (not force and type == POST_TYPE):
                    checklist = discussion_service.service.findPosts(loginhash, "\"%s\"" % item.title,
                                                               project.id, False)
                elif (not force and type == NEWS_POST_TYPE):
                    checklist = news_service.service.findNewsPosts(loginhash, "\"%s\"" % item.title,
                                                               project.id, False)


                # if we find it there's no need to re-index it
                if (force):
                    _log(LOG_ATTENTION, " %s will be reindexed" % item.id)
                    # schedule this item for re-indexing
                    if not dry_run:
                        _reIndexItem(item)
                elif len([r for r in checklist.dataRows if r.id == item.id]) > 0:
                     _log(LOG_SUCCESS, " %s is searchable" % item.id)
                else:
                    _log(LOG_ATTENTION, " %s is NOT searchable" % item.id)
                    # schedule this item for re-indexing
                    if not dry_run:
                        _reIndexItem(item)

                # throttle requests when they're being handled faster than the threshold
                if (time.time() - start_time) < 1.0/REQUESTS_PER_SECOND_THROTTLE:
                    time.sleep(1.0/REQUESTS_PER_SECOND_THROTTLE)

        except Exception as e:
            if item:
                _log(LOG_ERROR, " Unexpected error. %s was handed to us by the server but " \
                      "something went wrong on our way to searching for it." % item.id)
                print e
            else:
                _log(LOG_ERROR, " Unexpected error!")
                print e
    _log(LOG_INFO, " END %s indexing" % type.capitalize())
def _reIndexItemByID(item_id):
    """ Re-indexes the item referenced by the given ID. Certain
        item types can be verified and some are not yet supported.
    """
    if item_id.find("artf") == 0 or item_id.find("doc") == 0 or item_id.find("docf") == 0 or item_id.find("tracker") == 0\
    or item_id.find("forum") == 0 or item_id.find("topc") == 0 or item_id.find("news") == 0 or item_id.find("wiki") == 0\
    or item_id.find("page") == 0 or item_id.find("rel") == 0 or item_id.find("cmmt") == 0 or item_id.find("pkg") == 0\
    or item_id.find("post") == 0 or item_id.find("frs") == 0:
        _reIndexItem(_getItemData(item_id))
    else:
        collabnet_service.service.reindexObject(loginhash, item_id)
        _log(LOG_INFO, " %s was scheduled for re-indexing" % item_id)

        if verify_indexed:
            _log(LOG_ERROR, " Cannot verify %s. Its type is not yet supported" % item_id)


def _reIndexItem(item_data):
    """ Re-indexes the given item
    """
    if item_data is None:
        return False
    try:
        starttime = time.time()
        collabnet_service.service.reindexObject(loginhash, item_data.id)
        _log(LOG_INFO, " %s was scheduled for re-indexing" % item_data.id)

        if verify_indexed:
            # Now that we've submitted it for reindexing, let's wait until
            # we're completely sure that it returns in the search queries...
            _log(LOG_INFO, " Now searching for %s..." % item_data.id)
            counter = 1

            while True:
                if item_data.id.find("artf") == 0:
                    checklist = tracker_service.service.findArtifacts(loginhash, "\"%s\"" % item_data.title,
                                                              '', False)
                elif item_data.id.find("doc") == 0:
                    # note that this does a global search, not just within a project
                    checklist = document_service.service.findDocuments(loginhash, "\"%s\"" % item_data.title,
                                                               None, True, True)
                else:
                    _log(LOG_ERROR, " Cannot verify %s. Its type is not yet supported" % item_data.id)

                if len([r for r in checklist.dataRows if r.id == item_data.id]) == 0:
                    counter += 1
                    time.sleep(3)

                    # report every once in a while that we're still waiting around for item
                    # to be available on the server. Give up after abouta minute.
                    if counter % 3 == 0:
                        _log(LOG_WARN, " Still waiting for %s to be searchable" % item_data.id)
                    elif counter > 21:
                        elapsedtime = time.time() - starttime
                        _log(LOG_WARN, " Giving up search for %s" % item_data.id)
                        _log(LOG_ATTENTION,
                             " %s was not searchable. Please check manually" % item_data.id)
                        break
                else:
                    # found it! Let's get out of this loop
                    _log(LOG_SUCCESS, " %s is now searchable" % item_data.id)
                    break
    except Exception:
        _log(LOG_ERROR, " Unexpected error. %s was handed to us by the server but " \
              "something went wrong on our way to searching for it." % item_data.id)
        return False

def _reIndexSite():
    """ Send re-index events for every project and user on the site
    """
    _log(LOG_INFO, "")
    _log(LOG_INFO, " Re-indexing Projects")
    projects = collabnet_service.service.getProjectList(loginhash, False)

    for project in projects.dataRows:
        try:
            collabnet_service.service.reindexObject(loginhash, project.id)
            _log(LOG_INFO, " Scheduled project (%s) %s for re-indexing" % (project.id, project.title))
        except:
            if verbose:
                _log(LOG_ERROR, " There was an error re-indexing %s" % project.title)

    _log(LOG_INFO, "")
    _log(LOG_INFO, " Re-indexing Users")
    try:
        user_list = collabnet_service.service.getUserList(loginhash, None)
    except:
        _log(LOG_ERROR, " There was a problem fetching users from the site")
        _exit(2)

    # since there is no pagination available for getting all user list
    # we'll explain here what the throttling approach will be
    _log(LOG_INFO, " User re-indexing will be throttled by sending %s requests " \
          "per second" % int(REQUESTS_PER_SECOND_THROTTLE))
    count = 0
    for user in user_list.dataRows:
        user_data = collabnet_service.service.getUserData(loginhash, user.userName)
        try:
            collabnet_service.service.reindexObject(loginhash, user_data.id)
            _log(LOG_INFO, " Scheduled user %s for re-indexing" % user_data.id)
        except Exception:
            if verbose:
                _log(LOG_ERROR, " There was an error re-indexing %%%s" % user_data.id)

        v = 1.0/REQUESTS_PER_SECOND_THROTTLE
        time.sleep(v)
        count += 1

def _getItemData(item_id):
    """ Retrieves the given item's data from the server.
    """
    try:
        item_data = None
        if item_id.find("artf") == 0:
            item_data = tracker_service.service.getArtifactData(loginhash, item_id)
        elif item_id.find("doc") == 0:
            item_data = document_service.service.getDocumentData(loginhash, item_id, 0)
        else:
            _log(LOG_ERROR, " Unable to get Data for %s. Please check that the item prefix is correct. " \
                  "If the prefix is valid this tool is not yet capable or re-indexing that type" \
                  % item_id)

        return item_data
    except Exception:
        _log(LOG_ERROR, " Was not able to retrieve data for %s" % item_id)
        return None

def _log(level, line):
    """ Prints the given line to stdout with the given log level marker
    """
    if log_file:
        log_file.write("%s%s\n" % (level, line))
    else:
        print "%s%s" % (level, line)

def _exit(code):
    """ Performs proper clean up before exiting the program
    """
    # Make sure we terminate the user session if one was established
    if loginhash:
        _logOff()

    if log_file:
        log_file.write("ENDED at %s\n\n" % datetime.datetime.now())
        log_file.close()

    sys.exit(code)

def _logOff():
    """ Properly terminates the current user session
    """
    _log(LOG_INFO, " Logging off from CollabNet services")
    collabnet_service.service.logoff(username, loginhash)

if __name__ == "__main__":
    main()
